/* -*- mode: c; c-basic-offset: 4; indent-tabs-mode: nil; tab-width: 4 -*-
 * ex: set softtabstop=4 tabstop=8 expandtab shiftwidth=4: *
 * Editor Settings: expandtabs and use 4 spaces for indentation */

/*
 * Copyright Likewise Software    2004-2008
 * All rights reserved.
 *
 * This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 2.1 of the license, or (at
 * your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser
 * General Public License for more details.  You should have received a copy
 * of the GNU Lesser General Public License along with this program.  If
 * not, see <http://www.gnu.org/licenses/>.
 *
 * LIKEWISE SOFTWARE MAKES THIS SOFTWARE AVAILABLE UNDER OTHER LICENSING
 * TERMS AS WELL.  IF YOU HAVE ENTERED INTO A SEPARATE LICENSE AGREEMENT
 * WITH LIKEWISE SOFTWARE, THEN YOU MAY ELECT TO USE THE SOFTWARE UNDER THE
 * TERMS OF THAT SOFTWARE LICENSE AGREEMENT INSTEAD OF THE TERMS OF THE GNU
 * LESSER GENERAL PUBLIC LICENSE, NOTWITHSTANDING THE ABOVE NOTICE.  IF YOU
 * HAVE QUESTIONS, OR WISH TO REQUEST A COPY OF THE ALTERNATE LICENSING
 * TERMS OFFERED BY LIKEWISE SOFTWARE, PLEASE CONTACT LIKEWISE SOFTWARE AT
 * license@likewisesoftware.com
 */

/*
 * Copyright (C) Likewise Software. All rights reserved.
 *
 * Module Name:
 *
 *        samr_crypto.c
 *
 * Abstract:
 *
 *        Remote Procedure Call (RPC) Client Interface
 *
 *        Samr cryptographic functions
 *
 * Authors: Rafal Szczesniak (rafal@likewise.com)
 */

#include "includes.h"


NTSTATUS
SamrCreateNtOwfPassword(
    IN PCWSTR Password,
    OUT PBYTE Hash,
    IN DWORD HashSize
    )
{
    NTSTATUS ntStatus = STATUS_SUCCESS;
    DWORD error = ERROR_SUCCESS;
    size_t passwordLength = 0;
    PBYTE retHash = NULL;

    BAIL_ON_INVALID_PTR(Hash, ntStatus);
    BAIL_ON_INVALID_PTR(Password, ntStatus);

    if (HashSize != 16)
    {
        ntStatus = STATUS_INVALID_PARAMETER;
        BAIL_ON_NT_STATUS(ntStatus);
    }

    error = LwWc16sLen(Password, &passwordLength);
    BAIL_ON_WIN_ERROR(error);

    memset(Hash, 0, HashSize);

    retHash = MD4((PBYTE)Password,
                  sizeof(Password[0]) * passwordLength,
                  Hash);
    if (retHash == NULL)
    {
        ntStatus = STATUS_INTERNAL_ERROR;
    }

error:
    if (ntStatus || error)
    {
        memset(Hash, 0, HashSize);
    }

    if (ntStatus == STATUS_SUCCESS &&
        error != ERROR_SUCCESS)
    {
        ntStatus = LwWin32ErrorToNtStatus(error);
    }

    return ntStatus;

}


NTSTATUS
SamrCreateLmOwfPassword(
    IN PCWSTR Password,
    OUT PBYTE Hash,
    IN DWORD HashSize
    )
{
    NTSTATUS ntStatus = STATUS_SUCCESS;

    BAIL_ON_INVALID_PTR(Hash, ntStatus);
    BAIL_ON_INVALID_PTR(Password, ntStatus);

    if (HashSize != 16)
    {
        ntStatus = STATUS_INVALID_PARAMETER;
        BAIL_ON_NT_STATUS(ntStatus);
    }

    // LmOwfPassword is considered obsolete so it's not set
    memset(Hash, 0, HashSize);

error:
    return ntStatus;
}
