/* -*- mode: c; c-basic-offset: 4; indent-tabs-mode: nil; tab-width: 4 -*-
 * ex: set softtabstop=4 tabstop=8 expandtab shiftwidth=4: *
 * Editor Settings: expandtabs and use 4 spaces for indentation */

/*
 * Copyright (c) Likewise Software.  All rights Reserved.
 *
 * This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 2.1 of the license, or (at
 * your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser
 * General Public License for more details.  You should have received a copy
 * of the GNU Lesser General Public License along with this program.  If
 * not, see <http://www.gnu.org/licenses/>.
 *
 * LIKEWISE SOFTWARE MAKES THIS SOFTWARE AVAILABLE UNDER OTHER LICENSING
 * TERMS AS WELL.  IF YOU HAVE ENTERED INTO A SEPARATE LICENSE AGREEMENT
 * WITH LIKEWISE SOFTWARE, THEN YOU MAY ELECT TO USE THE SOFTWARE UNDER THE
 * TERMS OF THAT SOFTWARE LICENSE AGREEMENT INSTEAD OF THE TERMS OF THE GNU
 * LESSER GENERAL PUBLIC LICENSE, NOTWITHSTANDING THE ABOVE NOTICE.  IF YOU
 * HAVE QUESTIONS, OR WISH TO REQUEST A COPY OF THE ALTERNATE LICENSING
 * TERMS OFFERED BY LIKEWISE SOFTWARE, PLEASE CONTACT LIKEWISE SOFTWARE AT
 * license@likewise.com
 */

/*
 * Copyright (C) Likewise Software. All rights reserved.
 *
 * Module Name:
 *
 *     lwkeytab.h
 *
 * Abstract:
 *
 *     Kerberos Keytab Helper API
 *
 * Authors: Danilo Almeida (dalmeida@likewise.com)
 */

#ifndef __LW_KEYTAB_H__
#define __LW_KEYTAB_H__

#include <lw/types.h>
#include <lw/attrs.h>

LW_BEGIN_EXTERN_C

///
/// Get default principals to store in a computer's keytab.
///
/// @param[in] SamAccountName - The SAM acocunt name of the computer.
/// @param[in] DnsDomainName - The computer's AD domain name in DNS form.
/// @param[in] Fqdn - The computer's FQDN.
/// @param[out] PrincipalNames - Returns array of principal names.  Free with
///    #LwFreeStringArray().
///
/// @return Windows error code
/// @retval ERROR_SUCCESS on success
/// @retval !ERROR_SUCCESS on failure
///
DWORD
LwKrbGetKeytabDefaultComputerPrincipals(
    IN PCSTR SamAccountName,
    IN PCSTR DnsDomainName,
    IN PCSTR Fqdn,
    OUT PSTR** PrincipalNames,
    OUT PULONG PrincipalNamesCount
    );

///
/// Store principals in the keytab.
///
/// @param[in] SamAccountName - The SAM acocunt name of the computer.
/// @param[in] DnsDomainName - The computer's AD domain name in DNS form.
/// @param[in] KeyVersionNumber - The computer password's key version number.
/// @param[in] Password - The computer account's password.
/// @param[in] KeytabName - Optionally specifies non-default keytab.
/// @param[in] PrincipalNames -  Array of rincipal names to store.
/// @param[in] PricipalNamesCount - Elements in PricipalNames array.
///
/// @return Windows error code
/// @retval ERROR_SUCCESS on success
/// @retval !ERROR_SUCCESS on failure
///
DWORD
LwKrbAddKeytabComputerPrincipals(
    IN PCSTR SamAccountName,
    IN PCSTR DnsDomainName,
    IN ULONG KeyVersionNumber,
    IN PCSTR Password,
    IN PCSTR OPTIONAL KeytabName,
    IN PSTR* PrincipalNames,
    IN ULONG PrincipalNamesCount
    );

LW_END_EXTERN_C

#endif /* __LW_KEYTAB_H__ */
