/* Editor Settings: expandtabs and use 4 spaces for indentation
 * ex: set softtabstop=4 tabstop=8 expandtab shiftwidth=4: *
 * -*- mode: c, c-basic-offset: 4 -*- */

/*
 * Copyright Likewise Software    2004-2008
 * All rights reserved.
 *
 * This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 2.1 of the license, or (at
 * your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser
 * General Public License for more details.  You should have received a copy
 * of the GNU Lesser General Public License along with this program.  If
 * not, see <http://www.gnu.org/licenses/>.
 *
 * LIKEWISE SOFTWARE MAKES THIS SOFTWARE AVAILABLE UNDER OTHER LICENSING
 * TERMS AS WELL.  IF YOU HAVE ENTERED INTO A SEPARATE LICENSE AGREEMENT
 * WITH LIKEWISE SOFTWARE, THEN YOU MAY ELECT TO USE THE SOFTWARE UNDER THE
 * TERMS OF THAT SOFTWARE LICENSE AGREEMENT INSTEAD OF THE TERMS OF THE GNU
 * LESSER GENERAL PUBLIC LICENSE, NOTWITHSTANDING THE ABOVE NOTICE.  IF YOU
 * HAVE QUESTIONS, OR WISH TO REQUEST A COPY OF THE ALTERNATE LICENSING
 * TERMS OFFERED BY LIKEWISE SOFTWARE, PLEASE CONTACT LIKEWISE SOFTWARE AT
 * license@likewisesoftware.com
 */

/*
 * Copyright (C) Likewise Software. All rights reserved.
 *
 * Module Name:
 *
 *        cfg.c
 *
 * Abstract:
 *
 *        NTLM Security registry settings
 *
 * Authors: Kyle Stemen <kstemen@likewise.com>
 *
 */

#include "ntlmsrvapi.h"
#include "lsasrvutils.h"

static struct {

    LSA_REFRESH_CALLBACK_HANDLE NtlmRefreshCallbackHandle;

    pthread_mutex_t NtlmConfigMutex;

    NTLM_CONFIG NtlmConfig;
    BOOLEAN NtlmConfigInitialized;
} gNtlmConfigState = {
    .NtlmRefreshCallbackHandle = NULL,
    .NtlmConfigMutex = PTHREAD_MUTEX_INITIALIZER,
    .NtlmConfig = { 0 },
    .NtlmConfigInitialized = FALSE
};

static
DWORD
NtlmReadRegistry(
    OUT PNTLM_CONFIG pConfig
    )
{
    DWORD dwError = LW_ERROR_SUCCESS;
    NTLM_CONFIG config;

    config.bSendNTLMv2 = FALSE;
    config.bSupportUnicode = TRUE;
    config.bSupportNTLM2SessionSecurity = TRUE;
    config.bSupportKeyExchange = TRUE;
    config.bSupport56bit = TRUE;
    config.bSupport128bit = TRUE;

    LSA_CONFIG configItems[] =
    {
        {
            "SendNTLMv2",
            TRUE,
            LsaTypeBoolean,
            0,
            MAXDWORD,
            NULL,
            &config.bSendNTLMv2,
            NULL
        },
        {
            "SupportUnicode",
            TRUE,
            LsaTypeBoolean,
            0,
            MAXDWORD,
            NULL,
            &config.bSupportUnicode,
            NULL
        },
        {
            "SupportNTLM2SessionSecurity",
            TRUE,
            LsaTypeBoolean,
            0,
            MAXDWORD,
            NULL,
            &config.bSupportNTLM2SessionSecurity,
            NULL
        },
        {
            "SupportKeyExchange",
            TRUE,
            LsaTypeBoolean,
            0,
            MAXDWORD,
            NULL,
            &config.bSupportKeyExchange,
            NULL
        },
        {
            "Support56bit",
            TRUE,
            LsaTypeBoolean,
            0,
            MAXDWORD,
            NULL,
            &config.bSupport56bit,
            NULL
        },
        {
            "Support128bit",
            TRUE,
            LsaTypeBoolean,
            0,
            MAXDWORD,
            NULL,
            &config.bSupport128bit,
            NULL
        },
    };

    dwError = LsaProcessConfig(
                "Services\\lsass\\Parameters\\NTLM",
                "Policy\\Services\\lsass\\Parameters\\NTLM",
                configItems,
                sizeof(configItems)/sizeof(configItems[0]));
    BAIL_ON_LSA_ERROR(dwError);

    *pConfig = config;

cleanup:
    return dwError;

error:
    goto cleanup;
}

DWORD
NtlmGetConfiguration(
    OUT PNTLM_CONFIG pConfig
    )
{
    DWORD dwError = 0;
    BOOLEAN bConfigLocked = FALSE;

    NTLM_LOCK_MUTEX(bConfigLocked, &gNtlmConfigState.NtlmConfigMutex);
    if (!gNtlmConfigState.NtlmConfigInitialized)
    {
        dwError = NtlmReadRegistry(&gNtlmConfigState.NtlmConfig);
        BAIL_ON_LSA_ERROR(dwError);
        gNtlmConfigState.NtlmConfigInitialized = TRUE;
    }

    *pConfig = gNtlmConfigState.NtlmConfig;

error:

    NTLM_UNLOCK_MUTEX(bConfigLocked, &gNtlmConfigState.NtlmConfigMutex);

    return dwError;
}

static
DWORD
NtlmRefreshConfigurationCallback(
    IN PVOID pContext
    )
{
    DWORD dwError = 0;
    BOOLEAN bConfigLocked = FALSE;

    NTLM_LOCK_MUTEX(bConfigLocked, &gNtlmConfigState.NtlmConfigMutex);
    dwError = NtlmReadRegistry(&gNtlmConfigState.NtlmConfig);
    gNtlmConfigState.NtlmConfigInitialized = TRUE;
    BAIL_ON_LSA_ERROR(dwError);

error:

    NTLM_UNLOCK_MUTEX(bConfigLocked, &gNtlmConfigState.NtlmConfigMutex);

    return dwError;
}

DWORD
NtlmServerInitialize(
    VOID
    )
{
    DWORD dwError = 0;

    dwError = LsaSrvRefreshRegisterCallback(
                  NtlmRefreshConfigurationCallback,
                  NULL,
                  &gNtlmConfigState.NtlmRefreshCallbackHandle);
    BAIL_ON_LSA_ERROR(dwError);

error:

    return dwError;
}

VOID
NtlmServerShutdown(
    VOID
    )
{
    LsaSrvRefreshUnregisterCallback(gNtlmConfigState.NtlmRefreshCallbackHandle);
    gNtlmConfigState.NtlmRefreshCallbackHandle = NULL;
}
