/* Editor Settings: expandtabs and use 4 spaces for indentation
 * ex: set softtabstop=4 tabstop=8 expandtab shiftwidth=4: *
 */

/*
 * Copyright Likewise Software    2004-2009
 * All rights reserved.
 *
 * This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 2.1 of the license, or (at
 * your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser
 * General Public License for more details.  You should have received a copy
 * of the GNU Lesser General Public License along with this program.  If
 * not, see <http://www.gnu.org/licenses/>.
 *
 * LIKEWISE SOFTWARE MAKES THIS SOFTWARE AVAILABLE UNDER OTHER LICENSING
 * TERMS AS WELL.  IF YOU HAVE ENTERED INTO A SEPARATE LICENSE AGREEMENT
 * WITH LIKEWISE SOFTWARE, THEN YOU MAY ELECT TO USE THE SOFTWARE UNDER THE
 * TERMS OF THAT SOFTWARE LICENSE AGREEMENT INSTEAD OF THE TERMS OF THE GNU
 * LESSER GENERAL PUBLIC LICENSE, NOTWITHSTANDING THE ABOVE NOTICE.  IF YOU
 * HAVE QUESTIONS, OR WISH TO REQUEST A COPY OF THE ALTERNATE LICENSING
 * TERMS OFFERED BY LIKEWISE SOFTWARE, PLEASE CONTACT LIKEWISE SOFTWARE AT
 * license@likewisesoftware.com
 */

/*
 * Copyright (C) Likewise Software. All rights reserved.
 *
 * Module Name:
 *
 *        samr_getmembersinalias.c
 *
 * Abstract:
 *
 *        Remote Procedure Call (RPC) Server Interface
 *
 *        SamrGetMembersInAlias function
 *
 * Authors: Rafal Szczesniak (rafal@likewise.com)
 */

#include "includes.h"


NTSTATUS
SamrSrvGetMembersInAlias(
    IN  handle_t        hBinding,
    IN  ACCOUNT_HANDLE  hAlias,
    OUT SID_ARRAY      *pSids
    )
{
    NTSTATUS ntStatus = STATUS_SUCCESS;
    DWORD dwError = ERROR_SUCCESS;
    PACCOUNT_CONTEXT pAcctCtx = NULL;
    DWORD dwMembersNum = 0;
    PSID pGroupSid = NULL;
    PSID *ppMemberSids = NULL;
    DWORD i = 0;

    pAcctCtx = (PACCOUNT_CONTEXT)hAlias;

    if (pAcctCtx == NULL || pAcctCtx->Type != SamrContextAccount)
    {
        ntStatus = STATUS_INVALID_HANDLE;
        BAIL_ON_NT_STATUS(ntStatus);
    }

    if (!(pAcctCtx->dwAccessGranted & ALIAS_ACCESS_GET_MEMBERS))
    {
        ntStatus = STATUS_ACCESS_DENIED;
        BAIL_ON_NT_STATUS(ntStatus);
    }

    pGroupSid = pAcctCtx->pSid;
    if (pGroupSid == NULL)
    {
        ntStatus = STATUS_INVALID_SID;
        BAIL_ON_NT_STATUS(ntStatus);
    }

    ntStatus = SamrSrvAllocateMemory(OUT_PPVOID(&ppMemberSids),
                                     sizeof(*ppMemberSids));
    BAIL_ON_NT_STATUS(ntStatus);

    dwError = SamDbQueryMembers(
                         pGroupSid,
                         &dwMembersNum,
                         &ppMemberSids);
    BAIL_ON_LSA_ERROR(dwError);

    ntStatus = SamrSrvAllocateMemory(OUT_PPVOID(&pSids->pSids),
                                     dwMembersNum * sizeof(*pSids->pSids));
    BAIL_ON_NT_STATUS(ntStatus);

    pSids->dwNumSids = 0;

    for (i = 0; i < dwMembersNum; i++)
    {
        ntStatus = SamrSrvDuplicateSid(&pSids->pSids[i].pSid,
                                       ppMemberSids[i]);
        BAIL_ON_NT_STATUS(ntStatus);
        pSids->dwNumSids++;
    }

error:
    if (ntStatus || dwError)
    {
        if (pSids->pSids)
        {
            for (i = 0; i < pSids->dwNumSids; i++)
            {
                SamDbFreeSidList(1, &pSids->pSids[i].pSid);
            }

            SamrSrvFreeMemory(pSids->pSids);
        }

        pSids->dwNumSids = 0;
        pSids->pSids = NULL;
    }

    if (ppMemberSids)
    {
        SamDbFreeSidList(dwMembersNum, ppMemberSids);
    }

    if (ntStatus == STATUS_SUCCESS &&
        dwError != ERROR_SUCCESS)
    {
        ntStatus = LwWin32ErrorToNtStatus(dwError);
    }

    return ntStatus;
}



/*
local variables:
mode: c
c-basic-offset: 4
indent-tabs-mode: nil
tab-width: 4
end:
*/
