/* Editor Settings: expandtabs and use 4 spaces for indentation
 * ex: set softtabstop=4 tabstop=8 expandtab shiftwidth=4: *
 */

/*
 * Copyright Likewise Software    2004-2009
 * All rights reserved.
 *
 * This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 2.1 of the license, or (at
 * your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser
 * General Public License for more details.  You should have received a copy
 * of the GNU Lesser General Public License along with this program.  If
 * not, see <http://www.gnu.org/licenses/>.
 *
 * LIKEWISE SOFTWARE MAKES THIS SOFTWARE AVAILABLE UNDER OTHER LICENSING
 * TERMS AS WELL.  IF YOU HAVE ENTERED INTO A SEPARATE LICENSE AGREEMENT
 * WITH LIKEWISE SOFTWARE, THEN YOU MAY ELECT TO USE THE SOFTWARE UNDER THE
 * TERMS OF THAT SOFTWARE LICENSE AGREEMENT INSTEAD OF THE TERMS OF THE GNU
 * LESSER GENERAL PUBLIC LICENSE, NOTWITHSTANDING THE ABOVE NOTICE.  IF YOU
 * HAVE QUESTIONS, OR WISH TO REQUEST A COPY OF THE ALTERNATE LICENSING
 * TERMS OFFERED BY LIKEWISE SOFTWARE, PLEASE CONTACT LIKEWISE SOFTWARE AT
 * license@likewisesoftware.com
 */

/*
 * Copyright (C) Likewise Software. All rights reserved.
 *
 * Module Name:
 *
 *        samr_queryaliasinfo.c
 *
 * Abstract:
 *
 *        Remote Procedure Call (RPC) Server Interface
 *
 *        SamrQueryAliasInfo function
 *
 * Authors: Rafal Szczesniak (rafal@likewise.com)
 */

#include "includes.h"


static
NTSTATUS
SamrFillAliasInfo1(
    PACCOUNT_CONTEXT pAcctCtx,
    DWORD dwNumMembers,
    AliasInfo *pInfo
    );


static
NTSTATUS
SamrFillAliasInfo2(
    PACCOUNT_CONTEXT pAcctCtx,
    AliasInfo *pInfo
    );


static
NTSTATUS
SamrFillAliasInfo3(
    PACCOUNT_CONTEXT pAcctCtx,
    AliasInfo *pInfo
    );


NTSTATUS
SamrSrvQueryAliasInfo(
    IN handle_t hBinding,
    IN ACCOUNT_HANDLE hAlias,
    IN UINT16 level,
    OUT AliasInfo **info
    )
{
    NTSTATUS ntStatus = STATUS_SUCCESS;
    DWORD dwError = ERROR_SUCCESS;
    PACCOUNT_CONTEXT pAcctCtx = NULL;
    DWORD dwNumMembers = 0;
    PSID *ppMembers = NULL;
    AliasInfo *pAliasInfo = NULL;

    pAcctCtx = (PACCOUNT_CONTEXT)hAlias;

    if (pAcctCtx == NULL || pAcctCtx->Type != SamrContextAccount)
    {
        ntStatus = STATUS_INVALID_HANDLE;
        BAIL_ON_NT_STATUS(ntStatus);
    }

    if (!(pAcctCtx->dwAccessGranted & ALIAS_ACCESS_LOOKUP_INFO))
    {
        ntStatus = STATUS_ACCESS_DENIED;
        BAIL_ON_NT_STATUS(ntStatus);
    }

    if (level == ALIAS_INFO_ALL)
    {
        dwError = SamDbQueryMembers(pAcctCtx->pSid, &dwNumMembers, &ppMembers);
        BAIL_ON_LSA_ERROR(dwError);

        if (dwNumMembers)
        {
            SamDbFreeSidList(dwNumMembers, ppMembers);
        }
    }

    ntStatus = SamrSrvAllocateMemory((void**)&pAliasInfo,
                                     sizeof(*pAliasInfo));
    BAIL_ON_NT_STATUS(ntStatus);

    switch (level)
    {
    case ALIAS_INFO_ALL:
        ntStatus = SamrFillAliasInfo1(pAcctCtx, dwNumMembers, pAliasInfo);
        break;

    case ALIAS_INFO_NAME:
        ntStatus = SamrFillAliasInfo2(pAcctCtx, pAliasInfo);
        break;

    case ALIAS_INFO_DESCRIPTION:
        ntStatus = SamrFillAliasInfo3(pAcctCtx, pAliasInfo);
        break;

    default:
        ntStatus = STATUS_INVALID_INFO_CLASS;
    }

    BAIL_ON_NT_STATUS(ntStatus);

    *info = pAliasInfo;

error:
    if (ntStatus)
    {
        // NOTE: We are freeing allocated strings (name, domain) on error
        //       within each SamrFillAliasInfoX().  If we add other error paths,
        //       we may need to free these here too.
        SamrSrvFreeMemory(pAliasInfo);
        *info = NULL;
    }

    return ntStatus;
}


static
NTSTATUS
SamrFillAliasInfo1(
    PACCOUNT_CONTEXT pAcctCtx,
    DWORD dwNumMembers,
    AliasInfo *pInfo
    )
{
    NTSTATUS ntStatus = STATUS_SUCCESS;
    AliasInfoAll *pInfoAll = NULL;

    pInfoAll = &(pInfo->all);

    /* name */
    ntStatus = SamrSrvInitUnicodeString(&pInfoAll->name, pAcctCtx->pwszName);
    BAIL_ON_NT_STATUS(ntStatus);

    /* num_members */
    pInfoAll->num_members = dwNumMembers;

    /* description */
    ntStatus = SamrSrvInitUnicodeString(&pInfoAll->description,
                                        pAcctCtx->pwszDescription);
    BAIL_ON_NT_STATUS(ntStatus);

error:
    if (ntStatus)
    {
        SamrSrvFreeUnicodeString(&pInfoAll->name);
        SamrSrvFreeUnicodeString(&pInfoAll->description);

        memset(pInfo, 0, sizeof(*pInfo));
    }

    return ntStatus;
}


static
NTSTATUS
SamrFillAliasInfo2(
    PACCOUNT_CONTEXT pAcctCtx,
    AliasInfo *pInfo
    )
{
    NTSTATUS ntStatus = STATUS_SUCCESS;

    /* name */
    ntStatus = SamrSrvInitUnicodeString(&pInfo->name, pAcctCtx->pwszName);
    BAIL_ON_NT_STATUS(ntStatus);

error:
    if (ntStatus)
    {
        SamrSrvFreeUnicodeString(&pInfo->name);

        memset(pInfo, 0, sizeof(*pInfo));
    }

    return ntStatus;
}


static
NTSTATUS
SamrFillAliasInfo3(
    PACCOUNT_CONTEXT pAcctCtx,
    AliasInfo *pInfo
    )
{
    NTSTATUS ntStatus = STATUS_SUCCESS;

    /* description */
    ntStatus = SamrSrvInitUnicodeString(&pInfo->description,
                                        pAcctCtx->pwszDescription);
    BAIL_ON_NT_STATUS(ntStatus);

error:
    if (ntStatus)
    {
        SamrSrvFreeUnicodeString(&pInfo->description);

        memset(pInfo, 0, sizeof(*pInfo));
    }

    return ntStatus;
}





/*
local variables:
mode: c
c-basic-offset: 4
indent-tabs-mode: nil
tab-width: 4
end:
*/
