/* Editor Settings: expandtabs and use 4 spaces for indentation
 * ex: set softtabstop=4 tabstop=8 expandtab shiftwidth=4: *
 */

/*
 * Copyright Likewise Software    2004-2009
 * All rights reserved.
 *
 * This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 2.1 of the license, or (at
 * your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser
 * General Public License for more details.  You should have received a copy
 * of the GNU Lesser General Public License along with this program.  If
 * not, see <http://www.gnu.org/licenses/>.
 *
 * LIKEWISE SOFTWARE MAKES THIS SOFTWARE AVAILABLE UNDER OTHER LICENSING
 * TERMS AS WELL.  IF YOU HAVE ENTERED INTO A SEPARATE LICENSE AGREEMENT
 * WITH LIKEWISE SOFTWARE, THEN YOU MAY ELECT TO USE THE SOFTWARE UNDER THE
 * TERMS OF THAT SOFTWARE LICENSE AGREEMENT INSTEAD OF THE TERMS OF THE GNU
 * LESSER GENERAL PUBLIC LICENSE, NOTWITHSTANDING THE ABOVE NOTICE.  IF YOU
 * HAVE QUESTIONS, OR WISH TO REQUEST A COPY OF THE ALTERNATE LICENSING
 * TERMS OFFERED BY LIKEWISE SOFTWARE, PLEASE CONTACT LIKEWISE SOFTWARE AT
 * license@likewisesoftware.com
 */

/*
 * Copyright (C) Likewise Software. All rights reserved.
 *
 * Module Name:
 *
 *        samr_setaliasinfo.c
 *
 * Abstract:
 *
 *        Remote Procedure Call (RPC) Server Interface
 *
 *        SamrSrvSetAliasInfo function
 *
 * Authors: Rafal Szczesniak (rafal@likewise.com)
 */

#include "includes.h"


NTSTATUS
SamrSrvSetAliasInfo(
    /* [in] */ handle_t hBinding,
    /* [in] */ ACCOUNT_HANDLE hAlias,
    /* [in] */ UINT16 level,
    /* [in] */ AliasInfo *pInfo
    )
{
    NTSTATUS ntStatus = STATUS_SUCCESS;
    DWORD dwError = ERROR_SUCCESS;
    PACCOUNT_CONTEXT pAcctCtx = NULL;
    SAMDB_OBJECT_INFO updateObject = {0};
    SAMDB_FIELDS updateFields = 0;

    pAcctCtx = (PACCOUNT_CONTEXT)hAlias;

    if (pAcctCtx == NULL || pAcctCtx->Type != SamrContextAccount)
    {
        ntStatus = STATUS_INVALID_HANDLE;
        BAIL_ON_NT_STATUS(ntStatus);
    }

    if (!(pAcctCtx->dwAccessGranted & ALIAS_ACCESS_SET_INFO))
    {
        ntStatus = STATUS_ACCESS_DENIED;
        BAIL_ON_NT_STATUS(ntStatus);
    }

    switch (level)
    {
    case ALIAS_INFO_ALL:
        ntStatus = STATUS_INVALID_INFO_CLASS;
        BAIL_ON_NT_STATUS(ntStatus);
        break;

    case ALIAS_INFO_NAME:
        ntStatus = SamrSrvRenameAccount(hAlias,
                                        &pInfo->name);
        BAIL_ON_NT_STATUS(ntStatus);
        break;

    case ALIAS_INFO_DESCRIPTION:
        if (pAcctCtx->pSid == NULL)
        {
            ntStatus = STATUS_INVALID_HANDLE;
            BAIL_ON_NT_STATUS(ntStatus);
        }

        updateObject.Type = SAMDB_OBJECT_TYPE_LOCAL_GROUP;

        ntStatus = RtlDuplicateSid(
                            &updateObject.Sid,
                            pAcctCtx->pSid);
        BAIL_ON_NT_STATUS(ntStatus);

        //
        // Yes, some clients can send an empty string
        //
        if (pInfo->description.Buffer)
        {
            dwError = LwAllocateWc16StringFromUnicodeString(
                                     &updateObject.Info.LocalGroup.Description,
                                     &pInfo->description);
            BAIL_ON_LSA_ERROR(dwError);
        }
        else
        {
            updateObject.Info.LocalGroup.Description = NULL;
        }

        updateFields |= SAMDB_FIELD_LOCAL_GROUP_DESCRIPTION;

        dwError = SamDbUpdateObject(&updateObject, updateFields);
        BAIL_ON_LSA_ERROR(dwError);
        break;
    }

error:
    RTL_FREE(&updateObject.Sid);
    SamDbReleaseObject(&updateObject);

    if (ntStatus == STATUS_SUCCESS &&
        dwError != ERROR_SUCCESS)
    {
        ntStatus = LwWin32ErrorToNtStatus(dwError);
    }

    return ntStatus;
}


/*
local variables:
mode: c
c-basic-offset: 4
indent-tabs-mode: nil
tab-width: 4
end:
*/
