/* -*- mode: c; c-basic-offset: 4; indent-tabs-mode: nil; tab-width: 4 -*-
 * ex: set softtabstop=4 tabstop=8 expandtab shiftwidth=4: *
 * Editor Settings: expandtabs and use 4 spaces for indentation */

/*
 * Copyright Likewise Software
 * All rights reserved.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * for more details.  You should have received a copy of the GNU General
 * Public License along with this program.  If not, see
 * <http://www.gnu.org/licenses/>.
 *
 * LIKEWISE SOFTWARE MAKES THIS SOFTWARE AVAILABLE UNDER OTHER LICENSING
 * TERMS AS WELL.  IF YOU HAVE ENTERED INTO A SEPARATE LICENSE AGREEMENT
 * WITH LIKEWISE SOFTWARE, THEN YOU MAY ELECT TO USE THE SOFTWARE UNDER THE
 * TERMS OF THAT SOFTWARE LICENSE AGREEMENT INSTEAD OF THE TERMS OF THE GNU
 * GENERAL PUBLIC LICENSE, NOTWITHSTANDING THE ABOVE NOTICE.  IF YOU
 * HAVE QUESTIONS, OR WISH TO REQUEST A COPY OF THE ALTERNATE LICENSING
 * TERMS OFFERED BY LIKEWISE SOFTWARE, PLEASE CONTACT LIKEWISE SOFTWARE AT
 * license@likewisesoftware.com
 */

/*
 * Copyright (C) Likewise Software. All rights reserved.
 *
 * Module Name:
 *
 *        samdbtrans.c
 *
 * Abstract:
 *
 *
 *      Likewise SAM Database Provider
 *
 *      SQL Transactions
 *
 * Authors: Krishna Ganugapati (krishnag@likewisesoftware.com)
 * Authors: Rafal Szczesniak (rafal@likewise.com)
 *
 */

#include "includes.h"


DWORD
SamDbBeginTransaction(
    PSAMDB_CONTEXT DbContext
    )
{
    DWORD error = ERROR_SUCCESS;
    PSTR errorMessage = NULL;

    error = sqlite3_exec(
                    DbContext->DbHandle,
                    "BEGIN",
                    NULL,
                    NULL,
                    &errorMessage);
    BAIL_ON_SAMDB_ERROR(error);

cleanup:
    if (error)
    {
        SAMDB_LOG_ERROR("Sqlite3 error when beginning transaction "
                        "(code: %u): %s",
                        error,
                        LSA_SAFE_LOG_STRING(errorMessage));
        //
        // If sqlite can't process transactions then it's
        // an internal error
        //
        error = ERROR_INTERNAL_ERROR;
    }

    if (errorMessage)
    {
        sqlite3_free(errorMessage);
    }

    return error;
}


DWORD
SamDbRollbackTransaction(
    PSAMDB_CONTEXT DbContext
    )
{
    DWORD error = ERROR_SUCCESS;
    PSTR errorMessage = NULL;

    error = sqlite3_exec(
                    DbContext->DbHandle,
                    "ROLLBACK",
                    NULL,
                    NULL,
                    &errorMessage);
    BAIL_ON_SAMDB_ERROR(error);

cleanup:
    if (error)
    {
        SAMDB_LOG_ERROR("Sqlite3 error when rolling transaction back "
                        "(code: %u): %s",
                        error,
                        LSA_SAFE_LOG_STRING(errorMessage));
        //
        // If sqlite can't process transactions then it's
        // an internal error
        //
        error = ERROR_INTERNAL_ERROR;
    }

    if (errorMessage)
    {
        sqlite3_free(errorMessage);
    }

    return error;
}


DWORD
SamDbCommitTransaction(
    PSAMDB_CONTEXT DbContext
    )
{
    DWORD error = ERROR_SUCCESS;
    PSTR errorMessage = NULL;

    error = sqlite3_exec(
                    DbContext->DbHandle,
                    "COMMIT",
                    NULL,
                    NULL,
                    &errorMessage);
    BAIL_ON_SAMDB_ERROR(error);

cleanup:
    if (error)
    {
        SAMDB_LOG_ERROR("Sqlite3 error when commiting transaction "
                        "(code: %u): %s",
                        error,
                        LSA_SAFE_LOG_STRING(errorMessage));
        //
        // If sqlite can't process transactions then it's
        // an internal error
        //
        error = ERROR_INTERNAL_ERROR;
    }

    if (errorMessage)
    {
        sqlite3_free(errorMessage);
    }

    return error;
}
