/*
 * Copyright (c) Likewise Software.  All rights Reserved.
 *
 * This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 2.1 of the license, or (at
 * your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser
 * General Public License for more details.  You should have received a copy
 * of the GNU Lesser General Public License along with this program.  If
 * not, see <http://www.gnu.org/licenses/>.
 *
 * LIKEWISE SOFTWARE MAKES THIS SOFTWARE AVAILABLE UNDER OTHER LICENSING
 * TERMS AS WELL.  IF YOU HAVE ENTERED INTO A SEPARATE LICENSE AGREEMENT
 * WITH LIKEWISE SOFTWARE, THEN YOU MAY ELECT TO USE THE SOFTWARE UNDER THE
 * TERMS OF THAT SOFTWARE LICENSE AGREEMENT INSTEAD OF THE TERMS OF THE GNU
 * LESSER GENERAL PUBLIC LICENSE, NOTWITHSTANDING THE ABOVE NOTICE.  IF YOU
 * HAVE QUESTIONS, OR WISH TO REQUEST A COPY OF THE ALTERNATE LICENSING
 * TERMS OFFERED BY LIKEWISE SOFTWARE, PLEASE CONTACT LIKEWISE SOFTWARE AT
 * license@likewisesoftware.com
 */

/*
 * Module Name:
 *
 *        eventlog-types.h
 *
 * Abstract:
 *
 *        Low-level event logging API types
 *
 * Authors: Brian Koropoff (bkoropoff@likewise.com)
 *
 */

#ifndef __LWBASE_EVENTLOG_TYPES_H__
#define __LWBASE_EVENTLOG_TYPES_H__

#include <lw/types.h>
#include <lw/attrs.h>

/**
 * @file eventlog-types.h
 * @brief Eventlog types
 */

/**
 * @brief Event type information
 * @ingroup eventlog
 *
 *
 * Information describing the format of an event type.
 * The particular format is distinguished by an
 * #LW_RTL_EVENTLOG_TYPE_INFO_FORMAT.
 */
typedef LW_PCVOID PLW_RTL_EVENTLOG_TYPE_INFO;

/**
 * @brief Event type information format
 * @ingroup eventlog
 *
 * Distinguishes the format of event type information.
 */
typedef enum _LW_RTL_EVENTLOG_TYPE_INFO_FORMAT
{
    /**
     * @brief Type has no associated information
     */
    LW_RTL_EVENTLOG_TYPE_INFO_NONE,
    /**
     * @brief Type information is an lwmsg type spec
     */
    LW_RTL_EVENTLOG_TYPE_INFO_LWMSG
} LW_RTL_EVENTLOG_TYPE_INFO_FORMAT;

/**
 * @brief Event record
 * @ingroup eventlog
 */
typedef struct _LW_RTL_EVENTLOG_RECORD
{
    LW_LONG64 Timestamp;
    LW_GUID Type;
    LW_PVOID pData;
} LW_RTL_EVENTLOG_RECORD, *PLW_RTL_EVENTLOG_RECORD;

/* Backwards compatibility with old names */
#ifndef DOXYGEN
typedef struct _LW_RTL_EVENTLOG_RECORD LW_EVENTLOG_RECORD, *PLW_EVENTLOG_RECORD;
#endif

#endif
