/* -*- mode: c; c-basic-offset: 4; indent-tabs-mode: nil; tab-width: 4 -*-
 * ex: set softtabstop=4 tabstop=8 expandtab shiftwidth=4: *
 * Editor Settings: expandtabs and use 4 spaces for indentation */

/*
 * Copyright Likewise Software
 * All rights reserved.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * for more details.  You should have received a copy of the GNU General
 * Public License along with this program.  If not, see
 * <http://www.gnu.org/licenses/>.
 *
 * LIKEWISE SOFTWARE MAKES THIS SOFTWARE AVAILABLE UNDER OTHER LICENSING
 * TERMS AS WELL.  IF YOU HAVE ENTERED INTO A SEPARATE LICENSE AGREEMENT
 * WITH LIKEWISE SOFTWARE, THEN YOU MAY ELECT TO USE THE SOFTWARE UNDER THE
 * TERMS OF THAT SOFTWARE LICENSE AGREEMENT INSTEAD OF THE TERMS OF THE GNU
 * GENERAL PUBLIC LICENSE, NOTWITHSTANDING THE ABOVE NOTICE.  IF YOU
 * HAVE QUESTIONS, OR WISH TO REQUEST A COPY OF THE ALTERNATE LICENSING
 * TERMS OFFERED BY LIKEWISE SOFTWARE, PLEASE CONTACT LIKEWISE SOFTWARE AT
 * license@likewisesoftware.com
 */

/*
 * Copyright (C) Likewise Software. All rights reserved.
 *
 * Module Name:
 *
 *        filePositionInfo.c
 *
 * Abstract:
 *
 *        Likewise Posix File System Driver (PVFS)
 *
 *        FilePositionInformation Handler
 *
 * Authors: Sriram Nambakam <snambakam@likewise.com>
 */

#include "pvfs.h"

/* Forward declarations */

static
NTSTATUS
PvfsQueryFilePositionInfo(
    PPVFS_IRP_CONTEXT pIrpContext
    );

static
NTSTATUS
PvfsSetFilePositionInfo(
    IN PPVFS_IRP_CONTEXT pIrpContext
    );

////////////////////////////////////////////////////////////////////////

NTSTATUS
PvfsFilePositionInfo(
    PVFS_INFO_TYPE Type,
    PPVFS_IRP_CONTEXT pIrpContext
    )
{
    NTSTATUS ntError = STATUS_UNSUCCESSFUL;

    switch(Type)
    {
        case PVFS_SET:
            ntError = PvfsSetFilePositionInfo(pIrpContext);
            break;

        case PVFS_QUERY:
            ntError = PvfsQueryFilePositionInfo(pIrpContext);
            break;

        default:
            ntError = STATUS_INVALID_PARAMETER;
            break;
    }
    BAIL_ON_NT_STATUS(ntError);

error:

    return ntError;
}

static
NTSTATUS
PvfsQueryFilePositionInfo(
    PPVFS_IRP_CONTEXT pIrpContext
    )
{
    NTSTATUS ntError = STATUS_UNSUCCESSFUL;
    PIRP pIrp = pIrpContext->pIrp;
    PPVFS_CCB pCcb = NULL;
    PFILE_POSITION_INFORMATION pFileInfo = NULL;
    const IRP_ARGS_QUERY_SET_INFORMATION* pArgs =
        &pIrpContext->pIrp->Args.QuerySetInformation;
    off_t CurrentOffset = 0;

    /* Sanity checks */

    ntError = PvfsAcquireCCB(pIrp->FileHandle, &pCcb);
    BAIL_ON_NT_STATUS(ntError);

    BAIL_ON_INVALID_PTR(pArgs->FileInformation, ntError);

    ntError = PvfsAccessCheckAnyFileHandle(
                  pCcb,
                  (FILE_READ_DATA|FILE_WRITE_DATA));
    BAIL_ON_NT_STATUS(ntError);

    if (pArgs->Length != sizeof(*pFileInfo))
    {
        ntError = LW_STATUS_INFO_LENGTH_MISMATCH;
        BAIL_ON_NT_STATUS(ntError);
    }

    pFileInfo = (PFILE_POSITION_INFORMATION)pArgs->FileInformation;

    /* Real work starts here */

    /* Position */
    ntError = PvfsSysLseek(pCcb->fd, 0, SEEK_CUR, &CurrentOffset);
    BAIL_ON_NT_STATUS(ntError);

    pFileInfo->CurrentByteOffset = CurrentOffset;

    pIrp->IoStatusBlock.BytesTransferred = sizeof(*pFileInfo);

cleanup:

    if (pCcb)
    {
        PvfsReleaseCCB(pCcb);
    }

    return ntError;

error:
    goto cleanup;
}

////////////////////////////////////////////////////////////////////////

static
NTSTATUS
PvfsSetFilePositionInfo(
    IN PPVFS_IRP_CONTEXT pIrpContext
    )
{
    NTSTATUS status = STATUS_SUCCESS;
    PIRP pIrp = pIrpContext->pIrp;
    PPVFS_CCB pCcb = NULL;
    PFILE_POSITION_INFORMATION pFileInfo = NULL;
    const IRP_ARGS_QUERY_SET_INFORMATION* pArgs =
        &pIrpContext->pIrp->Args.QuerySetInformation;

    /* Sanity checks */

    status =  PvfsAcquireCCB(pIrp->FileHandle, &pCcb);
    BAIL_ON_NT_STATUS(status);

    BAIL_ON_INVALID_PTR(pArgs->FileInformation, status);

    status = PvfsAccessCheckAnyFileHandle(
                  pCcb,
                  (FILE_READ_DATA|FILE_WRITE_DATA));
    BAIL_ON_NT_STATUS(status);

    if (pArgs->Length != sizeof(*pFileInfo))
    {
        status = STATUS_INFO_LENGTH_MISMATCH;
        BAIL_ON_NT_STATUS(status);
    }

    pFileInfo = (PFILE_POSITION_INFORMATION)pArgs->FileInformation;

    /* Real work starts here */

    status = PvfsSysLseek(
                 pCcb->fd,
                 (off_t)pFileInfo->CurrentByteOffset,
                 SEEK_SET,
                 NULL);
    BAIL_ON_NT_STATUS(status);

    pIrp->IoStatusBlock.BytesTransferred = sizeof(*pFileInfo);
    status = STATUS_SUCCESS;

error:

    if (pCcb)
    {
        PvfsReleaseCCB(pCcb);
    }

    return status;
}
