/* -*- mode: c; c-basic-offset: 4; indent-tabs-mode: nil; tab-width: 4 -*-
 * ex: set softtabstop=4 tabstop=8 expandtab shiftwidth=4: *
 * Editor Settings: expandtabs and use 4 spaces for indentation */

/*
 * Copyright Likewise Software
 * All rights reserved.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * for more details.  You should have received a copy of the GNU General
 * Public License along with this program.  If not, see
 * <http://www.gnu.org/licenses/>.
 *
 * LIKEWISE SOFTWARE MAKES THIS SOFTWARE AVAILABLE UNDER OTHER LICENSING
 * TERMS AS WELL.  IF YOU HAVE ENTERED INTO A SEPARATE LICENSE AGREEMENT
 * WITH LIKEWISE SOFTWARE, THEN YOU MAY ELECT TO USE THE SOFTWARE UNDER THE
 * TERMS OF THAT SOFTWARE LICENSE AGREEMENT INSTEAD OF THE TERMS OF THE GNU
 * GENERAL PUBLIC LICENSE, NOTWITHSTANDING THE ABOVE NOTICE.  IF YOU
 * HAVE QUESTIONS, OR WISH TO REQUEST A COPY OF THE ALTERNATE LICENSING
 * TERMS OFFERED BY LIKEWISE SOFTWARE, PLEASE CONTACT LIKEWISE SOFTWARE AT
 * license@likewisesoftware.com
 */

/*
 * Copyright (C) Likewise Software. All rights reserved.
 *
 * Module Name:
 *
 *        libmain.c
 *
 * Abstract:
 *
 *        Likewise IO (LWIO) - SRV
 *
 *        Protocols API - SMBV1
 *
 *        Process QUERY_INFORMATION2 request
 *
 * Authors: Sriram Nambakam (snambakam@likewise.com)
 *
 */

#include "includes.h"

static
NTSTATUS
SrvBuildQueryInfo2State(
    PQUERY_INFO2_REQUEST_HEADER    pRequestHeader,
    PLWIO_SRV_FILE                 pFile,
    PSRV_QUERY_INFO2_STATE_SMB_V1* ppInfo2State
    );

static
NTSTATUS
SrvExecuteQueryFileInfo2(
    PSRV_EXEC_CONTEXT pExecContext
    );

static
NTSTATUS
SrvBuildQueryInfo2Response(
    PSRV_EXEC_CONTEXT pExecContext
    );

static
VOID
SrvFreeQueryInfo2StateHandle(
    HANDLE hInfo2State
    );

static
VOID
SrvFreeQueryInfo2State(
    PSRV_QUERY_INFO2_STATE_SMB_V1 pInfo2State
    );

NTSTATUS
SrvProcessQueryInformation2(
    PSRV_EXEC_CONTEXT pExecContext
    )
{
    NTSTATUS                   ntStatus     = 0;
    PLWIO_SRV_CONNECTION       pConnection  = pExecContext->pConnection;
    PSRV_EXEC_CONTEXT_SMB_V1   pCtxSmb1     = pExecContext->pProtocolContext;
    ULONG                      iMsg         = pCtxSmb1->iMsg;
    PSRV_MESSAGE_SMB_V1        pSmbRequest  = &pCtxSmb1->pRequests[iMsg];
    PLWIO_SRV_SESSION          pSession     = NULL;
    PLWIO_SRV_TREE             pTree        = NULL;
    PLWIO_SRV_FILE             pFile        = NULL;
    PSRV_QUERY_INFO2_STATE_SMB_V1 pInfo2State  = pCtxSmb1->hState;

    if (pInfo2State == NULL)
    {
        PBYTE pBuffer          = pSmbRequest->pBuffer + pSmbRequest->usHeaderSize;
        ULONG ulOffset         = pSmbRequest->usHeaderSize;
        ULONG ulBytesAvailable = pSmbRequest->ulMessageSize - pSmbRequest->usHeaderSize;
        PQUERY_INFO2_REQUEST_HEADER pRequestHeader = NULL; // Do not free

        ntStatus = SrvConnectionFindSession_SMB_V1(
                        pCtxSmb1,
                        pConnection,
                        pSmbRequest->pHeader->uid,
                        &pSession);
        BAIL_ON_NT_STATUS(ntStatus);

        ntStatus = SrvSetStatSessionInfo(pExecContext, pSession);
        BAIL_ON_NT_STATUS(ntStatus);

        ntStatus = SrvSessionFindTree_SMB_V1(
                        pCtxSmb1,
                        pSession,
                        pSmbRequest->pHeader->tid,
                        &pTree);
        BAIL_ON_NT_STATUS(ntStatus);

        ntStatus = WireUnmarshalQueryInfo2Request(
                        pBuffer,
                        ulBytesAvailable,
                        ulOffset,
                        &pRequestHeader);
        BAIL_ON_NT_STATUS(ntStatus);

        SRV_LOG_DEBUG(
                pExecContext->pLogContext,
                SMB_PROTOCOL_VERSION_1,
                pSmbRequest->pHeader->command,
                "Query file info: "
                "command(%u),uid(%u),mid(%u),pid(%u),tid(%u),file-id(%u)",
                pSmbRequest->pHeader->command,
                pSmbRequest->pHeader->uid,
                pSmbRequest->pHeader->mid,
                SMB_V1_GET_PROCESS_ID(pSmbRequest->pHeader),
                pSmbRequest->pHeader->tid,
                pRequestHeader->usFid);

        ntStatus = SrvTreeFindFile_SMB_V1(
                        pCtxSmb1,
                        pTree,
                        pRequestHeader->usFid,
                        &pFile);
        BAIL_ON_NT_STATUS(ntStatus);

        ntStatus = SrvBuildQueryInfo2State(
                        pRequestHeader,
                        pFile,
                        &pInfo2State);
        BAIL_ON_NT_STATUS(ntStatus);

        pCtxSmb1->hState = pInfo2State;
        pCtxSmb1->pfnStateRelease = &SrvFreeQueryInfo2StateHandle;
    }

    switch (pInfo2State->stage)
    {
        case SRV_QUERY_INFO2_STAGE_SMB_V1_INITIAL:

            pInfo2State->stage = SRV_QUERY_INFO2_STAGE_SMB_V1_ATTEMPT_IO;

            // intentional fall through

        case SRV_QUERY_INFO2_STAGE_SMB_V1_ATTEMPT_IO:

            ntStatus = SrvExecuteQueryFileInfo2(pExecContext);
            BAIL_ON_NT_STATUS(ntStatus);

            pInfo2State->stage = SRV_QUERY_INFO2_STAGE_SMB_V1_BUILD_RESPONSE;

            // intentional fall through

        case SRV_QUERY_INFO2_STAGE_SMB_V1_BUILD_RESPONSE:

            ntStatus = SrvBuildQueryInfo2Response(pExecContext);
            BAIL_ON_NT_STATUS(ntStatus);

            pInfo2State->stage = SRV_QUERY_INFO2_STAGE_SMB_V1_DONE;

        case SRV_QUERY_INFO2_STAGE_SMB_V1_DONE:

            break;
    }

cleanup:

    if (pFile)
    {
        SrvFileRelease(pFile);
    }

    if (pTree)
    {
        SrvTreeRelease(pTree);
    }

    if (pSession)
    {
        SrvSessionRelease(pSession);
    }

    return ntStatus;

error:

    switch (ntStatus)
    {
        case STATUS_PENDING:

            // TODO: Add an indicator to the file object to trigger a
            //       cleanup if the connection gets closed and all the
            //       files involved have to be closed

            break;

        default:

            SrvReleaseExecContextAsync(pExecContext);

            break;
    }

    goto cleanup;
}

static
NTSTATUS
SrvBuildQueryInfo2State(
    PQUERY_INFO2_REQUEST_HEADER  pRequestHeader,
    PLWIO_SRV_FILE               pFile,
    PSRV_QUERY_INFO2_STATE_SMB_V1* ppInfo2State
    )
{
    NTSTATUS ntStatus = STATUS_SUCCESS;
    PSRV_QUERY_INFO2_STATE_SMB_V1 pInfo2State = NULL;

    ntStatus = SrvAllocateMemory(
                    sizeof(SRV_QUERY_INFO2_STATE_SMB_V1),
                    (PVOID*)&pInfo2State);
    BAIL_ON_NT_STATUS(ntStatus);

    pInfo2State->stage = SRV_QUERY_INFO2_STAGE_SMB_V1_INITIAL;

    pInfo2State->pRequestHeader = pRequestHeader;
    pInfo2State->pFile          = SrvFileAcquire(pFile);

    *ppInfo2State = pInfo2State;

cleanup:

    return ntStatus;

error:

    *ppInfo2State = NULL;

    if (pInfo2State)
    {
        SrvFreeQueryInfo2State(pInfo2State);
    }

    goto cleanup;
}

static
NTSTATUS
SrvExecuteQueryFileInfo2(
    PSRV_EXEC_CONTEXT pExecContext
    )
{
    NTSTATUS                      ntStatus    = 0;
    PSRV_EXEC_CONTEXT_SMB_V1      pCtxSmb1    = pExecContext->pProtocolContext;
    PSRV_QUERY_INFO2_STATE_SMB_V1 pInfo2State = pCtxSmb1->hState;

    ntStatus = pExecContext->ioStatusBlock.Status;
    BAIL_ON_NT_STATUS(ntStatus);

    if (!pInfo2State->pFileBasicInfo)
    {
        pInfo2State->pFileBasicInfo = &pInfo2State->fileBasicInfo;

        SrvPrepareExecContextAsync(pExecContext);

        ntStatus = IoQueryInformationFile(
                        pInfo2State->pFile->hFile,
                        pExecContext->pAcb,
                        &pExecContext->ioStatusBlock,
                        pInfo2State->pFileBasicInfo,
                        sizeof(pInfo2State->fileBasicInfo),
                        FileBasicInformation);
        BAIL_ON_NT_STATUS(ntStatus);

        SrvReleaseExecContextAsync(pExecContext); // completed synchronously
    }

    if (!pInfo2State->pFileStandardInfo)
    {
        pInfo2State->pFileStandardInfo = &pInfo2State->fileStandardInfo;

        SrvPrepareExecContextAsync(pExecContext);

        ntStatus = IoQueryInformationFile(
                        pInfo2State->pFile->hFile,
                        pExecContext->pAcb,
                        &pExecContext->ioStatusBlock,
                        pInfo2State->pFileStandardInfo,
                        sizeof(pInfo2State->fileStandardInfo),
                        FileStandardInformation);
        BAIL_ON_NT_STATUS(ntStatus);

        SrvReleaseExecContextAsync(pExecContext); // completed synchronously
    }

error:

    return ntStatus;
}

static
NTSTATUS
SrvBuildQueryInfo2Response(
    PSRV_EXEC_CONTEXT pExecContext
    )
{
    NTSTATUS                    ntStatus     = 0;
    PLWIO_SRV_CONNECTION        pConnection  = pExecContext->pConnection;
    PSRV_EXEC_CONTEXT_SMB_V1    pCtxSmb1     = pExecContext->pProtocolContext;
    ULONG                       iMsg         = pCtxSmb1->iMsg;
    PSRV_MESSAGE_SMB_V1         pSmbRequest  = &pCtxSmb1->pRequests[iMsg];
    PSRV_MESSAGE_SMB_V1         pSmbResponse = &pCtxSmb1->pResponses[iMsg];
    PQUERY_INFO2_RESPONSE_HEADER pResponseHeader = NULL; // Do not free
    PBYTE pOutBuffer           = pSmbResponse->pBuffer;
    ULONG ulBytesAvailable     = pSmbResponse->ulBytesAvailable;
    ULONG ulOffset             = 0;
    ULONG ulTotalBytesUsed     = 0;
    PSRV_QUERY_INFO2_STATE_SMB_V1 pInfo2State = pCtxSmb1->hState;

    if (!pSmbResponse->ulSerialNum)
    {
        ntStatus = SrvMarshalHeader_SMB_V1(
                        pOutBuffer,
                        ulOffset,
                        ulBytesAvailable,
                        COM_QUERY_INFORMATION2,
                        STATUS_SUCCESS,
                        TRUE,
                        (pSmbRequest->pHeader->flags2 & FLAG2_ERR_STATUS) != 0,
                        pConnection->serverProperties.ulCapabilities,
                        pCtxSmb1->pTree->tid,
                        SMB_V1_GET_PROCESS_ID(pSmbRequest->pHeader),
                        pCtxSmb1->pSession->uid,
                        pSmbRequest->pHeader->mid,
                        &pSmbResponse->pHeader,
                        &pSmbResponse->pWordCount,
                        &pSmbResponse->pAndXHeader,
                        &pSmbResponse->usHeaderSize);
    }
    else
    {
        ntStatus = SrvMarshalHeaderAndX_SMB_V1(
                        pOutBuffer,
                        ulOffset,
                        ulBytesAvailable,
                        COM_QUERY_INFORMATION2,
                        &pSmbResponse->pWordCount,
                        &pSmbResponse->pAndXHeader,
                        &pSmbResponse->usHeaderSize);
    }
    BAIL_ON_NT_STATUS(ntStatus);

    pOutBuffer       += pSmbResponse->usHeaderSize;
    ulOffset         += pSmbResponse->usHeaderSize;
    ulBytesAvailable -= pSmbResponse->usHeaderSize;
    ulTotalBytesUsed += pSmbResponse->usHeaderSize;

    *pSmbResponse->pWordCount = 11;

    if (ulBytesAvailable < sizeof(QUERY_INFO2_RESPONSE_HEADER))
    {
        ntStatus = STATUS_INVALID_BUFFER_SIZE;
        BAIL_ON_NT_STATUS(ntStatus);
    }

    pResponseHeader = (PQUERY_INFO2_RESPONSE_HEADER)pOutBuffer;

    ntStatus = WireNTTimeToSMBDateTime(
                        pInfo2State->fileBasicInfo.CreationTime,
                        &pResponseHeader->creationDate,
                        &pResponseHeader->creationTime);
    BAIL_ON_NT_STATUS(ntStatus);

    ntStatus = WireNTTimeToSMBDateTime(
                        pInfo2State->fileBasicInfo.LastAccessTime,
                        &pResponseHeader->lastAccessDate,
                        &pResponseHeader->lastAccessTime);
    BAIL_ON_NT_STATUS(ntStatus);

    ntStatus = WireNTTimeToSMBDateTime(
                        pInfo2State->fileBasicInfo.LastWriteTime,
                        &pResponseHeader->lastWriteDate,
                        &pResponseHeader->lastWriteTime);
    BAIL_ON_NT_STATUS(ntStatus);

    pResponseHeader->usFileAttributes =
            SMB_FILE_ATTRIBUTES_FROM_NATIVE(pInfo2State->fileBasicInfo.FileAttributes);

    pResponseHeader->ulFileAllocationSize =
            SMB_MIN(UINT32_MAX, pInfo2State->fileStandardInfo.AllocationSize);

    pResponseHeader->ulFileDataSize =
            SMB_MIN(UINT32_MAX, pInfo2State->fileStandardInfo.EndOfFile);

    // pOutBuffer       += sizeof(QUERY_INFO2_RESPONSE_HEADER);
    // ulOffset         += sizeof(QUERY_INFO2_RESPONSE_HEADER);
    // ulBytesAvailable -= sizeof(QUERY_INFO2_RESPONSE_HEADER);
    ulTotalBytesUsed += sizeof(QUERY_INFO2_RESPONSE_HEADER);

    pSmbResponse->ulMessageSize = ulTotalBytesUsed;

cleanup:

    return ntStatus;

error:

    if (ulTotalBytesUsed)
    {
        pSmbResponse->pHeader = NULL;
        pSmbResponse->pAndXHeader = NULL;
        memset(pSmbResponse->pBuffer, 0, ulTotalBytesUsed);
    }

    pSmbResponse->ulMessageSize = 0;

    goto cleanup;
}

static
VOID
SrvFreeQueryInfo2StateHandle(
    HANDLE hInfo2State
    )
{
    SrvFreeQueryInfo2State(hInfo2State);
}

static
VOID
SrvFreeQueryInfo2State(
    PSRV_QUERY_INFO2_STATE_SMB_V1 pInfo2State
    )
{
    if (pInfo2State->pFile)
    {
        SrvFileRelease(pInfo2State->pFile);
    }

    SrvFreeMemory(pInfo2State);
}
