/* -*- mode: c; c-basic-offset: 4; indent-tabs-mode: nil; tab-width: 4 -*-
 * ex: set softtabstop=4 tabstop=8 expandtab shiftwidth=4: *
 * Editor Settings: expandtabs and use 4 spaces for indentation */

/*
 * Copyright Likewise Software
 * All rights reserved.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * for more details.  You should have received a copy of the GNU General
 * Public License along with this program.  If not, see
 * <http://www.gnu.org/licenses/>.
 *
 * LIKEWISE SOFTWARE MAKES THIS SOFTWARE AVAILABLE UNDER OTHER LICENSING
 * TERMS AS WELL.  IF YOU HAVE ENTERED INTO A SEPARATE LICENSE AGREEMENT
 * WITH LIKEWISE SOFTWARE, THEN YOU MAY ELECT TO USE THE SOFTWARE UNDER THE
 * TERMS OF THAT SOFTWARE LICENSE AGREEMENT INSTEAD OF THE TERMS OF THE GNU
 * GENERAL PUBLIC LICENSE, NOTWITHSTANDING THE ABOVE NOTICE.  IF YOU
 * HAVE QUESTIONS, OR WISH TO REQUEST A COPY OF THE ALTERNATE LICENSING
 * TERMS OFFERED BY LIKEWISE SOFTWARE, PLEASE CONTACT LIKEWISE SOFTWARE AT
 * license@likewisesoftware.com
 */

/*
 * Copyright (C) Likewise Software. All rights reserved.
 *
 * Module Name:
 *
 *        connection.c
 *
 * Abstract:
 *
 *        Likewise IO (LWIO) - SRV
 *
 *        Protocols API - SMBV2
 *
 *        Connection
 *
 * Authors: Sriram Nambakam (snambakam@likewise.com)
 *
 */

#include "includes.h"

NTSTATUS
SrvConnection2FindSession_SMB_V2(
    PSRV_EXEC_CONTEXT_SMB_V2 pSmb2Context,
    PLWIO_SRV_CONNECTION     pConnection,
    ULONG64                  ullUid,
    PLWIO_SRV_SESSION_2*     ppSession
    )
{
    NTSTATUS ntStatus = STATUS_SUCCESS;
    PLWIO_SRV_SESSION_2 pSession = NULL;
    ULONG iMsg = pSmb2Context->iMsg;
    PSRV_MESSAGE_SMB_V2 pSmbRequest = &pSmb2Context->pRequests[iMsg];
    BOOLEAN bRelated = LwIsSetFlag(pSmbRequest->pHeader->ulFlags,
                                   SMB2_FLAGS_RELATED_OPERATION);

    if (ullUid)
    {
        if (bRelated)
        {
            if (!pSmb2Context->pSession)
            {
                ntStatus = STATUS_INVALID_PARAMETER;
                BAIL_ON_NT_STATUS(ntStatus);
            }
            else
            {
                pSession = SrvSession2Acquire(pSmb2Context->pSession);
            }
        }
        else if (pSmb2Context->pSession)
        {
            if (pSmb2Context->pSession->ullUid != ullUid)
            {
                ntStatus = STATUS_INVALID_PARAMETER;
                BAIL_ON_NT_STATUS(ntStatus);
            }
            else
            {
                pSession = SrvSession2Acquire(pSmb2Context->pSession);
            }
        }
        else
        {
            ntStatus = SrvConnection2FindSession(
                            pConnection,
                            ullUid,
                            &pSession);
            BAIL_ON_NT_STATUS(ntStatus);

            pSmb2Context->pSession = SrvSession2Acquire(pSession);
        }
    }
    else if (pSmb2Context->pSession)
    {
        pSession = SrvSession2Acquire(pSmb2Context->pSession);
    }
    else
    {
        ntStatus = STATUS_NO_SUCH_LOGON_SESSION;
        BAIL_ON_NT_STATUS(ntStatus);
    }

    if (SrvSession2IsRundown(pSession))
    {
        SrvSession2Release(pSession);
        pSession = NULL;
        ntStatus = STATUS_USER_SESSION_DELETED;
        BAIL_ON_NT_STATUS(ntStatus);
    }

    *ppSession = pSession;

cleanup:

    return ntStatus;

error:

    *ppSession = NULL;

    goto cleanup;
}
