/* -*- mode: c; c-basic-offset: 4; indent-tabs-mode: nil; tab-width: 4 -*-
 * ex: set softtabstop=4 tabstop=8 expandtab shiftwidth=4: *
 * Editor Settings: expandtabs and use 4 spaces for indentation */

/*
 * Copyright Likewise Software
 * All rights reserved.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * for more details.  You should have received a copy of the GNU General
 * Public License along with this program.  If not, see
 * <http://www.gnu.org/licenses/>.
 *
 * LIKEWISE SOFTWARE MAKES THIS SOFTWARE AVAILABLE UNDER OTHER LICENSING
 * TERMS AS WELL.  IF YOU HAVE ENTERED INTO A SEPARATE LICENSE AGREEMENT
 * WITH LIKEWISE SOFTWARE, THEN YOU MAY ELECT TO USE THE SOFTWARE UNDER THE
 * TERMS OF THAT SOFTWARE LICENSE AGREEMENT INSTEAD OF THE TERMS OF THE GNU
 * GENERAL PUBLIC LICENSE, NOTWITHSTANDING THE ABOVE NOTICE.  IF YOU
 * HAVE QUESTIONS, OR WISH TO REQUEST A COPY OF THE ALTERNATE LICENSING
 * TERMS OFFERED BY LIKEWISE SOFTWARE, PLEASE CONTACT LIKEWISE SOFTWARE AT
 * license@likewisesoftware.com
 */

/*
 * Copyright (C) Likewise Software. All rights reserved.
 *
 * Module Name:
 *
 *        sharerepository.h
 *
 * Abstract:
 *
 *        Likewise IO (LWIO) - SRV
 *
 *        Share repository interface
 *
 * Authors: Sriram Nambakam (snambakam@likewise.com)
 *
 */

#ifndef __SRV_SHARE_REPOSITORY_H__
#define __SRV_SHARE_REPOSITORY_H__

#define SRV_SHARE_REPOSITORY_SPECIAL_SHARE_HOME_W {'?','h','o','m','e',0}

// Define SRV_SHARE_REPOSITORY_HANDLE_DATA in share repository db implementation
typedef struct SRV_SHARE_REPOSITORY_HANDLE_DATA *SRV_SHARE_REPOSITORY_HANDLE,
                                                **PSRV_SHARE_REPOSITORY_HANDLE;

// Share repository must NOT hold any locks when calling this callback - to make
// sure there are no deadlocks.
typedef NTSTATUS (*SRV_SHARE_REPOSITORY_RELOAD_CALLBACK)(PVOID pContext);


#define SRV_SHARE_MAX_USES_UNLIMITED    0xFFFFFFFF

typedef struct _SRV_SHARE_REPOSITORY_SHARE
{
    PWSTR pwszShareName; // Required
    PWSTR pwszPath; // Required
    PWSTR pwszComment; // Optional
    PSECURITY_DESCRIPTOR_RELATIVE pSecurityDescriptor; // Required
    ULONG SecurityDescriptorLength; // Required
    SHARE_STYPE Service; // Required
    SHARE_INFO_FLAGS Flags; // Required
    ULONG MaxUses; // Required

} SRV_SHARE_REPOSITORY_SHARE, *PSRV_SHARE_REPOSITORY_SHARE;


NTSTATUS
SrvShareRepositoryInitialize(
    OUT PSRV_SHARE_REPOSITORY_HANDLE pHandle,
    IN SRV_SHARE_REPOSITORY_RELOAD_CALLBACK ReloadCallback,
    IN PVOID pReloadCallbackContext
    );

VOID
SrvShareRepositoryCleanup(
    IN SRV_SHARE_REPOSITORY_HANDLE Handle
    );

NTSTATUS
SrvShareRepositoryAddShare(
    IN SRV_SHARE_REPOSITORY_HANDLE Handle,
    IN PSRV_SHARE_REPOSITORY_SHARE pRepositoryShare
    );

// A share has to be looked up by ShareName.  In a typical backend
// implementation, the ShareName, Path and Service fields are immutable.
NTSTATUS
SrvShareRepositoryUpdateShare(
    IN SRV_SHARE_REPOSITORY_HANDLE Handle,
    IN PSRV_SHARE_REPOSITORY_SHARE pRepositoryShare
    );

NTSTATUS
SrvShareRepositoryDeleteShare(
    IN SRV_SHARE_REPOSITORY_HANDLE Handle,
    IN PWSTR pwszShareName
    );

NTSTATUS
SrvShareRepositoryGetShareList(
    IN SRV_SHARE_REPOSITORY_HANDLE Handle,
    OUT PSRV_SHARE_REPOSITORY_SHARE** pppRepositoryShareList,
    OUT PULONG pRepositoryShareListCount
    );

VOID
SrvShareRepositoryFreeShareList(
    IN PSRV_SHARE_REPOSITORY_SHARE* ppRepositoryShareList,
    IN ULONG RepositoryShareListCount
    );


#endif /* __SRV_SHARE_REPOSITORY_H__ */
