/* Editor Settings: expandtabs and use 4 spaces for indentation
 * ex: set softtabstop=4 tabstop=8 expandtab shiftwidth=4: *
 */

/*
 * Copyright Likewise Software    2004-2008
 * All rights reserved.
 *
 * This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 2.1 of the license, or (at
 * your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser
 * General Public License for more details.  You should have received a copy
 * of the GNU Lesser General Public License along with this program.  If
 * not, see <http://www.gnu.org/licenses/>.
 *
 * LIKEWISE SOFTWARE MAKES THIS SOFTWARE AVAILABLE UNDER OTHER LICENSING
 * TERMS AS WELL.  IF YOU HAVE ENTERED INTO A SEPARATE LICENSE AGREEMENT
 * WITH LIKEWISE SOFTWARE, THEN YOU MAY ELECT TO USE THE SOFTWARE UNDER THE
 * TERMS OF THAT SOFTWARE LICENSE AGREEMENT INSTEAD OF THE TERMS OF THE GNU
 * LESSER GENERAL PUBLIC LICENSE, NOTWITHSTANDING THE ABOVE NOTICE.  IF YOU
 * HAVE QUESTIONS, OR WISH TO REQUEST A COPY OF THE ALTERNATE LICENSING
 * TERMS OFFERED BY LIKEWISE SOFTWARE, PLEASE CONTACT LIKEWISE SOFTWARE AT
 * license@likewisesoftware.com
 */

#ifndef _TESTRPC_H_
#define _TESTRPC_H_

#include <config.h>

#include <lw/ntstatus.h>
#include <lw/winerror.h>
#include <wc16str.h>
#include <wc16printf.h>
#include "params.h"


struct _TEST;
struct _CREDENTIALS;
struct _RPCSERVER;

typedef DWORD (*test_fn)(
    struct _TEST *pTest,
    struct _RPCSERVER* Server,
    struct _CREDENTIALS *pCreds,
    struct _PARAMETER *pOptions,
    DWORD dwOptcount
    );


typedef struct _TEST
{
    PCSTR         pszName;
    test_fn       Function;

    struct _TEST *pNext;
} TEST, *PTEST;


typedef enum _CREDENTIALS_TYPE
{
    CREDENTIALS_TYPE_UNKNOWN = 0,
    CREDENTIALS_TYPE_KRB5,
    CREDENTIALS_TYPE_NTLM
}
CREDENTIALS_TYPE, *PCREDENTIALS_TYPE;

typedef struct _CREDENTIALS
{
    CREDENTIALS_TYPE Type;

    struct _KRB5
    {
        PWSTR   pwszPrincipal;
        PWSTR   pwszCredsCache;
    } Krb5;

    struct _NTLM
    {
        PWSTR   pwszDomain;
        PWSTR   pwszWorkstation;
        PWSTR   pwszUsername;
        PWSTR   pwszPassword;
    } Ntlm;
} CREDENTIALS, *PCREDENTIALS;


VOID
AddTest(
    PTEST  pFt,
    PCSTR  pszName,
    test_fn function
    );


VOID SetupNetApiTests(PTEST t);
VOID SetupSamrTests(PTEST t);
VOID SetupLsaTests(PTEST t);
VOID SetupNetlogonTests(PTEST t);
VOID SetupDsrTests(PTEST t);
VOID SetupWkssvcTests(PTEST t);


extern BOOLEAN verboseMode;
extern BOOLEAN helpMode;


#if defined(BAIL_ON_WIN_ERROR)
#undef BAIL_ON_WIN_ERROR
#endif

#define BAIL_ON_WIN_ERROR(err) \
    if ((err) != ERROR_SUCCESS) \
    { \
        printf("Error: %s (%u=0x%08x) in %s:%u\n", \
               LwWin32ErrorToName((err)), (err), (err), \
               __FILE__, __LINE__); \
        goto error; \
    }

#if defined(BAIL_ON_NT_STATUS)
#undef BAIL_ON_NT_STATUS
#endif

#define BAIL_ON_NT_STATUS(status) \
    if ((status) != STATUS_SUCCESS) \
    { \
        printf("Error: %s (0x%08x) in %s:%u\n", \
               LwNtStatusToName((status)), (status), \
               __FILE__, __LINE__); \
        goto error; \
    }


#define TESTINFO(test, description) \
    do { \
        if (verboseMode) \
        { \
            printf("#\n# Test: %s\n#\n\nParameters:\n", (test)->pszName); \
        } \
        else \
        { \
            printf("%s: ", (test)->pszName); \
        } \
        \
        if (helpMode) \
        { \
            printf("%s\nDefault Parameters:\n", description); \
        } \
    } while (0)


#define DUMP_WSTR(pfx, v)                           \
    if (verboseMode) {                              \
        w16printfw(L"%hhs%hhs = \"%ws\"\n", pfx, #v, (v)); \
    }

#define DUMP_INT(pfx, v)                                    \
    if (verboseMode) {                                     \
        printf("%s%s = %d (0x%08x)\n", pfx, #v, (v), (v));  \
    }

#define DUMP_UINT(pfx, v)                                   \
    if (verboseMode) {                                     \
        printf("%s%s = %u (0x%08x)\n", pfx, #v, (v), (v));  \
    }

#define DUMP_UNICODE_STRING(pfx, v)                     \
    if (verboseMode) {                                 \
        wchar16_t *str = GetFromUnicodeString((v));     \
        w16printfw(L"%hhs%hhs = \"%ws\"\n", pfx, #v, str);     \
        SAFE_FREE(str);                                 \
    }

#define DUMP_UNICODE_STRING_EX(pfx, v)                  \
    if (verboseMode) {                                 \
        wchar16_t *str = GetFromUnicodeStringEx((v));   \
        w16printfw(L"%hhs%hhs = \"%ws\"\n", pfx, #v, str);     \
        SAFE_FREE(str);                                 \
    }


#define INPUT_ARG_WSTR(v)            DUMP_WSTR("> ", (v));
#define INPUT_ARG_INT(v)             DUMP_INT("> ", (v));
#define INPUT_ARG_UINT(v)            DUMP_UINT("> ", (v));
#define INPUT_ARG_UNICODE_STRING(v)  DUMP_UNICODE_STRING("> ", (v));

#define OUTPUT_ARG_WSTR(v)           DUMP_WSTR("< ", (v));
#define OUTPUT_ARG_INT(v)            DUMP_INT("< ", (v));
#define OUTPUT_ARG_UINT(v)           DUMP_UINT("< ", (v));
#define OUTPUT_ARG_UNICODE_STRING(v) DUMP_UNICODE_STRING("< ", (v));

#define CALL_NT_API(status, nt_call) \
    do { \
        status = nt_call; \
        if (status != STATUS_SUCCESS) \
        { \
            printf("%s = %s (0x%08x)\n", \
                   #nt_call, \
                   NtStatusToName((status)), (status)); \
        } \
    } while (0)


#define CALL_WIN_API(status, win_call) \
    do { \
        status = win_call; \
        if (status != ERROR_SUCCESS) \
        { \
            printf("%s = %s (0x%08x)\n", \
                   #win_call, \
                   Win32ErrorToName((status)), (status)); \
        } \
    } while (0)


#define DISPLAY_ERROR(msg) \
    do { \
        printf("%s:%d ", __FILE__, __LINE__); \
        printf msg; \
    } while (0)


#define DISPLAY_NTSTATUS(status) \
    do { \
        if (!verboseMode) break; \
        printf("%s:%d ", __FILE__, __LINE__); \
        printf("NTSTATUS code = %s (0x%08x)\n", \
               LwNtStatusToName(status), \
               (status)); \
    } while (0)


#define DISPLAY_NTSTATUS_EXPECTED(status, expected) \
    do { \
        if (!verboseMode || \
            (status) == (expected)) break; \
        printf("%s:%d ", __FILE__, __LINE__); \
        printf("NTSTATUS code = %s (0x%08x), ", \
               LwNtStatusToName(status), \
               (status)); \
        printf("expected = %s (0x%08x)\n", \
               LwNtStatusToName(expected), \
               (expected)); \
    } while (0)


#define DISPLAY_COMMENT(msg) \
    do { \
        if (!verboseMode) break; \
        printf("%s:%d ", __FILE__, __LINE__); \
        printf msg; \
    } while (0)


#define ASSERT_TEST(expr) \
    if (!(expr)) \
    { \
        DISPLAY_ERROR(("assert failed: %s\n", #expr)); \
        bRet = FALSE; \
    }


#define TEST_PASSED(msg, status) \
    do { \
        printf msg; \
        printf(" %s", (status) ? "PASSED\n" : "FAILED\n"); \
    } while (0);


#define ASSERT_TEST_MSG(expr, msg)                                 \
    if (!(expr))                                                   \
    {                                                              \
        DISPLAY_ERROR(("assert failed: %s ", #expr));              \
        printf msg;                                                \
        bRet = FALSE;                                              \
    }

#define ASSERT_DWORD_EQUAL(val1, val2)                             \
    if ((val1) != (val2))                                          \
    {                                                              \
        DISPLAY_ERROR(("assert failed: %s != %s\n",                \
                       #val1, #val2));                             \
        bRet = FALSE;                                              \
    }


#define ASSERT_WC16STRING_EQUAL(str1_ptr, str2_ptr)                \
    if (((str1_ptr) == NULL && (str2_ptr) != NULL) ||              \
        ((str1_ptr) != NULL && (str2_ptr) == NULL) ||              \
        (((str1_ptr) != (str2_ptr)) &&                             \
         !LwRtlWC16StringIsEqual((str1_ptr), (str2_ptr), FALSE)))  \
    {                                                              \
        DISPLAY_ERROR(("assert failed: %s != %s\n",                \
                       #str1_ptr, #str2_ptr));                     \
        bRet = FALSE;                                              \
    }


#define ASSERT_UNICODE_STRING_EQUAL(str1_ptr, str2_ptr)            \
    if (((str1_ptr) != (str2_ptr)) &&                              \
        !LwRtlUnicodeStringIsEqual((str1_ptr), (str2_ptr),         \
                                   FALSE))                         \
    {                                                              \
        DISPLAY_ERROR(("assert failed: %s != %s\n",                \
                       #str1_ptr, #str2_ptr));                     \
        bRet = FALSE;                                              \
    }


#define ASSERT_UNICODE_STRING_VALID_MSG(name_ptr, msg)             \
    do {                                                           \
        DWORD _iWC = 0;                                            \
                                                                   \
        if (!((name_ptr)->Length == (name_ptr)->MaximumLength) &&  \
            !((name_ptr)->Length + sizeof((name_ptr)->Buffer[0])   \
              == (name_ptr)->MaximumLength))                       \
        {                                                          \
            DISPLAY_ERROR(("assert failed: "                       \
                           "invalid unicode name len/size\n"));    \
            printf msg;                                            \
            bRet = FALSE;                                          \
        }                                                          \
                                                                   \
        if ((name_ptr)->MaximumLength == 0 &&                      \
            (name_ptr)->Buffer != NULL)                            \
        {                                                          \
            DISPLAY_ERROR(("assert failed: "                       \
                           "invalid unicode name "                 \
                           "size == 0 while buffer != NULL\n"));   \
            printf msg;                                            \
            bRet = FALSE;                                          \
        }                                                          \
                                                                   \
        for (_iWC = 0;                                             \
             _iWC < ((name_ptr)->Length                            \
                     / sizeof((name_ptr)->Buffer[0]));             \
             _iWC++)                                               \
        {                                                          \
            if ((name_ptr)->Buffer[_iWC] == 0)                     \
            {                                                      \
                DISPLAY_ERROR(("assert failed: "                   \
                               "invalid unicode name "             \
                               "NULL-termination "                 \
                               "before i < size\n"));              \
                printf msg;                                        \
                bRet = FALSE;                                      \
            }                                                      \
        }                                                          \
                                                                   \
    } while (0)


#define ASSERT_UNICODE_STRING_VALID(name_ptr)                      \
    do {                                                           \
        DWORD _iWC = 0;                                            \
                                                                   \
        if (!((name_ptr)->Length == (name_ptr)->MaximumLength) &&  \
            !((name_ptr)->Length + sizeof((name_ptr)->Buffer[0])   \
              == (name_ptr)->MaximumLength))                       \
        {                                                          \
            DISPLAY_ERROR(("assert failed: "                       \
                           "invalid unicode name len/size\n"));    \
            bRet = FALSE;                                          \
        }                                                          \
                                                                   \
        if ((name_ptr)->MaximumLength == 0 &&                      \
            (name_ptr)->Buffer != NULL)                            \
        {                                                          \
            DISPLAY_ERROR(("assert failed: "                       \
                           "invalid unicode name "                 \
                           "size == 0 while buffer != NULL\n"));   \
            bRet = FALSE;                                          \
        }                                                          \
                                                                   \
        for (_iWC = 0;                                             \
             _iWC < ((name_ptr)->Length                            \
                     / sizeof((name_ptr)->Buffer[0]));             \
             _iWC++)                                               \
        {                                                          \
            if ((name_ptr)->Buffer[_iWC] == 0)                     \
            {                                                      \
                DISPLAY_ERROR(("assert failed: "                   \
                               "invalid unicode name "             \
                               "NULL-termination "                 \
                               "before i < size\n"));              \
                bRet = FALSE;                                      \
            }                                                      \
        }                                                          \
                                                                   \
    } while (0)


#define ASSERT_SID_EQUAL(sid1_ptr, sid2_ptr)                       \
    do {                                                           \
        if ((sid1_ptr) == (sid2_ptr))                              \
        {                                                          \
            break;                                                 \
        }                                                          \
                                                                   \
        if ((sid1_ptr) == NULL && (sid2_ptr) != NULL)              \
        {                                                          \
            DISPLAY_ERROR(("invalid SID: %s == NULL\n",            \
                           #sid1_ptr));                            \
            bRet = FALSE;                                          \
            break;                                                 \
        }                                                          \
                                                                   \
        if ((sid1_ptr) != NULL && (sid2_ptr) == NULL)              \
        {                                                          \
            DISPLAY_ERROR(("invalid SID: %s == NULL\n",            \
                           #sid2_ptr));                            \
            bRet = FALSE;                                          \
            break;                                                 \
        }                                                          \
                                                                   \
        if (!RtlEqualSid((sid1_ptr), (sid2_ptr)))                  \
        {                                                          \
            DISPLAY_ERROR(("assert failed: %s == %s\n",            \
                           #sid1_ptr, #sid2_ptr));                 \
            bRet = FALSE;                                          \
        }                                                          \
    } while (0)


#define ASSERT_SID_EQUAL_MSG(sid1_ptr, sid2_ptr, msg)              \
    do {                                                           \
        if ((sid1_ptr) == (sid2_ptr))                              \
        {                                                          \
            break;                                                 \
        }                                                          \
                                                                   \
        if ((sid1_ptr) == NULL && (sid2_ptr) != NULL)              \
        {                                                          \
            DISPLAY_ERROR(("assert failed: %s == NULL\n",          \
                           #sid1_ptr));                            \
            printf msg;                                            \
            bRet = FALSE;                                          \
            break;                                                 \
        }                                                          \
                                                                   \
        if ((sid1_ptr) != NULL && (sid2_ptr) == NULL)              \
        {                                                          \
            DISPLAY_ERROR(("assert failed: %s == NULL\n",          \
                           #sid2_ptr));                            \
            printf msg;                                            \
            bRet = FALSE;                                          \
            break;                                                 \
        }                                                          \
                                                                   \
        if (!RtlEqualSid((sid1_ptr), (sid2_ptr)))                  \
        {                                                          \
            DISPLAY_ERROR(("assert failed: %s != %s\n",            \
                           #sid1_ptr, #sid2_ptr));                 \
            printf msg;                                            \
            bRet = FALSE;                                          \
        }                                                          \
    } while (0)


#endif /* _TESTRPC_H_ */


/*
local variables:
mode: c
c-basic-offset: 4
indent-tabs-mode: nil
tab-width: 4
end:
*/
