/*
 * Copyright (c) Likewise Software.  All rights Reserved.
 *
 * This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 2.1 of the license, or (at
 * your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser
 * General Public License for more details.  You should have received a copy
 * of the GNU Lesser General Public License along with this program.  If
 * not, see <http://www.gnu.org/licenses/>.
 *
 * LIKEWISE SOFTWARE MAKES THIS SOFTWARE AVAILABLE UNDER OTHER LICENSING
 * TERMS AS WELL.  IF YOU HAVE ENTERED INTO A SEPARATE LICENSE AGREEMENT
 * WITH LIKEWISE SOFTWARE, THEN YOU MAY ELECT TO USE THE SOFTWARE UNDER THE
 * TERMS OF THAT SOFTWARE LICENSE AGREEMENT INSTEAD OF THE TERMS OF THE GNU
 * LESSER GENERAL PUBLIC LICENSE, NOTWITHSTANDING THE ABOVE NOTICE.  IF YOU
 * HAVE QUESTIONS, OR WISH TO REQUEST A COPY OF THE ALTERNATE LICENSING
 * TERMS OFFERED BY LIKEWISE SOFTWARE, PLEASE CONTACT LIKEWISE SOFTWARE AT
 * license@likewise.com
 */

/*
 * Module Name:
 *
 *        error-table.h
 *
 *   NT status and Win32 error code mapping table
 *
 *   Some of these definitions were written from scratch.
 *
 *   NT status codes are unsigned 32-bit hex values
 *
 *   Win32 error codes are unsigned 32-bit decimal values
 *
 */

/* Equivalence table*/
#ifdef __ERROR_XMACRO__
#define S STATUS_CODE

S ( LW_STATUS_SUCCESS,                ERROR_SUCCESS,            ERRSUCCESS, 0           , 0      , "Success" )
S ( LW_STATUS_MORE_ENTRIES,           ERROR_MORE_DATA,          ERRSRV,     ERRerror    , EAGAIN , "More data available" )
S ( LW_STATUS_NOT_IMPLEMENTED,        ERROR_INVALID_FUNCTION,   ERRDOS,     ERRbadfunc  , ENOTSUP, "Not implemented." )
S ( LW_STATUS_INVALID_INFO_CLASS,     -1,                       ERRDOS,     ERRbadpipe  , -1     , "" )
S ( LW_STATUS_INVALID_PARAMETER,      ERROR_INVALID_PARAMETER,  ERRSRV,     ERRerror    , EINVAL , "Invalid parameter passed" )
S ( LW_STATUS_NO_MEMORY,              ERROR_OUTOFMEMORY,        ERRSRV,     ERRerror    , ENOMEM , "Out of memory" )
S ( LW_STATUS_INVALID_SID,            ERROR_INVALID_SID,        ERRSRV,     ERRerror    , EINVAL , "" )
S ( LW_STATUS_INSUFFICIENT_RESOURCES, -1,                       ERRSRV,     ERRerror    , ENOSPC , "" )
S ( LW_STATUS_UNHANDLED_EXCEPTION,    -1,                       ERRSRV,     ERRerror    , -1     , "" )
S ( LW_STATUS_ACCESS_DENIED,          ERROR_ACCESS_DENIED,      ERRDOS,     ERRbadaccess, EACCES , "Access is denied" )
S ( LW_STATUS_NONE_MAPPED,            ERROR_NONE_MAPPED,        ERRSRV,     ERRerror    , -1     , "Specified account has no mapping to a valid SID" )
S ( LW_STATUS_CONNECTION_REFUSED,     ERROR_CONNECTION_REFUSED, ERRSRV,     ERRerror    , 111    , "" )
S ( LW_STATUS_SPECIAL_ACCOUNT,        ERROR_SPECIAL_ACCOUNT,    ERRSRV,     ERRerror    , -1     , "Cannot perform this operation on built-in accounts" )

S ( LW_STATUS_WAIT_1, ERROR_MR_MID_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_WAIT_2, ERROR_MR_MID_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_WAIT_3, ERROR_MR_MID_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_WAIT_63, ERROR_MR_MID_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_ABANDONED_WAIT_0, ERROR_MR_MID_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_ABANDONED_WAIT_63, ERROR_MR_MID_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_USER_APC, ERROR_MR_MID_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_KERNEL_APC, ERROR_MR_MID_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_ALERTED, ERROR_MR_MID_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_TIMEOUT, ERROR_MR_MID_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_PENDING, ERROR_IO_PENDING, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_REPARSE, ERROR_MR_MID_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_NOT_ALL_ASSIGNED, ERROR_NOT_ALL_ASSIGNED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_SOME_NOT_MAPPED, ERROR_SOME_NOT_MAPPED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_OPLOCK_BREAK_IN_PROGRESS, ERROR_OPLOCK_BREAK_IN_PROGRESS, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_VOLUME_MOUNTED, ERROR_MR_MID_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_RXACT_COMMITTED, ERROR_MR_MID_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_NOTIFY_CLEANUP, ERROR_MR_MID_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_NOTIFY_ENUM_DIR, ERROR_NOTIFY_ENUM_DIR, ERRDOS, ERROR_NOTIFY_ENUM_DIR, -1, "" )
S ( LW_STATUS_NO_QUOTAS_FOR_ACCOUNT, ERROR_NO_QUOTAS_FOR_ACCOUNT, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_PRIMARY_TRANSPORT_CONNECT_FAILED, ERROR_MR_MID_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_PAGE_FAULT_TRANSITION, ERROR_MR_MID_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_PAGE_FAULT_DEMAND_ZERO, ERROR_MR_MID_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_PAGE_FAULT_COPY_ON_WRITE, ERROR_MR_MID_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_PAGE_FAULT_GUARD_PAGE, ERROR_MR_MID_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_PAGE_FAULT_PAGING_FILE, ERROR_MR_MID_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_CACHE_PAGE_LOCKED, ERROR_MR_MID_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_CRASH_DUMP, ERROR_MR_MID_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_BUFFER_ALL_ZEROS, ERROR_MR_MID_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_REPARSE_OBJECT, ERROR_MR_MID_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_RESOURCE_REQUIREMENTS_CHANGED, ERROR_MR_MID_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_TRANSLATION_COMPLETE, ERROR_MR_MID_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_DS_MEMBERSHIP_EVALUATED_LOCALLY, ERROR_DS_MEMBERSHIP_EVALUATED_LOCALLY, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_NOTHING_TO_TERMINATE, ERROR_MR_MID_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_PROCESS_NOT_IN_JOB, ERROR_MR_MID_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_PROCESS_IN_JOB, ERROR_MR_MID_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_VOLSNAP_HIBERNATE_READY, ERROR_MR_MID_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_FSFILTER_OP_COMPLETED_SUCCESSFULLY, ERROR_MR_MID_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_INTERRUPT_VECTOR_ALREADY_CONNECTED, ERROR_MR_MID_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_INTERRUPT_STILL_CONNECTED, ERROR_MR_MID_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_PROCESS_CLONED, ERROR_MR_MID_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_FILE_LOCKED_WITH_ONLY_READERS, ERROR_MR_MID_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_FILE_LOCKED_WITH_WRITERS, ERROR_MR_MID_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_RESOURCEMANAGER_READ_ONLY, ERROR_MR_MID_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_WAIT_FOR_OPLOCK, ERROR_WAIT_FOR_OPLOCK, ERRSRV, ERRerror, -1, "" )
S ( LW_DBG_EXCEPTION_HANDLED, ERROR_MR_MID_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_DBG_CONTINUE, ERROR_MR_MID_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_OBJECT_NAME_EXISTS, ERROR_MR_MID_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_THREAD_WAS_SUSPENDED, ERROR_MR_MID_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_WORKING_SET_LIMIT_RANGE, ERROR_INVALID_PARAMETER, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_IMAGE_NOT_AT_BASE, ERROR_MR_MID_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_RXACT_STATE_CREATED, ERROR_MR_MID_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_SEGMENT_NOTIFICATION, ERROR_MR_MID_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_LOCAL_USER_SESSION_KEY, ERROR_LOCAL_USER_SESSION_KEY, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_BAD_CURRENT_DIRECTORY, ERROR_MR_MID_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_SERIAL_MORE_WRITES, ERROR_MORE_WRITES, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_REGISTRY_RECOVERED, ERROR_REGISTRY_RECOVERED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_FT_READ_RECOVERY_FROM_BACKUP, ERROR_MR_MID_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_FT_WRITE_RECOVERY, ERROR_MR_MID_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_SERIAL_COUNTER_TIMEOUT, ERROR_COUNTER_TIMEOUT, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_NULL_LM_PASSWORD, ERROR_NULL_LM_PASSWORD, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_IMAGE_MACHINE_TYPE_MISMATCH, ERROR_MR_MID_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_RECEIVE_PARTIAL, ERROR_MR_MID_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_RECEIVE_EXPEDITED, ERROR_MR_MID_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_RECEIVE_PARTIAL_EXPEDITED, ERROR_MR_MID_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_EVENT_DONE, ERROR_MR_MID_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_EVENT_PENDING, ERROR_MR_MID_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_CHECKING_FILE_SYSTEM, ERROR_MR_MID_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_FATAL_APP_EXIT, ERROR_MR_MID_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_PREDEFINED_HANDLE, ERROR_MR_MID_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_WAS_UNLOCKED, ERROR_MR_MID_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_SERVICE_NOTIFICATION, ERROR_MR_MID_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_WAS_LOCKED, ERROR_MR_MID_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_LOG_HARD_ERROR, ERROR_MR_MID_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_ALREADY_WIN32, ERROR_MR_MID_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_WX86_UNSIMULATE, ERROR_MR_MID_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_WX86_CONTINUE, ERROR_MR_MID_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_WX86_SINGLE_STEP, ERROR_MR_MID_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_WX86_BREAKPOINT, ERROR_MR_MID_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_WX86_EXCEPTION_CONTINUE, ERROR_MR_MID_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_WX86_EXCEPTION_LASTCHANCE, ERROR_MR_MID_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_WX86_EXCEPTION_CHAIN, ERROR_MR_MID_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_IMAGE_MACHINE_TYPE_MISMATCH_EXE, ERROR_MR_MID_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_NO_YIELD_PERFORMED, ERROR_MR_MID_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_TIMER_RESUME_IGNORED, ERROR_MR_MID_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_ARBITRATION_UNHANDLED, ERROR_MR_MID_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_CARDBUS_NOT_SUPPORTED, ERROR_MR_MID_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_WX86_CREATEWX86TIB, ERROR_MR_MID_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MP_PROCESSOR_MISMATCH, ERROR_MR_MID_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_HIBERNATED, ERROR_MR_MID_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_RESUME_HIBERNATION, ERROR_MR_MID_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_FIRMWARE_UPDATED, ERROR_MR_MID_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_DRIVERS_LEAKING_LOCKED_PAGES, ERROR_MR_MID_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_RETRIEVED, ERROR_MR_MID_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_SYSTEM_POWERSTATE_TRANSITION, ERROR_MR_MID_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_ALPC_CHECK_COMPLETION_LIST, ERROR_MR_MID_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_SYSTEM_POWERSTATE_COMPLEX_TRANSITION, ERROR_MR_MID_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_ACCESS_AUDIT_BY_POLICY, ERROR_MR_MID_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_ABANDON_HIBERFILE, ERROR_MR_MID_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_BIZRULES_NOT_ENABLED, ERROR_MR_MID_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_WAKE_SYSTEM, ERROR_MR_MID_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_DS_SHUTTING_DOWN, ERROR_DS_SHUTTING_DOWN, ERRSRV, ERRerror, -1, "" )
S ( LW_DBG_REPLY_LATER, ERROR_MR_MID_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_DBG_UNABLE_TO_PROVIDE_HANDLE, ERROR_MR_MID_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_DBG_TERMINATE_THREAD, ERROR_MR_MID_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_DBG_TERMINATE_PROCESS, ERROR_MR_MID_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_DBG_CONTROL_C, ERROR_MR_MID_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_DBG_PRINTEXCEPTION_C, ERROR_MR_MID_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_DBG_RIPEXCEPTION, ERROR_MR_MID_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_DBG_CONTROL_BREAK, ERROR_MR_MID_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_DBG_COMMAND_EXCEPTION, ERROR_MR_MID_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_RPC_NT_UUID_LOCAL_ONLY, RPC_S_UUID_LOCAL_ONLY, ERRSRV, ERRerror, -1, "" )
S ( LW_RPC_NT_SEND_INCOMPLETE, RPC_S_SEND_INCOMPLETE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_GUARD_PAGE_VIOLATION, -2147483647, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_DATATYPE_MISALIGNMENT, ERROR_NOACCESS, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_BREAKPOINT, -2147483645, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_SINGLE_STEP, -2147483644, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_BUFFER_OVERFLOW, ERROR_MORE_DATA, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_NO_MORE_FILES, ERROR_NO_MORE_FILES, ERRDOS, ERRnofiles, -1, "" )
S ( LW_STATUS_WAKE_SYSTEM_DEBUGGER, ERROR_MR_MID_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_HANDLES_CLOSED, ERROR_MR_MID_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_NO_INHERITANCE, ERROR_NO_INHERITANCE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_GUID_SUBSTITUTION_MADE, ERROR_MR_MID_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_PARTIAL_COPY, ERROR_PARTIAL_COPY, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_DEVICE_PAPER_EMPTY, ERROR_OUT_OF_PAPER, ERRHRD, ERRnopaper, -1, "" )
S ( LW_STATUS_DEVICE_POWERED_OFF, ERROR_NOT_READY, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_DEVICE_OFF_LINE, ERROR_NOT_READY, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_DEVICE_BUSY, ERROR_BUSY, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_NO_MORE_EAS, ERROR_NO_MORE_ITEMS, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_INVALID_EA_NAME, ERROR_INVALID_EA_NAME, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_EA_LIST_INCONSISTENT, ERROR_EA_LIST_INCONSISTENT, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_INVALID_EA_FLAG, ERROR_EA_LIST_INCONSISTENT, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_VERIFY_REQUIRED, ERROR_MEDIA_CHANGED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_EXTRANEOUS_INFORMATION, ERROR_MR_MID_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_RXACT_COMMIT_NECESSARY, ERROR_MR_MID_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_NO_MORE_ENTRIES, ERROR_NO_MORE_ITEMS, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_FILEMARK_DETECTED, ERROR_FILEMARK_DETECTED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MEDIA_CHANGED, ERROR_MEDIA_CHANGED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_BUS_RESET, ERROR_BUS_RESET, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_END_OF_MEDIA, ERROR_END_OF_MEDIA, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_BEGINNING_OF_MEDIA, ERROR_BEGINNING_OF_MEDIA, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MEDIA_CHECK, ERROR_MR_MID_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_SETMARK_DETECTED, ERROR_SETMARK_DETECTED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_NO_DATA_DETECTED, ERROR_NO_DATA_DETECTED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_REDIRECTOR_HAS_OPEN_HANDLES, ERROR_MR_MID_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_SERVER_HAS_OPEN_HANDLES, ERROR_MR_MID_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_ALREADY_DISCONNECTED, NERR_ActiveConns, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_LONGJUMP, ERROR_MR_MID_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_CLEANER_CARTRIDGE_INSTALLED, ERROR_CLEANER_CARTRIDGE_INSTALLED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_PLUGPLAY_QUERY_VETOED, ERROR_MR_MID_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_UNWIND_CONSOLIDATE, ERROR_MR_MID_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_REGISTRY_HIVE_RECOVERED, ERROR_MR_MID_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_DLL_MIGHT_BE_INSECURE, ERROR_MR_MID_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_DLL_MIGHT_BE_INCOMPATIBLE, ERROR_MR_MID_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_STOPPED_ON_SYMLINK, ERROR_MR_MID_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_DEVICE_REQUIRES_CLEANING, ERROR_DEVICE_REQUIRES_CLEANING, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_DEVICE_DOOR_OPEN, ERROR_DEVICE_DOOR_OPEN, ERRSRV, ERRerror, -1, "" )
S ( LW_DBG_EXCEPTION_NOT_HANDLED, ERROR_MR_MID_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_UNSUCCESSFUL, ERROR_GEN_FAILURE, ERRDOS, ERRgeneral, -1, "" )
S ( LW_STATUS_INFO_LENGTH_MISMATCH, ERROR_BAD_LENGTH, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_ACCESS_VIOLATION, ERROR_NOACCESS, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_IN_PAGE_ERROR, ERROR_SWAPERROR, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_PAGEFILE_QUOTA, ERROR_PAGEFILE_QUOTA, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_INVALID_HANDLE, ERROR_INVALID_HANDLE, ERRDOS, ERRbadfid, EBADF, "" )
S ( LW_STATUS_BAD_INITIAL_STACK, ERROR_STACK_OVERFLOW, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_BAD_INITIAL_PC, ERROR_BAD_EXE_FORMAT, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_INVALID_CID, ERROR_INVALID_PARAMETER, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_TIMER_NOT_CANCELED, ERROR_MR_MID_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_NO_SUCH_FILE, ERROR_FILE_NOT_FOUND, ERRDOS, ERRbadfile, ENOENT, "File not found." )
S ( LW_STATUS_NO_SUCH_DEVICE, ERROR_FILE_NOT_FOUND, ERRDOS, ERRbadfile, ENOENT, "Device not found." )
S ( LW_STATUS_INVALID_DEVICE_REQUEST, ERROR_INVALID_FUNCTION, ERRDOS, ERRbadfunc, -1, "Invalid device request." )
S ( LW_STATUS_END_OF_FILE, ERROR_HANDLE_EOF, ERRDOS, ERReof, -1, "" )
S ( LW_STATUS_WRONG_VOLUME, ERROR_WRONG_DISK, ERRHRD, ERRwrongdisk, -1, "" )
S ( LW_STATUS_NO_MEDIA_IN_DEVICE, ERROR_NOT_READY, ERRHRD, ERRnotready, -1, "" )
S ( LW_STATUS_UNRECOGNIZED_MEDIA, ERROR_UNRECOGNIZED_MEDIA, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_NONEXISTENT_SECTOR, ERROR_SECTOR_NOT_FOUND, ERRHRD, ERRbadsector, EIO, "" )
S ( LW_STATUS_MORE_PROCESSING_REQUIRED, ERROR_MORE_DATA, ERRDOS, ERRmoredata, -1, "" )
S ( LW_STATUS_CONFLICTING_ADDRESSES, ERROR_INVALID_ADDRESS, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_NOT_MAPPED_VIEW, ERROR_INVALID_ADDRESS, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_UNABLE_TO_FREE_VM, ERROR_INVALID_PARAMETER, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_UNABLE_TO_DELETE_SECTION, ERROR_INVALID_PARAMETER, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_INVALID_SYSTEM_SERVICE, ERROR_INVALID_FUNCTION, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_ILLEGAL_INSTRUCTION, -1073741795, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_INVALID_LOCK_SEQUENCE, ERROR_ACCESS_DENIED, ERRDOS, ERRnoaccess, EPERM, "" )
S ( LW_STATUS_INVALID_VIEW_SIZE, ERROR_ACCESS_DENIED, ERRDOS, ERRnoaccess, EPERM, "" )
S ( LW_STATUS_INVALID_FILE_FOR_SECTION, ERROR_BAD_EXE_FORMAT, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_ALREADY_COMMITTED, ERROR_ACCESS_DENIED, ERRDOS, ERRnoaccess, EPERM, "" )
S ( LW_STATUS_BUFFER_TOO_SMALL, ERROR_INSUFFICIENT_BUFFER, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_OBJECT_TYPE_MISMATCH, ERROR_INVALID_HANDLE, ERRDOS, ERRbadfid, EBADF, "" )
S ( LW_STATUS_NONCONTINUABLE_EXCEPTION, -1073741787, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_INVALID_DISPOSITION, -1073741786, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_UNWIND, ERROR_MR_MID_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_BAD_STACK, ERROR_MR_MID_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_INVALID_UNWIND_TARGET, ERROR_MR_MID_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_NOT_LOCKED, ERROR_NOT_LOCKED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_PARITY_ERROR, -1073741781, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_UNABLE_TO_DECOMMIT_VM, ERROR_INVALID_ADDRESS, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_NOT_COMMITTED, ERROR_INVALID_ADDRESS, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_INVALID_PORT_ATTRIBUTES, ERROR_MR_MID_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_PORT_MESSAGE_TOO_LONG, ERROR_MR_MID_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_INVALID_PARAMETER_MIX, ERROR_INVALID_PARAMETER, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_INVALID_QUOTA_LOWER, ERROR_MR_MID_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_DISK_CORRUPT_ERROR, ERROR_DISK_CORRUPT, ERRHRD, ERRbadmedia, EIO, "" )
S ( LW_STATUS_OBJECT_NAME_INVALID, ERROR_INVALID_NAME, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_OBJECT_NAME_NOT_FOUND, ERROR_FILE_NOT_FOUND, ERRDOS, ERRbadfile, ENOENT, "Object not found." )
S ( LW_STATUS_OBJECT_NAME_COLLISION, ERROR_ALREADY_EXISTS, ERRDOS, ERRfilexists, EEXIST, "" )
S ( LW_STATUS_PORT_DISCONNECTED, ERROR_INVALID_HANDLE, ERRDOS, ERRbadfid, EBADF, "" )
S ( LW_STATUS_DEVICE_ALREADY_ATTACHED, ERROR_MR_MID_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_OBJECT_PATH_INVALID, ERROR_BAD_PATHNAME, ERRDOS, ERRbadpath, ENOENT, "" )
S ( LW_STATUS_OBJECT_PATH_NOT_FOUND, ERROR_PATH_NOT_FOUND, ERRDOS, ERRbadpath, ENOENT, "" )
S ( LW_STATUS_OBJECT_PATH_SYNTAX_BAD, ERROR_BAD_PATHNAME, ERRDOS, ERRbadpath, ENOENT, "" )
S ( LW_STATUS_DATA_OVERRUN, ERROR_IO_DEVICE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_DATA_LATE_ERROR, ERROR_IO_DEVICE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_DATA_ERROR, ERROR_CRC, ERRHRD, ERRdata, EIO, "" )
S ( LW_STATUS_CRC_ERROR, ERROR_CRC, ERRHRD, ERRdata, EIO, "" )
S ( LW_STATUS_SECTION_TOO_BIG, ERROR_NOT_ENOUGH_MEMORY, ERRDOS, ERRnomem, ENOMEM, "" )
S ( LW_STATUS_PORT_CONNECTION_REFUSED, ERROR_ACCESS_DENIED, ERRDOS, ERRnoaccess, EPERM, "" )
S ( LW_STATUS_INVALID_PORT_HANDLE, ERROR_INVALID_HANDLE, ERRDOS, ERRbadfid, EBADF, "" )
S ( LW_STATUS_SHARING_VIOLATION, ERROR_SHARING_VIOLATION, ERRDOS, ERRbadshare, ETXTBSY, "" )
S ( LW_STATUS_QUOTA_EXCEEDED, ERROR_NOT_ENOUGH_QUOTA, ERRSRV, ERRerror, -1, "Not enough quota is available" )
S ( LW_STATUS_INVALID_PAGE_PROTECTION, ERROR_INVALID_PARAMETER, ERRDOS, ERRinvalidparam, -1, "" )
S ( LW_STATUS_MUTANT_NOT_OWNED, ERROR_NOT_OWNER, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_SEMAPHORE_LIMIT_EXCEEDED, ERROR_TOO_MANY_POSTS, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_PORT_ALREADY_SET, ERROR_INVALID_PARAMETER, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_SECTION_NOT_IMAGE, ERROR_INVALID_PARAMETER, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_SUSPEND_COUNT_EXCEEDED, ERROR_SIGNAL_REFUSED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_THREAD_IS_TERMINATING, ERROR_ACCESS_DENIED, ERRDOS, ERRnoaccess, EPERM, "" )
S ( LW_STATUS_BAD_WORKING_SET_LIMIT, ERROR_INVALID_PARAMETER, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_INCOMPATIBLE_FILE_MAP, ERROR_INVALID_PARAMETER, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_SECTION_PROTECTION, ERROR_INVALID_PARAMETER, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_EAS_NOT_SUPPORTED, ERROR_EAS_NOT_SUPPORTED, ERRDOS, ERROR_EAS_NOT_SUPPORTED, -1, "" )
S ( LW_STATUS_EA_TOO_LARGE, ERROR_EA_LIST_INCONSISTENT, ERRDOS, ERROR_EAS_DIDNT_FIT, -1, "" )
S ( LW_STATUS_NONEXISTENT_EA_ENTRY, ERROR_FILE_CORRUPT, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_NO_EAS_ON_FILE, ERROR_FILE_CORRUPT, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_EA_CORRUPT_ERROR, ERROR_FILE_CORRUPT, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_FILE_LOCK_CONFLICT, ERROR_LOCK_VIOLATION, ERRDOS, ERRlock, EDEADLK, "" )
S ( LW_STATUS_LOCK_NOT_GRANTED, ERROR_LOCK_VIOLATION, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_DELETE_PENDING, ERROR_ACCESS_DENIED, ERRDOS, ERRnoaccess, EPERM, "" )
S ( LW_STATUS_CTL_FILE_NOT_SUPPORTED, ERROR_NOT_SUPPORTED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_UNKNOWN_REVISION, ERROR_UNKNOWN_REVISION, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_REVISION_MISMATCH, ERROR_REVISION_MISMATCH, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_INVALID_OWNER, ERROR_INVALID_OWNER, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_INVALID_PRIMARY_GROUP, ERROR_INVALID_PRIMARY_GROUP, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_NO_IMPERSONATION_TOKEN, ERROR_NO_IMPERSONATION_TOKEN, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_CANT_DISABLE_MANDATORY, ERROR_CANT_DISABLE_MANDATORY, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_NO_LOGON_SERVERS, ERROR_NO_LOGON_SERVERS, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_NO_SUCH_LOGON_SESSION, ERROR_NO_SUCH_LOGON_SESSION, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_NO_SUCH_PRIVILEGE, ERROR_NO_SUCH_PRIVILEGE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_PRIVILEGE_NOT_HELD, ERROR_PRIVILEGE_NOT_HELD, ERRDOS, ERRnoaccess, EPERM, "" )
S ( LW_STATUS_INVALID_ACCOUNT_NAME, ERROR_INVALID_ACCOUNT_NAME, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_USER_EXISTS, ERROR_USER_EXISTS, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_NO_SUCH_USER, ERROR_NO_SUCH_USER, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_GROUP_EXISTS, ERROR_GROUP_EXISTS, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_NO_SUCH_GROUP, ERROR_NO_SUCH_GROUP, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MEMBER_IN_GROUP, ERROR_MEMBER_IN_GROUP, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MEMBER_NOT_IN_GROUP, ERROR_MEMBER_NOT_IN_GROUP, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_LAST_ADMIN, ERROR_LAST_ADMIN, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_WRONG_PASSWORD, ERROR_INVALID_PASSWORD, ERRSRV, ERRbadpw, -1, "" )
S ( LW_STATUS_ILL_FORMED_PASSWORD, ERROR_ILL_FORMED_PASSWORD, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_PASSWORD_RESTRICTION, ERROR_PASSWORD_RESTRICTION, ERRSRV, ERRerror, -1, "Password does not meet requirements" )
S ( LW_STATUS_LOGON_FAILURE, ERROR_LOGON_FAILURE, ERRDOS, ERRnoaccess, EPERM, "" )
S ( LW_STATUS_ACCOUNT_RESTRICTION, ERROR_ACCOUNT_RESTRICTION, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_INVALID_LOGON_HOURS, ERROR_INVALID_LOGON_HOURS, ERRSRV, ERRbadLogonTime, -1, "" )
S ( LW_STATUS_INVALID_WORKSTATION, ERROR_INVALID_WORKSTATION, ERRSRV, ERRbadClient, -1, "" )
S ( LW_STATUS_PASSWORD_EXPIRED, ERROR_PASSWORD_EXPIRED, ERRSRV, ERRpasswordExpired, -1, "" )
S ( LW_STATUS_ACCOUNT_DISABLED, ERROR_ACCOUNT_DISABLED, ERRSRV, ERRaccountExpired, -1, "" )
S ( LW_STATUS_TOO_MANY_LUIDS_REQUESTED, ERROR_TOO_MANY_LUIDS_REQUESTED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_LUIDS_EXHAUSTED, ERROR_LUIDS_EXHAUSTED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_INVALID_SUB_AUTHORITY, ERROR_INVALID_SUB_AUTHORITY, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_INVALID_ACL, ERROR_INVALID_ACL, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_INVALID_SECURITY_DESCR, ERROR_INVALID_SECURITY_DESCR, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_PROCEDURE_NOT_FOUND, ERROR_PROC_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_INVALID_IMAGE_FORMAT, ERROR_BAD_EXE_FORMAT, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_NO_TOKEN, ERROR_NO_TOKEN, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_BAD_INHERITANCE_ACL, ERROR_BAD_INHERITANCE_ACL, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_RANGE_NOT_LOCKED, ERROR_NOT_LOCKED, ERRDOS, ERROR_NOT_LOCKED, -1, "" )
S ( LW_STATUS_DISK_FULL, ERROR_DISK_FULL, ERRHRD, ERRdiskfull, ENOSPC, "" )
S ( LW_STATUS_SERVER_DISABLED, ERROR_SERVER_DISABLED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_SERVER_NOT_DISABLED, ERROR_SERVER_NOT_DISABLED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_TOO_MANY_GUIDS_REQUESTED, ERROR_TOO_MANY_NAMES, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_GUIDS_EXHAUSTED, ERROR_NO_MORE_ITEMS, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_INVALID_ID_AUTHORITY, ERROR_INVALID_ID_AUTHORITY, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_AGENTS_EXHAUSTED, ERROR_NO_MORE_ITEMS, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_INVALID_VOLUME_LABEL, ERROR_LABEL_TOO_LONG, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_SECTION_NOT_EXTENDED, ERROR_OUTOFMEMORY, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_NOT_MAPPED_DATA, ERROR_INVALID_ADDRESS, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_RESOURCE_DATA_NOT_FOUND, ERROR_RESOURCE_DATA_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_RESOURCE_TYPE_NOT_FOUND, ERROR_RESOURCE_TYPE_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_RESOURCE_NAME_NOT_FOUND, ERROR_RESOURCE_NAME_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_ARRAY_BOUNDS_EXCEEDED, -1073741684, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_FLOAT_DENORMAL_OPERAND, -1073741683, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_FLOAT_DIVIDE_BY_ZERO, -1073741682, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_FLOAT_INEXACT_RESULT, -1073741681, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_FLOAT_INVALID_OPERATION, -1073741680, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_FLOAT_OVERFLOW, -1073741679, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_FLOAT_STACK_CHECK, -1073741678, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_FLOAT_UNDERFLOW, -1073741677, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_INTEGER_DIVIDE_BY_ZERO, -1073741676, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_INTEGER_OVERFLOW, ERROR_ARITHMETIC_OVERFLOW, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_PRIVILEGED_INSTRUCTION, -1073741674, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_TOO_MANY_PAGING_FILES, ERROR_NOT_ENOUGH_MEMORY, ERRDOS, ERRnomem, ENOMEM, "" )
S ( LW_STATUS_FILE_INVALID, ERROR_FILE_INVALID, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_ALLOTTED_SPACE_EXCEEDED, ERROR_ALLOTTED_SPACE_EXCEEDED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_DFS_EXIT_PATH_FOUND, ERROR_PATH_NOT_FOUND, ERRDOS, ERRbadpath, ENOENT, "" )
S ( LW_STATUS_DEVICE_DATA_ERROR, ERROR_CRC, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_DEVICE_NOT_CONNECTED, ERROR_DEVICE_NOT_CONNECTED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_DEVICE_POWER_FAILURE, ERROR_NOT_READY, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_FREE_VM_NOT_AT_BASE, ERROR_INVALID_ADDRESS, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MEMORY_NOT_ALLOCATED, ERROR_INVALID_ADDRESS, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_WORKING_SET_QUOTA, ERROR_WORKING_SET_QUOTA, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MEDIA_WRITE_PROTECTED, ERROR_WRITE_PROTECT, ERRHRD, ERRnowrite, EROFS, "" )
S ( LW_STATUS_DEVICE_NOT_READY, ERROR_NOT_READY, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_INVALID_GROUP_ATTRIBUTES, ERROR_INVALID_GROUP_ATTRIBUTES, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_BAD_IMPERSONATION_LEVEL, ERROR_BAD_IMPERSONATION_LEVEL, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_CANT_OPEN_ANONYMOUS, ERROR_CANT_OPEN_ANONYMOUS, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_BAD_VALIDATION_CLASS, ERROR_BAD_VALIDATION_CLASS, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_BAD_TOKEN_TYPE, ERROR_BAD_TOKEN_TYPE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_BAD_MASTER_BOOT_RECORD, ERROR_INVALID_PARAMETER, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_INSTRUCTION_MISALIGNMENT, ERROR_MR_MID_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_INSTANCE_NOT_AVAILABLE, ERROR_PIPE_BUSY, ERRDOS, ERRpipebusy, -1, "" )
S ( LW_STATUS_PIPE_NOT_AVAILABLE, ERROR_PIPE_BUSY, ERRDOS, ERRpipebusy, -1, "" )
S ( LW_STATUS_INVALID_PIPE_STATE, ERROR_BAD_PIPE, ERRDOS, ERRbadpipe, -1, "" )
S ( LW_STATUS_PIPE_BUSY, ERROR_PIPE_BUSY, ERRDOS, ERRpipebusy, -1, "" )
S ( LW_STATUS_ILLEGAL_FUNCTION, ERROR_INVALID_FUNCTION, ERRDOS, ERRbadfunc, -1, "Illegal function." )
S ( LW_STATUS_PIPE_DISCONNECTED, ERROR_PIPE_NOT_CONNECTED, ERRDOS, ERRnotconnected, -1, "" )
S ( LW_STATUS_PIPE_CLOSING, ERROR_NO_DATA, ERRDOS, ERRpipeclosing, -1, "" )
S ( LW_STATUS_PIPE_CONNECTED, ERROR_PIPE_CONNECTED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_PIPE_LISTENING, ERROR_PIPE_LISTENING, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_INVALID_READ_MODE, ERROR_BAD_PIPE, ERRDOS, ERRbadpipe, -1, "" )
S ( LW_STATUS_IO_TIMEOUT, ERROR_SEM_TIMEOUT, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_FILE_FORCED_CLOSED, ERROR_HANDLE_EOF, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_PROFILING_NOT_STARTED, ERROR_MR_MID_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_PROFILING_NOT_STOPPED, ERROR_MR_MID_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_COULD_NOT_INTERPRET, ERROR_MR_MID_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_FILE_IS_A_DIRECTORY, ERROR_ACCESS_DENIED, ERRDOS, ERRnoaccess, EPERM, "" )
S ( LW_STATUS_NOT_SUPPORTED, ERROR_NOT_SUPPORTED, ERRDOS, ERRunsup, -1, "" )
S ( LW_STATUS_REMOTE_NOT_LISTENING, ERROR_REM_NOT_LIST, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_DUPLICATE_NAME, ERROR_DUP_NAME, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_BAD_NETWORK_PATH, ERROR_BAD_NETPATH, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_NETWORK_BUSY, ERROR_NETWORK_BUSY, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_DEVICE_DOES_NOT_EXIST, ERROR_DEV_NOT_EXIST, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_TOO_MANY_COMMANDS, ERROR_TOO_MANY_CMDS, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_ADAPTER_HARDWARE_ERROR, ERROR_ADAP_HDW_ERR, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_INVALID_NETWORK_RESPONSE, ERROR_BAD_NET_RESP, ERRSRV, ERRerror, EBADMSG, "" )
S ( LW_STATUS_UNEXPECTED_NETWORK_ERROR, ERROR_UNEXP_NET_ERR, ERRSRV, ERRsrverror, -1, "" )
S ( LW_STATUS_BAD_REMOTE_ADAPTER, ERROR_BAD_REM_ADAP, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_PRINT_QUEUE_FULL, ERROR_PRINTQ_FULL, ERRSRV, ERRqfull, -1, "" )
S ( LW_STATUS_NO_SPOOL_SPACE, ERROR_NO_SPOOL_SPACE, ERRSRV, ERRqtoobig, -1, "" )
S ( LW_STATUS_PRINT_CANCELLED, ERROR_PRINT_CANCELLED, ERRSRV, ERRinvpfid, -1, "" )
S ( LW_STATUS_NETWORK_NAME_DELETED, ERROR_NETNAME_DELETED, ERRSRV, ERRinvtid, -1, "" )
S ( LW_STATUS_NETWORK_ACCESS_DENIED, ERROR_NETWORK_ACCESS_DENIED, ERRSRV, ERRaccess, -1, "" )
S ( LW_STATUS_BAD_DEVICE_TYPE, ERROR_BAD_DEV_TYPE, ERRSRV, ERRinvdevice, -1, "" )
S ( LW_STATUS_BAD_NETWORK_NAME, ERROR_BAD_NET_NAME, ERRSRV, ERRinvnetname, -1, "Network name not found." )
S ( LW_STATUS_TOO_MANY_NAMES, ERROR_TOO_MANY_NAMES, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_TOO_MANY_SESSIONS, ERROR_TOO_MANY_SESS, ERRSRV, ERRtoomanyuids, -1, "" )
S ( LW_STATUS_SHARING_PAUSED, ERROR_SHARING_PAUSED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_REQUEST_NOT_ACCEPTED, ERROR_REQ_NOT_ACCEP, ERRSRV, ERRnoresource, -1, "" )
S ( LW_STATUS_REDIRECTOR_PAUSED, ERROR_REDIR_PAUSED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_NET_WRITE_FAULT, ERROR_NET_WRITE_FAULT, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_PROFILING_AT_LIMIT, ERROR_MR_MID_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_NOT_SAME_DEVICE, ERROR_NOT_SAME_DEVICE, ERRDOS, ERRdiffdevice, EXDEV, "" )
S ( LW_STATUS_FILE_RENAMED, ERROR_ACCESS_DENIED, ERRDOS, ERRnoaccess, EPERM, "" )
S ( LW_STATUS_VIRTUAL_CIRCUIT_CLOSED, ERROR_VC_DISCONNECTED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_NO_SECURITY_ON_OBJECT, ERROR_NO_SECURITY_ON_OBJECT, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_CANT_WAIT, ERROR_MR_MID_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_PIPE_EMPTY, ERROR_NO_DATA, ERRDOS, ERRpipeclosing, -1, "" )
S ( LW_STATUS_CANT_ACCESS_DOMAIN_INFO, ERROR_CANT_ACCESS_DOMAIN_INFO, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_CANT_TERMINATE_SELF, ERROR_MR_MID_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_INVALID_SERVER_STATE, ERROR_INVALID_SERVER_STATE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_INVALID_DOMAIN_STATE, ERROR_INVALID_DOMAIN_STATE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_INVALID_DOMAIN_ROLE, ERROR_INVALID_DOMAIN_ROLE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_NO_SUCH_DOMAIN, ERROR_NO_SUCH_DOMAIN, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_DOMAIN_EXISTS, ERROR_DOMAIN_EXISTS, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_DOMAIN_LIMIT_EXCEEDED, ERROR_DOMAIN_LIMIT_EXCEEDED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_OPLOCK_NOT_GRANTED, ERROR_OPLOCK_NOT_GRANTED, ERRDOS, ERRlock, EDEADLK, "" )
S ( LW_STATUS_INVALID_OPLOCK_PROTOCOL, ERROR_INVALID_OPLOCK_PROTOCOL, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_INTERNAL_DB_CORRUPTION, ERROR_INTERNAL_DB_CORRUPTION, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_INTERNAL_ERROR, ERROR_INTERNAL_ERROR, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_GENERIC_NOT_MAPPED, ERROR_GENERIC_NOT_MAPPED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_BAD_DESCRIPTOR_FORMAT, ERROR_BAD_DESCRIPTOR_FORMAT, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_INVALID_USER_BUFFER, ERROR_INVALID_USER_BUFFER, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_UNEXPECTED_IO_ERROR, ERROR_MR_MID_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_UNEXPECTED_MM_CREATE_ERR, ERROR_MR_MID_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_UNEXPECTED_MM_MAP_ERROR, ERROR_MR_MID_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_UNEXPECTED_MM_EXTEND_ERR, ERROR_MR_MID_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_NOT_LOGON_PROCESS, ERROR_NOT_LOGON_PROCESS, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_LOGON_SESSION_EXISTS, ERROR_LOGON_SESSION_EXISTS, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_INVALID_PARAMETER_1, ERROR_INVALID_PARAMETER, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_INVALID_PARAMETER_2, ERROR_INVALID_PARAMETER, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_INVALID_PARAMETER_3, ERROR_INVALID_PARAMETER, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_INVALID_PARAMETER_4, ERROR_INVALID_PARAMETER, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_INVALID_PARAMETER_5, ERROR_INVALID_PARAMETER, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_INVALID_PARAMETER_6, ERROR_INVALID_PARAMETER, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_INVALID_PARAMETER_7, ERROR_INVALID_PARAMETER, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_INVALID_PARAMETER_8, ERROR_INVALID_PARAMETER, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_INVALID_PARAMETER_9, ERROR_INVALID_PARAMETER, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_INVALID_PARAMETER_10, ERROR_INVALID_PARAMETER, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_INVALID_PARAMETER_11, ERROR_INVALID_PARAMETER, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_INVALID_PARAMETER_12, ERROR_INVALID_PARAMETER, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_REDIRECTOR_NOT_STARTED, ERROR_PATH_NOT_FOUND, ERRDOS, ERRbadpath, ENOENT, "" )
S ( LW_STATUS_REDIRECTOR_STARTED, ERROR_SERVICE_ALREADY_RUNNING, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_STACK_OVERFLOW, ERROR_STACK_OVERFLOW, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_NO_SUCH_PACKAGE, ERROR_NO_SUCH_PACKAGE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_BAD_FUNCTION_TABLE, ERROR_MR_MID_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_VARIABLE_NOT_FOUND, ERROR_ENVVAR_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_DIRECTORY_NOT_EMPTY, ERROR_DIR_NOT_EMPTY, ERRDOS, ERRremcd, EPERM, "" )
S ( LW_STATUS_FILE_CORRUPT_ERROR, ERROR_FILE_CORRUPT, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_NOT_A_DIRECTORY, ERROR_DIRECTORY, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_BAD_LOGON_SESSION_STATE, ERROR_BAD_LOGON_SESSION_STATE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_LOGON_SESSION_COLLISION, ERROR_LOGON_SESSION_COLLISION, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_NAME_TOO_LONG, ERROR_FILENAME_EXCED_RANGE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_FILES_OPEN, NERR_OpenFiles, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_CONNECTION_IN_USE, NERR_DevInUse, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_MR_MID_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_PROCESS_IS_TERMINATING, ERROR_ACCESS_DENIED, ERRDOS, ERRnoaccess, EPERM, "" )
S ( LW_STATUS_INVALID_LOGON_TYPE, ERROR_INVALID_LOGON_TYPE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_NO_GUID_TRANSLATION, ERROR_MR_MID_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_CANNOT_IMPERSONATE, ERROR_CANNOT_IMPERSONATE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_IMAGE_ALREADY_LOADED, ERROR_SERVICE_ALREADY_RUNNING, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_ABIOS_NOT_PRESENT, ERROR_MR_MID_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_ABIOS_LID_NOT_EXIST, ERROR_MR_MID_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_ABIOS_LID_ALREADY_OWNED, ERROR_MR_MID_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_ABIOS_NOT_LID_OWNER, ERROR_MR_MID_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_ABIOS_INVALID_COMMAND, ERROR_MR_MID_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_ABIOS_INVALID_LID, ERROR_MR_MID_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_ABIOS_SELECTOR_NOT_AVAILABLE, ERROR_MR_MID_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_ABIOS_INVALID_SELECTOR, ERROR_MR_MID_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_NO_LDT, ERROR_INVALID_THREAD_ID, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_INVALID_LDT_SIZE, ERROR_MR_MID_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_INVALID_LDT_OFFSET, ERROR_MR_MID_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_INVALID_LDT_DESCRIPTOR, ERROR_MR_MID_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_INVALID_IMAGE_NE_FORMAT, ERROR_BAD_EXE_FORMAT, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_RXACT_INVALID_STATE, ERROR_RXACT_INVALID_STATE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_RXACT_COMMIT_FAILURE, ERROR_RXACT_COMMIT_FAILURE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MAPPED_FILE_SIZE_ZERO, ERROR_FILE_INVALID, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_TOO_MANY_OPENED_FILES, ERROR_TOO_MANY_OPEN_FILES, ERRDOS, ERRnofids, EMFILE, "" )
S ( LW_STATUS_CANCELLED, ERROR_OPERATION_ABORTED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_CANNOT_DELETE, ERROR_ACCESS_DENIED, ERRDOS, ERRnoaccess, EPERM, "" )
S ( LW_STATUS_INVALID_COMPUTER_NAME, ERROR_INVALID_COMPUTERNAME, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_FILE_DELETED, ERROR_ACCESS_DENIED, ERRDOS, ERRnoaccess, EPERM, "" )
S ( LW_STATUS_SPECIAL_GROUP, ERROR_SPECIAL_GROUP, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_SPECIAL_USER, ERROR_SPECIAL_USER, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MEMBERS_PRIMARY_GROUP, ERROR_MEMBERS_PRIMARY_GROUP, ERRSRV, ERRerror, -1, "Group is a primary group of one or more users" )
S ( LW_STATUS_FILE_CLOSED, ERROR_INVALID_HANDLE, ERRDOS, ERRbadfid, EBADF, "" )
S ( LW_STATUS_TOO_MANY_THREADS, ERROR_MR_MID_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_THREAD_NOT_IN_PROCESS, ERROR_MR_MID_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_TOKEN_ALREADY_IN_USE, ERROR_TOKEN_ALREADY_IN_USE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_PAGEFILE_QUOTA_EXCEEDED, ERROR_MR_MID_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_COMMITMENT_LIMIT, ERROR_COMMITMENT_LIMIT, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_INVALID_IMAGE_LE_FORMAT, ERROR_BAD_EXE_FORMAT, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_INVALID_IMAGE_NOT_MZ, ERROR_BAD_EXE_FORMAT, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_INVALID_IMAGE_PROTECT, ERROR_BAD_EXE_FORMAT, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_INVALID_IMAGE_WIN_16, ERROR_BAD_EXE_FORMAT, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_LOGON_SERVER_CONFLICT, ERROR_MR_MID_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_TIME_DIFFERENCE_AT_DC, ERROR_TIME_SKEW, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_SYNCHRONIZATION_REQUIRED, ERROR_MR_MID_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_DLL_NOT_FOUND, ERROR_MOD_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_OPEN_FAILED, ERROR_MR_MID_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_IO_PRIVILEGE_FAILED, ERROR_MR_MID_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_ORDINAL_NOT_FOUND, ERROR_INVALID_ORDINAL, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_ENTRYPOINT_NOT_FOUND, ERROR_PROC_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_CONTROL_C_EXIT, ERROR_MR_MID_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_LOCAL_DISCONNECT, ERROR_NETNAME_DELETED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_REMOTE_DISCONNECT, ERROR_NETNAME_DELETED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_REMOTE_RESOURCES, ERROR_REM_NOT_LIST, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_LINK_FAILED, ERROR_UNEXP_NET_ERR, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_LINK_TIMEOUT, ERROR_UNEXP_NET_ERR, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_INVALID_CONNECTION, ERROR_UNEXP_NET_ERR, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_INVALID_ADDRESS, ERROR_UNEXP_NET_ERR, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_DLL_INIT_FAILED, ERROR_DLL_INIT_FAILED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MISSING_SYSTEMFILE, ERROR_MR_MID_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_APP_INIT_FAILURE, ERROR_MR_MID_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_PAGEFILE_CREATE_FAILED, ERROR_MR_MID_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_NO_PAGEFILE, ERROR_MR_MID_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_INVALID_LEVEL, ERROR_INVALID_LEVEL, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_WRONG_PASSWORD_CORE, ERROR_INVALID_PASSWORD, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_ILLEGAL_FLOAT_CONTEXT, ERROR_MR_MID_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_PIPE_BROKEN, ERROR_BROKEN_PIPE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_REGISTRY_CORRUPT, ERROR_BADDB, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_REGISTRY_IO_FAILED, ERROR_REGISTRY_IO_FAILED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_NO_EVENT_PAIR, ERROR_MR_MID_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_UNRECOGNIZED_VOLUME, ERROR_UNRECOGNIZED_VOLUME, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_SERIAL_NO_DEVICE_INITED, ERROR_SERIAL_NO_DEVICE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_NO_SUCH_ALIAS, ERROR_NO_SUCH_ALIAS, ERRSRV, ERRerror, -1, "No such local group" )
S ( LW_STATUS_MEMBER_NOT_IN_ALIAS, ERROR_MEMBER_NOT_IN_ALIAS, ERRSRV, ERRerror, -1, "User is not a member of local group" )
S ( LW_STATUS_MEMBER_IN_ALIAS, ERROR_MEMBER_IN_ALIAS, ERRSRV, ERRerror, -1, "User is already in local group" )
S ( LW_STATUS_ALIAS_EXISTS, ERROR_ALIAS_EXISTS, ERRSRV, ERRerror, -1, "Local group exists" )
S ( LW_STATUS_LOGON_NOT_GRANTED, ERROR_LOGON_NOT_GRANTED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_TOO_MANY_SECRETS, ERROR_TOO_MANY_SECRETS, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_SECRET_TOO_LONG, ERROR_SECRET_TOO_LONG, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_INTERNAL_DB_ERROR, ERROR_INTERNAL_DB_ERROR, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_FULLSCREEN_MODE, ERROR_FULLSCREEN_MODE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_TOO_MANY_CONTEXT_IDS, ERROR_TOO_MANY_CONTEXT_IDS, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_LOGON_TYPE_NOT_GRANTED, ERROR_LOGON_TYPE_NOT_GRANTED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_NOT_REGISTRY_FILE, ERROR_NOT_REGISTRY_FILE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_NT_CROSS_ENCRYPTION_REQUIRED, ERROR_NT_CROSS_ENCRYPTION_REQUIRED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_DOMAIN_CTRLR_CONFIG_ERROR, ERROR_MR_MID_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_FT_MISSING_MEMBER, ERROR_IO_DEVICE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_ILL_FORMED_SERVICE_ENTRY, ERROR_MR_MID_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_ILLEGAL_CHARACTER, ERROR_MR_MID_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_UNMAPPABLE_CHARACTER, ERROR_NO_UNICODE_TRANSLATION, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_UNDEFINED_CHARACTER, ERROR_MR_MID_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_FLOPPY_VOLUME, ERROR_MR_MID_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_FLOPPY_ID_MARK_NOT_FOUND, ERROR_FLOPPY_ID_MARK_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_FLOPPY_WRONG_CYLINDER, ERROR_FLOPPY_WRONG_CYLINDER, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_FLOPPY_UNKNOWN_ERROR, ERROR_FLOPPY_UNKNOWN_ERROR, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_FLOPPY_BAD_REGISTERS, ERROR_FLOPPY_BAD_REGISTERS, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_DISK_RECALIBRATE_FAILED, ERROR_DISK_RECALIBRATE_FAILED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_DISK_OPERATION_FAILED, ERROR_DISK_OPERATION_FAILED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_DISK_RESET_FAILED, ERROR_DISK_RESET_FAILED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_SHARED_IRQ_BUSY, ERROR_IRQ_BUSY, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_FT_ORPHANING, ERROR_IO_DEVICE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_BIOS_FAILED_TO_CONNECT_INTERRUPT, ERROR_MR_MID_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_PARTITION_FAILURE, ERROR_PARTITION_FAILURE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_INVALID_BLOCK_LENGTH, ERROR_INVALID_BLOCK_LENGTH, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_DEVICE_NOT_PARTITIONED, ERROR_DEVICE_NOT_PARTITIONED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_UNABLE_TO_LOCK_MEDIA, ERROR_UNABLE_TO_LOCK_MEDIA, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_UNABLE_TO_UNLOAD_MEDIA, ERROR_UNABLE_TO_UNLOAD_MEDIA, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_EOM_OVERFLOW, ERROR_EOM_OVERFLOW, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_NO_MEDIA, ERROR_NO_MEDIA_IN_DRIVE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_NO_SUCH_MEMBER, ERROR_NO_SUCH_MEMBER, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_INVALID_MEMBER, ERROR_INVALID_MEMBER, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_KEY_DELETED, ERROR_KEY_DELETED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_NO_LOG_SPACE, ERROR_NO_LOG_SPACE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_TOO_MANY_SIDS, ERROR_TOO_MANY_SIDS, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_LM_CROSS_ENCRYPTION_REQUIRED, ERROR_LM_CROSS_ENCRYPTION_REQUIRED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_KEY_HAS_CHILDREN, ERROR_KEY_HAS_CHILDREN, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_CHILD_MUST_BE_VOLATILE, ERROR_CHILD_MUST_BE_VOLATILE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_DEVICE_CONFIGURATION_ERROR, ERROR_INVALID_PARAMETER, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_DRIVER_INTERNAL_ERROR, ERROR_IO_DEVICE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_INVALID_DEVICE_STATE, ERROR_BAD_COMMAND, ERRHRD, ERRbadcmd, -1, "" )
S ( LW_STATUS_IO_DEVICE_ERROR, ERROR_IO_DEVICE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_DEVICE_PROTOCOL_ERROR, ERROR_IO_DEVICE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_BACKUP_CONTROLLER, ERROR_MR_MID_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_LOG_FILE_FULL, ERROR_LOG_FILE_FULL, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_TOO_LATE, ERROR_WRITE_PROTECT, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_NO_TRUST_LSA_SECRET, ERROR_NO_TRUST_LSA_SECRET, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_NO_TRUST_SAM_ACCOUNT, ERROR_NO_TRUST_SAM_ACCOUNT, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_TRUSTED_DOMAIN_FAILURE, ERROR_TRUSTED_DOMAIN_FAILURE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_TRUSTED_RELATIONSHIP_FAILURE, ERROR_TRUSTED_RELATIONSHIP_FAILURE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_EVENTLOG_FILE_CORRUPT, ERROR_EVENTLOG_FILE_CORRUPT, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_EVENTLOG_CANT_START, ERROR_EVENTLOG_CANT_START, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_TRUST_FAILURE, ERROR_TRUST_FAILURE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MUTANT_LIMIT_EXCEEDED, ERROR_MR_MID_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_NETLOGON_NOT_STARTED, ERROR_NETLOGON_NOT_STARTED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_ACCOUNT_EXPIRED, ERROR_ACCOUNT_EXPIRED, ERRSRV, ERRaccountExpired, -1, "" )
S ( LW_STATUS_POSSIBLE_DEADLOCK, ERROR_POSSIBLE_DEADLOCK, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_NETWORK_CREDENTIAL_CONFLICT, ERROR_SESSION_CREDENTIAL_CONFLICT, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_REMOTE_SESSION_LIMIT, ERROR_REMOTE_SESSION_LIMIT_EXCEEDED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_EVENTLOG_FILE_CHANGED, ERROR_EVENTLOG_FILE_CHANGED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_NOLOGON_INTERDOMAIN_TRUST_ACCOUNT, ERROR_NOLOGON_INTERDOMAIN_TRUST_ACCOUNT, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_NOLOGON_WORKSTATION_TRUST_ACCOUNT, ERROR_NOLOGON_WORKSTATION_TRUST_ACCOUNT, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_NOLOGON_SERVER_TRUST_ACCOUNT, ERROR_NOLOGON_SERVER_TRUST_ACCOUNT, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_DOMAIN_TRUST_INCONSISTENT, ERROR_DOMAIN_TRUST_INCONSISTENT, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_FS_DRIVER_REQUIRED, ERROR_MR_MID_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_IMAGE_ALREADY_LOADED_AS_DLL, ERROR_MR_MID_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_NETWORK_OPEN_RESTRICTION, ERROR_MR_MID_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_NO_USER_SESSION_KEY, ERROR_NO_USER_SESSION_KEY, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_USER_SESSION_DELETED, ERROR_UNEXP_NET_ERR, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_RESOURCE_LANG_NOT_FOUND, ERROR_RESOURCE_LANG_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_INSUFF_SERVER_RESOURCES, ERROR_NOT_ENOUGH_SERVER_MEMORY, ERRDOS, ERRnomem, ENOMEM, "" )
S ( LW_STATUS_INVALID_BUFFER_SIZE, ERROR_INVALID_USER_BUFFER, ERRSRV, ERRerror, EFAULT, "" )
S ( LW_STATUS_INVALID_ADDRESS_COMPONENT, ERROR_INVALID_NETNAME, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_INVALID_ADDRESS_WILDCARD, ERROR_INVALID_NETNAME, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_TOO_MANY_ADDRESSES, ERROR_TOO_MANY_NAMES, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_ADDRESS_ALREADY_EXISTS, ERROR_DUP_NAME, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_ADDRESS_CLOSED, ERROR_NETNAME_DELETED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_CONNECTION_DISCONNECTED, ERROR_NETNAME_DELETED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_CONNECTION_RESET, ERROR_NETNAME_DELETED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_TOO_MANY_NODES, ERROR_TOO_MANY_NAMES, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_TRANSACTION_ABORTED, ERROR_UNEXP_NET_ERR, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_TRANSACTION_TIMED_OUT, ERROR_UNEXP_NET_ERR, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_TRANSACTION_NO_RELEASE, ERROR_UNEXP_NET_ERR, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_TRANSACTION_NO_MATCH, ERROR_UNEXP_NET_ERR, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_TRANSACTION_RESPONDED, ERROR_UNEXP_NET_ERR, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_TRANSACTION_INVALID_ID, ERROR_UNEXP_NET_ERR, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_TRANSACTION_INVALID_TYPE, ERROR_UNEXP_NET_ERR, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_NOT_SERVER_SESSION, ERROR_NOT_SUPPORTED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_NOT_CLIENT_SESSION, ERROR_NOT_SUPPORTED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_CANNOT_LOAD_REGISTRY_FILE, ERROR_MR_MID_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_DEBUG_ATTACH_FAILED, ERROR_MR_MID_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_SYSTEM_PROCESS_TERMINATED, ERROR_MR_MID_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_DATA_NOT_ACCEPTED, ERROR_MR_MID_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_NO_BROWSER_SERVERS_FOUND, ERROR_NO_BROWSER_SERVERS_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_VDM_HARD_ERROR, ERROR_MR_MID_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_DRIVER_CANCEL_TIMEOUT, ERROR_MR_MID_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_REPLY_MESSAGE_MISMATCH, ERROR_MR_MID_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MAPPED_ALIGNMENT, ERROR_MAPPED_ALIGNMENT, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_IMAGE_CHECKSUM_MISMATCH, ERROR_BAD_EXE_FORMAT, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_LOST_WRITEBEHIND_DATA, ERROR_MR_MID_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_CLIENT_SERVER_PARAMETERS_INVALID, ERROR_MR_MID_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_PASSWORD_MUST_CHANGE, ERROR_PASSWORD_MUST_CHANGE, ERRSRV, ERRpasswordExpired, -1, "" )
S ( LW_STATUS_NOT_FOUND, ERROR_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_NOT_TINY_STREAM, ERROR_MR_MID_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_RECOVERY_FAILURE, ERROR_MR_MID_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_STACK_OVERFLOW_READ, ERROR_MR_MID_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_FAIL_CHECK, ERROR_INVALID_PARAMETER, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_DUPLICATE_OBJECTID, -1073741270, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_OBJECTID_EXISTS, -1073741269, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_CONVERT_TO_LARGE, ERROR_MR_MID_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_RETRY, ERROR_RETRY, ERRSRV, ERRerror, EAGAIN, "The request needs to be retried" )
S ( LW_STATUS_FOUND_OUT_OF_SCOPE, ERROR_MR_MID_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_ALLOCATE_BUCKET, ERROR_MR_MID_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_PROPSET_NOT_FOUND, ERROR_SET_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MARSHALL_OVERFLOW, ERROR_MR_MID_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_INVALID_VARIANT, ERROR_MR_MID_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_DOMAIN_CONTROLLER_NOT_FOUND, ERROR_DOMAIN_CONTROLLER_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_ACCOUNT_LOCKED_OUT, ERROR_ACCOUNT_LOCKED_OUT, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_HANDLE_NOT_CLOSABLE, ERROR_INVALID_HANDLE, ERRDOS, ERRbadfid, EBADF, "" )
S ( LW_STATUS_GRACEFUL_DISCONNECT, ERROR_GRACEFUL_DISCONNECT, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_ADDRESS_ALREADY_ASSOCIATED, ERROR_ADDRESS_ALREADY_ASSOCIATED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_ADDRESS_NOT_ASSOCIATED, ERROR_ADDRESS_NOT_ASSOCIATED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_CONNECTION_INVALID, ERROR_CONNECTION_INVALID, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_CONNECTION_ACTIVE, ERROR_CONNECTION_ACTIVE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_NETWORK_UNREACHABLE, ERROR_NETWORK_UNREACHABLE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_HOST_UNREACHABLE, ERROR_HOST_UNREACHABLE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_PROTOCOL_UNREACHABLE, ERROR_PROTOCOL_UNREACHABLE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_PORT_UNREACHABLE, ERROR_PORT_UNREACHABLE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_REQUEST_ABORTED, ERROR_REQUEST_ABORTED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_CONNECTION_ABORTED, ERROR_CONNECTION_ABORTED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_BAD_COMPRESSION_BUFFER, ERROR_MR_MID_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_USER_MAPPED_FILE, ERROR_USER_MAPPED_FILE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_AUDIT_FAILED, ERROR_MR_MID_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_TIMER_RESOLUTION_NOT_SET, ERROR_MR_MID_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_CONNECTION_COUNT_LIMIT, ERROR_CONNECTION_COUNT_LIMIT, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_LOGIN_TIME_RESTRICTION, ERROR_LOGIN_TIME_RESTRICTION, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_LOGIN_WKSTA_RESTRICTION, ERROR_LOGIN_WKSTA_RESTRICTION, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_IMAGE_MP_UP_MISMATCH, ERROR_BAD_EXE_FORMAT, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_INSUFFICIENT_LOGON_INFO, ERROR_MR_MID_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_BAD_DLL_ENTRYPOINT, ERROR_MR_MID_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_BAD_SERVICE_ENTRYPOINT, ERROR_MR_MID_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_LPC_REPLY_LOST, ERROR_INTERNAL_ERROR, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_IP_ADDRESS_CONFLICT1, ERROR_MR_MID_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_IP_ADDRESS_CONFLICT2, ERROR_MR_MID_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_REGISTRY_QUOTA_LIMIT, ERROR_MR_MID_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_PATH_NOT_COVERED, ERROR_HOST_UNREACHABLE, ERRSRV, ERRbadpath, -1, "" )
S ( LW_STATUS_NO_CALLBACK_ACTIVE, ERROR_MR_MID_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_LICENSE_QUOTA_EXCEEDED, ERROR_LICENSE_QUOTA_EXCEEDED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_PWD_TOO_SHORT, ERROR_MR_MID_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_PWD_TOO_RECENT, ERROR_MR_MID_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_PWD_HISTORY_CONFLICT, ERROR_MR_MID_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_PLUGPLAY_NO_DEVICE, ERROR_SERVICE_DISABLED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_UNSUPPORTED_COMPRESSION, ERROR_MR_MID_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_INVALID_HW_PROFILE, ERROR_MR_MID_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_INVALID_PLUGPLAY_DEVICE_PATH, ERROR_MR_MID_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_DRIVER_ORDINAL_NOT_FOUND, ERROR_INVALID_ORDINAL, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_DRIVER_ENTRYPOINT_NOT_FOUND, ERROR_PROC_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_RESOURCE_NOT_OWNED, ERROR_NOT_OWNER, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_TOO_MANY_LINKS, ERROR_TOO_MANY_LINKS, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_QUOTA_LIST_INCONSISTENT, ERROR_MR_MID_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_FILE_IS_OFFLINE, ERROR_FILE_OFFLINE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_EVALUATION_EXPIRATION, ERROR_MR_MID_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_ILLEGAL_DLL_RELOCATION, ERROR_MR_MID_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_LICENSE_VIOLATION, ERROR_CTX_LICENSE_NOT_AVAILABLE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_DLL_INIT_FAILED_LOGOFF, ERROR_MR_MID_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_DRIVER_UNABLE_TO_LOAD, ERROR_BAD_DRIVER, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_DFS_UNAVAILABLE, ERROR_CONNECTION_UNAVAIL, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_VOLUME_DISMOUNTED, ERROR_NOT_READY, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_WX86_INTERNAL_ERROR, ERROR_MR_MID_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_WX86_FLOAT_STACK_CHECK, ERROR_MR_MID_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_VALIDATE_CONTINUE, ERROR_MR_MID_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_NO_MATCH, ERROR_NO_MATCH, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_NO_MORE_MATCHES, ERROR_NO_MORE_MATCHES, ERRSRV, ERRerror, -1, "No more find results match query" )
S ( LW_STATUS_NOT_A_REPARSE_POINT, ERROR_NOT_A_REPARSE_POINT, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_IO_REPARSE_TAG_INVALID, ERROR_REPARSE_TAG_INVALID, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_IO_REPARSE_TAG_MISMATCH, ERROR_REPARSE_TAG_MISMATCH, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_IO_REPARSE_DATA_INVALID, ERROR_INVALID_REPARSE_DATA, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_IO_REPARSE_TAG_NOT_HANDLED, ERROR_CANT_ACCESS_FILE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_REPARSE_POINT_NOT_RESOLVED, ERROR_CANT_RESOLVE_FILENAME, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_DIRECTORY_IS_A_REPARSE_POINT, ERROR_BAD_PATHNAME, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_RANGE_LIST_CONFLICT, ERROR_MR_MID_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_SOURCE_ELEMENT_EMPTY, ERROR_SOURCE_ELEMENT_EMPTY, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_DESTINATION_ELEMENT_FULL, ERROR_DESTINATION_ELEMENT_FULL, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_ILLEGAL_ELEMENT_ADDRESS, ERROR_ILLEGAL_ELEMENT_ADDRESS, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MAGAZINE_NOT_PRESENT, ERROR_MAGAZINE_NOT_PRESENT, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_REINITIALIZATION_NEEDED, ERROR_DEVICE_REINITIALIZATION_NEEDED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_ENCRYPTION_FAILED, ERROR_ACCESS_DENIED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_DECRYPTION_FAILED, ERROR_ACCESS_DENIED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_RANGE_NOT_FOUND, ERROR_MR_MID_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_NO_RECOVERY_POLICY, ERROR_ACCESS_DENIED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_NO_EFS, ERROR_ACCESS_DENIED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_WRONG_EFS, ERROR_ACCESS_DENIED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_NO_USER_KEYS, ERROR_ACCESS_DENIED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_FILE_NOT_ENCRYPTED, ERROR_FILE_NOT_ENCRYPTED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_NOT_EXPORT_FORMAT, ERROR_NOT_EXPORT_FORMAT, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_FILE_ENCRYPTED, ERROR_FILE_ENCRYPTED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_WMI_GUID_NOT_FOUND, ERROR_WMI_GUID_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_WMI_INSTANCE_NOT_FOUND, ERROR_WMI_INSTANCE_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_WMI_ITEMID_NOT_FOUND, ERROR_WMI_ITEMID_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_WMI_TRY_AGAIN, ERROR_WMI_TRY_AGAIN, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_SHARED_POLICY, ERROR_SHARED_POLICY, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_POLICY_OBJECT_NOT_FOUND, ERROR_POLICY_OBJECT_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_POLICY_ONLY_IN_DS, ERROR_POLICY_ONLY_IN_DS, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_VOLUME_NOT_UPGRADED, ERROR_INVALID_FUNCTION, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_REMOTE_STORAGE_NOT_ACTIVE, ERROR_REMOTE_STORAGE_NOT_ACTIVE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_REMOTE_STORAGE_MEDIA_ERROR, ERROR_REMOTE_STORAGE_MEDIA_ERROR, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_NO_TRACKING_SERVICE, ERROR_NO_TRACKING_SERVICE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_SERVER_SID_MISMATCH, ERROR_MR_MID_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_DS_NO_ATTRIBUTE_OR_VALUE, ERROR_DS_NO_ATTRIBUTE_OR_VALUE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_DS_INVALID_ATTRIBUTE_SYNTAX, ERROR_DS_INVALID_ATTRIBUTE_SYNTAX, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_DS_ATTRIBUTE_TYPE_UNDEFINED, ERROR_DS_ATTRIBUTE_TYPE_UNDEFINED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_DS_ATTRIBUTE_OR_VALUE_EXISTS, ERROR_DS_ATTRIBUTE_OR_VALUE_EXISTS, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_DS_BUSY, ERROR_DS_BUSY, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_DS_UNAVAILABLE, ERROR_DS_UNAVAILABLE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_DS_NO_RIDS_ALLOCATED, ERROR_DS_NO_RIDS_ALLOCATED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_DS_NO_MORE_RIDS, ERROR_DS_NO_MORE_RIDS, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_DS_INCORRECT_ROLE_OWNER, ERROR_DS_INCORRECT_ROLE_OWNER, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_DS_RIDMGR_INIT_ERROR, ERROR_DS_RIDMGR_INIT_ERROR, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_DS_OBJ_CLASS_VIOLATION, ERROR_DS_OBJ_CLASS_VIOLATION, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_DS_CANT_ON_NON_LEAF, ERROR_DS_CANT_ON_NON_LEAF, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_DS_CANT_ON_RDN, ERROR_DS_CANT_ON_RDN, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_DS_CANT_MOD_OBJ_CLASS, ERROR_DS_CANT_MOD_OBJ_CLASS, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_DS_CROSS_DOM_MOVE_FAILED, ERROR_DS_CROSS_DOM_MOVE_ERROR, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_DS_GC_NOT_AVAILABLE, ERROR_DS_GC_NOT_AVAILABLE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_DIRECTORY_SERVICE_REQUIRED, ERROR_DS_DS_REQUIRED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_REPARSE_ATTRIBUTE_CONFLICT, ERROR_REPARSE_ATTRIBUTE_CONFLICT, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_CANT_ENABLE_DENY_ONLY, ERROR_MR_MID_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_FLOAT_MULTIPLE_FAULTS, ERROR_MR_MID_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_FLOAT_MULTIPLE_TRAPS, ERROR_MR_MID_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_DEVICE_REMOVED, ERROR_DEVICE_REMOVED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_JOURNAL_DELETE_IN_PROGRESS, ERROR_JOURNAL_DELETE_IN_PROGRESS, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_JOURNAL_NOT_ACTIVE, ERROR_JOURNAL_NOT_ACTIVE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_NOINTERFACE, ERROR_MR_MID_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_DS_ADMIN_LIMIT_EXCEEDED, ERROR_DS_ADMIN_LIMIT_EXCEEDED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_DRIVER_FAILED_SLEEP, ERROR_MR_MID_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MUTUAL_AUTHENTICATION_FAILED, ERROR_MUTUAL_AUTH_FAILED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_CORRUPT_SYSTEM_FILE, ERROR_MR_MID_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_DATATYPE_MISALIGNMENT_ERROR, ERROR_NOACCESS, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_WMI_READ_ONLY, ERROR_WMI_READ_ONLY, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_WMI_SET_FAILURE, ERROR_WMI_SET_FAILURE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_COMMITMENT_MINIMUM, ERROR_MR_MID_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_REG_NAT_CONSUMPTION, ERROR_REG_NAT_CONSUMPTION, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_TRANSPORT_FULL, ERROR_TRANSPORT_FULL, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_DS_SAM_INIT_FAILURE, ERROR_DS_SAM_INIT_FAILURE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_ONLY_IF_CONNECTED, ERROR_ONLY_IF_CONNECTED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_DS_SENSITIVE_GROUP_VIOLATION, ERROR_DS_SENSITIVE_GROUP_VIOLATION, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_PNP_RESTART_ENUMERATION, ERROR_MR_MID_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_JOURNAL_ENTRY_DELETED, ERROR_JOURNAL_ENTRY_DELETED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_DS_CANT_MOD_PRIMARYGROUPID, ERROR_DS_CANT_MOD_PRIMARYGROUPID, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_SYSTEM_IMAGE_BAD_SIGNATURE, ERROR_MR_MID_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_PNP_REBOOT_REQUIRED, ERROR_MR_MID_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_POWER_STATE_INVALID, ERROR_MR_MID_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_DS_INVALID_GROUP_TYPE, ERROR_DS_INVALID_GROUP_TYPE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_DS_NO_NEST_GLOBALGROUP_IN_MIXEDDOMAIN, ERROR_DS_NO_NEST_GLOBALGROUP_IN_MIXEDDOMAIN, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_DS_NO_NEST_LOCALGROUP_IN_MIXEDDOMAIN, ERROR_DS_NO_NEST_LOCALGROUP_IN_MIXEDDOMAIN, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_DS_GLOBAL_CANT_HAVE_LOCAL_MEMBER, ERROR_DS_GLOBAL_CANT_HAVE_LOCAL_MEMBER, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_DS_GLOBAL_CANT_HAVE_UNIVERSAL_MEMBER, ERROR_DS_GLOBAL_CANT_HAVE_UNIVERSAL_MEMBER, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_DS_UNIVERSAL_CANT_HAVE_LOCAL_MEMBER, ERROR_DS_UNIVERSAL_CANT_HAVE_LOCAL_MEMBER, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_DS_GLOBAL_CANT_HAVE_CROSSDOMAIN_MEMBER, ERROR_DS_GLOBAL_CANT_HAVE_CROSSDOMAIN_MEMBER, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_DS_LOCAL_CANT_HAVE_CROSSDOMAIN_LOCAL_MEMBER, ERROR_DS_LOCAL_CANT_HAVE_CROSSDOMAIN_LOCAL_MEMBER, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_DS_HAVE_PRIMARY_MEMBERS, ERROR_DS_HAVE_PRIMARY_MEMBERS, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_WMI_NOT_SUPPORTED, ERROR_NOT_SUPPORTED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_INSUFFICIENT_POWER, ERROR_MR_MID_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_SAM_NEED_BOOTKEY_PASSWORD, ERROR_DS_SAM_NEED_BOOTKEY_PASSWORD, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_SAM_NEED_BOOTKEY_FLOPPY, ERROR_DS_SAM_NEED_BOOTKEY_FLOPPY, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_DS_CANT_START, ERROR_DS_CANT_START, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_DS_INIT_FAILURE, ERROR_DS_INIT_FAILURE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_SAM_INIT_FAILURE, ERROR_SAM_INIT_FAILURE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_DS_GC_REQUIRED, ERROR_DS_GC_REQUIRED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_DS_LOCAL_MEMBER_OF_LOCAL_ONLY, ERROR_DS_LOCAL_MEMBER_OF_LOCAL_ONLY, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_DS_NO_FPO_IN_UNIVERSAL_GROUPS, ERROR_DS_NO_FPO_IN_UNIVERSAL_GROUPS, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_DS_MACHINE_ACCOUNT_QUOTA_EXCEEDED, ERROR_DS_MACHINE_ACCOUNT_QUOTA_EXCEEDED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MULTIPLE_FAULT_VIOLATION, ERROR_MR_MID_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_CURRENT_DOMAIN_NOT_ALLOWED, ERROR_CURRENT_DOMAIN_NOT_ALLOWED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_CANNOT_MAKE, ERROR_CANNOT_MAKE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_SYSTEM_SHUTDOWN, ERROR_MR_MID_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_DS_INIT_FAILURE_CONSOLE, ERROR_DS_INIT_FAILURE_CONSOLE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_DS_SAM_INIT_FAILURE_CONSOLE, ERROR_DS_SAM_INIT_FAILURE_CONSOLE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_UNFINISHED_CONTEXT_DELETED, -2146893005, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_NO_TGT_REPLY, -2146893004, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_OBJECTID_NOT_FOUND, ERROR_FILE_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_NO_IP_ADDRESSES, -2146893003, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_WRONG_CREDENTIAL_HANDLE, -2146893002, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_CRYPTO_SYSTEM_INVALID, -2146893001, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MAX_REFERRALS_EXCEEDED, -2146893000, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MUST_BE_KDC, -2146892999, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_STRONG_CRYPTO_NOT_SUPPORTED, -2146892998, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_TOO_MANY_PRINCIPALS, -2146892997, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_NO_PA_DATA, -2146892996, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_PKINIT_NAME_MISMATCH, -2146892995, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_SMARTCARD_LOGON_REQUIRED, SEC_E_SMARTCARD_LOGON_REQUIRED, ERRSRV, ERRerror, -1, "SEC_E_SMARTCARD_LOGON_REQUIRED" )
S ( LW_STATUS_KDC_INVALID_REQUEST, -2146892992, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_KDC_UNABLE_TO_REFER, -2146892991, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_KDC_UNKNOWN_ETYPE, -2146892990, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_SHUTDOWN_IN_PROGRESS, ERROR_SHUTDOWN_IN_PROGRESS, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_SERVER_SHUTDOWN_IN_PROGRESS, ERROR_SERVER_SHUTDOWN_IN_PROGRESS, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_NOT_SUPPORTED_ON_SBS, ERROR_NOT_SUPPORTED_ON_SBS, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_WMI_GUID_DISCONNECTED, ERROR_WMI_GUID_DISCONNECTED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_WMI_ALREADY_DISABLED, ERROR_WMI_ALREADY_DISABLED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_WMI_ALREADY_ENABLED, ERROR_WMI_ALREADY_ENABLED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MFT_TOO_FRAGMENTED, ERROR_DISK_TOO_FRAGMENTED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_COPY_PROTECTION_FAILURE, -2147286267, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_CSS_AUTHENTICATION_FAILURE, -2147286266, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_CSS_KEY_NOT_PRESENT, -2147286265, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_CSS_KEY_NOT_ESTABLISHED, -2147286264, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_CSS_SCRAMBLED_SECTOR, -2147286263, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_CSS_REGION_MISMATCH, -2147286262, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_CSS_RESETS_EXHAUSTED, -2147286261, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_PKINIT_FAILURE, ERROR_PKINIT_FAILURE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_SMARTCARD_SUBSYSTEM_FAILURE, ERROR_SMARTCARD_SUBSYSTEM_FAILURE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_NO_KERB_KEY, -2146892984, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_HOST_DOWN, ERROR_HOST_DOWN, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_UNSUPPORTED_PREAUTH, -2146892989, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_EFS_ALG_BLOB_TOO_BIG, ERROR_EFS_ALG_BLOB_TOO_BIG, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_PORT_NOT_SET, ERROR_MR_MID_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_DEBUGGER_INACTIVE, ERROR_MR_MID_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_DS_VERSION_CHECK_FAILURE, ERROR_MR_MID_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_AUDITING_DISABLED, -1073151999, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_PRENT4_MACHINE_ACCOUNT, ERROR_DS_MACHINE_ACCOUNT_CREATED_PRENT4, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_DS_AG_CANT_HAVE_UNIVERSAL_MEMBER, ERROR_DS_AG_CANT_HAVE_UNIVERSAL_MEMBER, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_INVALID_IMAGE_WIN_32, ERROR_BAD_EXE_FORMAT, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_INVALID_IMAGE_WIN_64, ERROR_BAD_EXE_FORMAT, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_BAD_BINDINGS, -2146892986, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_NETWORK_SESSION_EXPIRED, ERROR_NO_USER_SESSION_KEY, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_APPHELP_BLOCK, ERROR_MR_MID_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_ALL_SIDS_FILTERED, ERROR_MR_MID_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_NOT_SAFE_MODE_DRIVER, ERROR_MR_MID_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_ACCESS_DISABLED_BY_POLICY_DEFAULT, ERROR_ACCESS_DISABLED_BY_POLICY, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_ACCESS_DISABLED_BY_POLICY_PATH, ERROR_ACCESS_DISABLED_BY_POLICY, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_ACCESS_DISABLED_BY_POLICY_PUBLISHER, ERROR_ACCESS_DISABLED_BY_POLICY, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_ACCESS_DISABLED_BY_POLICY_OTHER, ERROR_ACCESS_DISABLED_BY_POLICY, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_FAILED_DRIVER_ENTRY, ERROR_MR_MID_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_DEVICE_ENUMERATION_ERROR, ERROR_MR_MID_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MOUNT_POINT_NOT_RESOLVED, ERROR_MR_MID_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_INVALID_DEVICE_OBJECT_PARAMETER, ERROR_MR_MID_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MCA_OCCURED, ERROR_MR_MID_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_DRIVER_BLOCKED_CRITICAL, ERROR_DRIVER_BLOCKED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_DRIVER_BLOCKED, ERROR_DRIVER_BLOCKED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_DRIVER_DATABASE_ERROR, ERROR_MR_MID_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_SYSTEM_HIVE_TOO_LARGE, ERROR_MR_MID_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_INVALID_IMPORT_OF_NON_DLL, ERROR_INVALID_IMPORT_OF_NON_DLL, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_NO_SECRETS, ERROR_MR_MID_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_ACCESS_DISABLED_NO_SAFER_UI_BY_POLICY, ERROR_MR_MID_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_FAILED_STACK_SWITCH, ERROR_MR_MID_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_HEAP_CORRUPTION, ERROR_MR_MID_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_SMARTCARD_WRONG_PIN, -2146434965, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_SMARTCARD_CARD_BLOCKED, -2146434964, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_SMARTCARD_CARD_NOT_AUTHENTICATED, -2146434961, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_SMARTCARD_NO_CARD, SCARD_E_NO_SMARTCARD, ERRSRV, ERRerror, -1, "SCARD_E_NO_SMARTCARD" )
S ( LW_STATUS_SMARTCARD_NO_KEY_CONTAINER, -2146893811, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_SMARTCARD_NO_CERTIFICATE, SCARD_E_NO_SUCH_CERTIFICATE, ERRSRV, ERRerror, -1, "SCARD_E_NO_SUCH_CERTIFICATE" )
S ( LW_STATUS_SMARTCARD_NO_KEYSET, -2146893802, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_SMARTCARD_IO_ERROR, -2146435025, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_DOWNGRADE_DETECTED, ERROR_DOWNGRADE_DETECTED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_SMARTCARD_CERT_REVOKED, 1266, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_ISSUING_CA_UNTRUSTED, 1267, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_REVOCATION_OFFLINE_C, 1268, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_PKINIT_CLIENT_FAILURE, 1269, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_SMARTCARD_CERT_EXPIRED, 1270, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_DRIVER_FAILED_PRIOR_UNLOAD, ERROR_MR_MID_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_SMARTCARD_SILENT_CONTEXT, ERROR_MR_MID_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_PER_USER_TRUST_QUOTA_EXCEEDED, ERROR_MR_MID_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_ALL_USER_TRUST_QUOTA_EXCEEDED, ERROR_MR_MID_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_USER_DELETE_TRUST_QUOTA_EXCEEDED, ERROR_MR_MID_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_DS_NAME_NOT_UNIQUE, ERROR_MR_MID_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_DS_DUPLICATE_ID_FOUND, ERROR_MR_MID_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_DS_GROUP_CONVERSION_ERROR, ERROR_MR_MID_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_VOLSNAP_PREPARE_HIBERNATE, ERROR_MR_MID_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_USER2USER_REQUIRED, ERROR_MR_MID_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_STACK_BUFFER_OVERRUN, ERROR_MR_MID_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_NO_S4U_PROT_SUPPORT, ERROR_MR_MID_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_CROSSREALM_DELEGATION_FAILURE, ERROR_MR_MID_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_REVOCATION_OFFLINE_KDC, ERROR_MR_MID_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_ISSUING_CA_UNTRUSTED_KDC, ERROR_MR_MID_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_KDC_CERT_EXPIRED, ERROR_MR_MID_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_KDC_CERT_REVOKED, SEC_E_KDC_CERT_REVOKED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_PARAMETER_QUOTA_EXCEEDED, ERROR_MR_MID_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_HIBERNATION_FAILURE, ERROR_MR_MID_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_DELAY_LOAD_FAILED, ERROR_MR_MID_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_AUTHENTICATION_FIREWALL_FAILED, ERROR_AUTHENTICATION_FIREWALL_FAILED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_VDM_DISALLOWED, ERROR_MR_MID_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_HUNG_DISPLAY_DRIVER_THREAD, ERROR_MR_MID_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_INSUFFICIENT_RESOURCE_FOR_SPECIFIED_SHARED_SECTION_SIZE, ERROR_MR_MID_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_INVALID_CRUNTIME_PARAMETER, ERROR_MR_MID_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_NTLM_BLOCKED, ERROR_MR_MID_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_ASSERTION_FAILURE, ERROR_MR_MID_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_VERIFIER_STOP, ERROR_MR_MID_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_CALLBACK_POP_STACK, ERROR_MR_MID_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_INCOMPATIBLE_DRIVER_BLOCKED, ERROR_MR_MID_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_HIVE_UNLOADED, ERROR_MR_MID_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_COMPRESSION_DISABLED, ERROR_MR_MID_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_FILE_SYSTEM_LIMITATION, ERROR_MR_MID_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_INVALID_IMAGE_HASH, ERROR_MR_MID_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_NOT_CAPABLE, ERROR_MR_MID_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_REQUEST_OUT_OF_SEQUENCE, ERROR_MR_MID_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_IMPLEMENTATION_LIMIT, ERROR_MR_MID_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_ELEVATION_REQUIRED, ERROR_MR_MID_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_BEYOND_VDL, ERROR_MR_MID_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_ENCOUNTERED_WRITE_IN_PROGRESS, ERROR_MR_MID_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_PTE_CHANGED, ERROR_MR_MID_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_PURGE_FAILED, ERROR_MR_MID_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_CRED_REQUIRES_CONFIRMATION, ERROR_MR_MID_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_CS_ENCRYPTION_INVALID_SERVER_RESPONSE, ERROR_MR_MID_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_CS_ENCRYPTION_UNSUPPORTED_SERVER, ERROR_MR_MID_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_CS_ENCRYPTION_EXISTING_ENCRYPTED_FILE, ERROR_MR_MID_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_CS_ENCRYPTION_NEW_ENCRYPTED_FILE, ERROR_MR_MID_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_CS_ENCRYPTION_FILE_NOT_CSE, ERROR_MR_MID_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_INVALID_LABEL, ERROR_MR_MID_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_DRIVER_PROCESS_TERMINATED, ERROR_MR_MID_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_AMBIGUOUS_SYSTEM_DEVICE, ERROR_MR_MID_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_SYSTEM_DEVICE_NOT_FOUND, ERROR_MR_MID_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_RESTART_BOOT_APPLICATION, ERROR_MR_MID_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_INVALID_TASK_NAME, ERROR_MR_MID_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_INVALID_TASK_INDEX, ERROR_MR_MID_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_THREAD_ALREADY_IN_TASK, ERROR_MR_MID_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_CALLBACK_BYPASS, ERROR_MR_MID_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_PORT_CLOSED, ERROR_MR_MID_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_LOST, ERROR_MR_MID_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_INVALID_MESSAGE, ERROR_MR_MID_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_REQUEST_CANCELED, ERROR_MR_MID_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_RECURSIVE_DISPATCH, ERROR_MR_MID_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_LPC_RECEIVE_BUFFER_EXPECTED, ERROR_MR_MID_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_LPC_INVALID_CONNECTION_USAGE, ERROR_MR_MID_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_LPC_REQUESTS_NOT_ALLOWED, ERROR_MR_MID_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_RESOURCE_IN_USE, ERROR_MR_MID_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_HARDWARE_MEMORY_ERROR, ERROR_MR_MID_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_THREADPOOL_HANDLE_EXCEPTION, ERROR_MR_MID_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_THREADPOOL_SET_EVENT_ON_COMPLETION_FAILED, ERROR_MR_MID_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_THREADPOOL_RELEASE_SEMAPHORE_ON_COMPLETION_FAILED, ERROR_MR_MID_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_THREADPOOL_RELEASE_MUTEX_ON_COMPLETION_FAILED, ERROR_MR_MID_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_THREADPOOL_FREE_LIBRARY_ON_COMPLETION_FAILED, ERROR_MR_MID_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_THREADPOOL_RELEASED_DURING_OPERATION, ERROR_MR_MID_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_CALLBACK_RETURNED_WHILE_IMPERSONATING, ERROR_MR_MID_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_APC_RETURNED_WHILE_IMPERSONATING, ERROR_MR_MID_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_PROCESS_IS_PROTECTED, ERROR_MR_MID_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MCA_EXCEPTION, ERROR_MR_MID_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_CERTIFICATE_MAPPING_NOT_UNIQUE, ERROR_MR_MID_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_SYMLINK_CLASS_DISABLED, ERROR_MR_MID_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_INVALID_IDN_NORMALIZATION, ERROR_MR_MID_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_NO_UNICODE_TRANSLATION, ERROR_MR_MID_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_ALREADY_REGISTERED, ERROR_MR_MID_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_CONTEXT_MISMATCH, ERROR_MR_MID_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_PORT_ALREADY_HAS_COMPLETION_LIST, ERROR_MR_MID_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_CALLBACK_RETURNED_THREAD_PRIORITY, ERROR_MR_MID_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_INVALID_THREAD, ERROR_MR_MID_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_CALLBACK_RETURNED_TRANSACTION, ERROR_MR_MID_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_CALLBACK_RETURNED_LDR_LOCK, ERROR_MR_MID_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_CALLBACK_RETURNED_LANG, ERROR_MR_MID_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_CALLBACK_RETURNED_PRI_BACK, ERROR_MR_MID_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_CALLBACK_RETURNED_THREAD_AFFINITY, ERROR_MR_MID_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_OPLOCK_SWITCHED_TO_NEW_HANDLE, ERROR_OPLOCK_SWITCHED_TO_NEW_HANDLE, ERRSRV, ERRerror, -1, "")
S ( LW_DBG_NO_STATE_CHANGE, ERROR_INVALID_FUNCTION, ERRSRV, ERRerror, -1, "" )
S ( LW_DBG_APP_NOT_IDLE, ERROR_FILE_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_RPC_NT_INVALID_STRING_BINDING, RPC_S_INVALID_STRING_BINDING, ERRSRV, ERRerror, -1, "" )
S ( LW_RPC_NT_WRONG_KIND_OF_BINDING, RPC_S_WRONG_KIND_OF_BINDING, ERRSRV, ERRerror, -1, "" )
S ( LW_RPC_NT_INVALID_BINDING, ERROR_INVALID_HANDLE, ERRSRV, ERRerror, -1, "" )
S ( LW_RPC_NT_PROTSEQ_NOT_SUPPORTED, RPC_S_PROTSEQ_NOT_SUPPORTED, ERRSRV, ERRerror, -1, "" )
S ( LW_RPC_NT_INVALID_RPC_PROTSEQ, RPC_S_INVALID_RPC_PROTSEQ, ERRSRV, ERRerror, -1, "" )
S ( LW_RPC_NT_INVALID_STRING_UUID, RPC_S_INVALID_STRING_UUID, ERRSRV, ERRerror, -1, "" )
S ( LW_RPC_NT_INVALID_ENDPOINT_FORMAT, RPC_S_INVALID_ENDPOINT_FORMAT, ERRSRV, ERRerror, -1, "" )
S ( LW_RPC_NT_INVALID_NET_ADDR, RPC_S_INVALID_NET_ADDR, ERRSRV, ERRerror, -1, "" )
S ( LW_RPC_NT_NO_ENDPOINT_FOUND, RPC_S_NO_ENDPOINT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_RPC_NT_INVALID_TIMEOUT, RPC_S_INVALID_TIMEOUT, ERRSRV, ERRerror, -1, "" )
S ( LW_RPC_NT_OBJECT_NOT_FOUND, RPC_S_OBJECT_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_RPC_NT_ALREADY_REGISTERED, RPC_S_ALREADY_REGISTERED, ERRSRV, ERRerror, -1, "" )
S ( LW_RPC_NT_TYPE_ALREADY_REGISTERED, RPC_S_TYPE_ALREADY_REGISTERED, ERRSRV, ERRerror, -1, "" )
S ( LW_RPC_NT_ALREADY_LISTENING, RPC_S_ALREADY_LISTENING, ERRSRV, ERRerror, -1, "" )
S ( LW_RPC_NT_NO_PROTSEQS_REGISTERED, RPC_S_NO_PROTSEQS_REGISTERED, ERRSRV, ERRerror, -1, "" )
S ( LW_RPC_NT_NOT_LISTENING, RPC_S_NOT_LISTENING, ERRSRV, ERRerror, -1, "" )
S ( LW_RPC_NT_UNKNOWN_MGR_TYPE, RPC_S_UNKNOWN_MGR_TYPE, ERRSRV, ERRerror, -1, "" )
S ( LW_RPC_NT_UNKNOWN_IF, RPC_S_UNKNOWN_IF, ERRSRV, ERRerror, -1, "" )
S ( LW_RPC_NT_NO_BINDINGS, RPC_S_NO_BINDINGS, ERRSRV, ERRerror, -1, "" )
S ( LW_RPC_NT_NO_PROTSEQS, RPC_S_NO_PROTSEQS, ERRSRV, ERRerror, -1, "" )
S ( LW_RPC_NT_CANT_CREATE_ENDPOINT, RPC_S_CANT_CREATE_ENDPOINT, ERRSRV, ERRerror, -1, "" )
S ( LW_RPC_NT_OUT_OF_RESOURCES, RPC_S_OUT_OF_RESOURCES, ERRSRV, ERRerror, -1, "" )
S ( LW_RPC_NT_SERVER_UNAVAILABLE, RPC_S_SERVER_UNAVAILABLE, ERRSRV, ERRerror, -1, "" )
S ( LW_RPC_NT_SERVER_TOO_BUSY, RPC_S_SERVER_TOO_BUSY, ERRSRV, ERRerror, -1, "" )
S ( LW_RPC_NT_INVALID_NETWORK_OPTIONS, RPC_S_INVALID_NETWORK_OPTIONS, ERRSRV, ERRerror, -1, "" )
S ( LW_RPC_NT_NO_CALL_ACTIVE, RPC_S_NO_CALL_ACTIVE, ERRSRV, ERRerror, -1, "" )
S ( LW_RPC_NT_CALL_FAILED, RPC_S_CALL_FAILED, ERRSRV, ERRerror, -1, "" )
S ( LW_RPC_NT_CALL_FAILED_DNE, RPC_S_CALL_FAILED_DNE, ERRSRV, ERRerror, -1, "" )
S ( LW_RPC_NT_PROTOCOL_ERROR, RPC_S_PROTOCOL_ERROR, ERRSRV, ERRerror, -1, "" )
S ( LW_RPC_NT_UNSUPPORTED_TRANS_SYN, RPC_S_UNSUPPORTED_TRANS_SYN, ERRSRV, ERRerror, -1, "" )
S ( LW_RPC_NT_UNSUPPORTED_TYPE, RPC_S_UNSUPPORTED_TYPE, ERRSRV, ERRerror, -1, "" )
S ( LW_RPC_NT_INVALID_TAG, RPC_S_INVALID_TAG, ERRSRV, ERRerror, -1, "" )
S ( LW_RPC_NT_INVALID_BOUND, RPC_S_INVALID_BOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_RPC_NT_NO_ENTRY_NAME, RPC_S_NO_ENTRY_NAME, ERRSRV, ERRerror, -1, "" )
S ( LW_RPC_NT_INVALID_NAME_SYNTAX, RPC_S_INVALID_NAME_SYNTAX, ERRSRV, ERRerror, -1, "" )
S ( LW_RPC_NT_UNSUPPORTED_NAME_SYNTAX, RPC_S_UNSUPPORTED_NAME_SYNTAX, ERRSRV, ERRerror, -1, "" )
S ( LW_RPC_NT_UUID_NO_ADDRESS, RPC_S_UUID_NO_ADDRESS, ERRSRV, ERRerror, -1, "" )
S ( LW_RPC_NT_DUPLICATE_ENDPOINT, RPC_S_DUPLICATE_ENDPOINT, ERRSRV, ERRerror, -1, "" )
S ( LW_RPC_NT_UNKNOWN_AUTHN_TYPE, RPC_S_UNKNOWN_AUTHN_TYPE, ERRSRV, ERRerror, -1, "" )
S ( LW_RPC_NT_MAX_CALLS_TOO_SMALL, RPC_S_MAX_CALLS_TOO_SMALL, ERRSRV, ERRerror, -1, "" )
S ( LW_RPC_NT_STRING_TOO_LONG, RPC_S_STRING_TOO_LONG, ERRSRV, ERRerror, -1, "" )
S ( LW_RPC_NT_PROTSEQ_NOT_FOUND, RPC_S_PROTSEQ_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_RPC_NT_PROCNUM_OUT_OF_RANGE, RPC_S_PROCNUM_OUT_OF_RANGE, ERRSRV, ERRerror, -1, "" )
S ( LW_RPC_NT_BINDING_HAS_NO_AUTH, RPC_S_BINDING_HAS_NO_AUTH, ERRSRV, ERRerror, -1, "" )
S ( LW_RPC_NT_UNKNOWN_AUTHN_SERVICE, RPC_S_UNKNOWN_AUTHN_SERVICE, ERRSRV, ERRerror, -1, "" )
S ( LW_RPC_NT_UNKNOWN_AUTHN_LEVEL, RPC_S_UNKNOWN_AUTHN_LEVEL, ERRSRV, ERRerror, -1, "" )
S ( LW_RPC_NT_INVALID_AUTH_IDENTITY, RPC_S_INVALID_AUTH_IDENTITY, ERRSRV, ERRerror, -1, "" )
S ( LW_RPC_NT_UNKNOWN_AUTHZ_SERVICE, RPC_S_UNKNOWN_AUTHZ_SERVICE, ERRSRV, ERRerror, -1, "" )
S ( LW_EPT_NT_INVALID_ENTRY, EPT_S_INVALID_ENTRY, ERRSRV, ERRerror, -1, "" )
S ( LW_EPT_NT_CANT_PERFORM_OP, EPT_S_CANT_PERFORM_OP, ERRSRV, ERRerror, -1, "" )
S ( LW_EPT_NT_NOT_REGISTERED, EPT_S_NOT_REGISTERED, ERRSRV, ERRerror, -1, "" )
S ( LW_RPC_NT_NOTHING_TO_EXPORT, RPC_S_NOTHING_TO_EXPORT, ERRSRV, ERRerror, -1, "" )
S ( LW_RPC_NT_INCOMPLETE_NAME, RPC_S_INCOMPLETE_NAME, ERRSRV, ERRerror, -1, "" )
S ( LW_RPC_NT_INVALID_VERS_OPTION, RPC_S_INVALID_VERS_OPTION, ERRSRV, ERRerror, -1, "" )
S ( LW_RPC_NT_NO_MORE_MEMBERS, RPC_S_NO_MORE_MEMBERS, ERRSRV, ERRerror, -1, "" )
S ( LW_RPC_NT_NOT_ALL_OBJS_UNEXPORTED, RPC_S_NOT_ALL_OBJS_UNEXPORTED, ERRSRV, ERRerror, -1, "" )
S ( LW_RPC_NT_INTERFACE_NOT_FOUND, RPC_S_INTERFACE_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_RPC_NT_ENTRY_ALREADY_EXISTS, RPC_S_ENTRY_ALREADY_EXISTS, ERRSRV, ERRerror, -1, "" )
S ( LW_RPC_NT_ENTRY_NOT_FOUND, RPC_S_ENTRY_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_RPC_NT_NAME_SERVICE_UNAVAILABLE, RPC_S_NAME_SERVICE_UNAVAILABLE, ERRSRV, ERRerror, -1, "" )
S ( LW_RPC_NT_INVALID_NAF_ID, RPC_S_INVALID_NAF_ID, ERRSRV, ERRerror, -1, "" )
S ( LW_RPC_NT_CANNOT_SUPPORT, RPC_S_CANNOT_SUPPORT, ERRSRV, ERRerror, -1, "" )
S ( LW_RPC_NT_NO_CONTEXT_AVAILABLE, RPC_S_NO_CONTEXT_AVAILABLE, ERRSRV, ERRerror, -1, "" )
S ( LW_RPC_NT_INTERNAL_ERROR, RPC_S_INTERNAL_ERROR, ERRSRV, ERRerror, -1, "" )
S ( LW_RPC_NT_ZERO_DIVIDE, RPC_S_ZERO_DIVIDE, ERRSRV, ERRerror, -1, "" )
S ( LW_RPC_NT_ADDRESS_ERROR, RPC_S_ADDRESS_ERROR, ERRSRV, ERRerror, -1, "" )
S ( LW_RPC_NT_FP_DIV_ZERO, RPC_S_FP_DIV_ZERO, ERRSRV, ERRerror, -1, "" )
S ( LW_RPC_NT_FP_UNDERFLOW, RPC_S_FP_UNDERFLOW, ERRSRV, ERRerror, -1, "" )
S ( LW_RPC_NT_FP_OVERFLOW, RPC_S_FP_OVERFLOW, ERRSRV, ERRerror, -1, "" )
S ( LW_RPC_NT_CALL_IN_PROGRESS, RPC_S_CALL_IN_PROGRESS, ERRSRV, ERRerror, -1, "" )
S ( LW_RPC_NT_NO_MORE_BINDINGS, RPC_S_NO_MORE_BINDINGS, ERRSRV, ERRerror, -1, "" )
S ( LW_RPC_NT_GROUP_MEMBER_NOT_FOUND, RPC_S_GROUP_MEMBER_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_EPT_NT_CANT_CREATE, EPT_S_CANT_CREATE, ERRSRV, ERRerror, -1, "" )
S ( LW_RPC_NT_INVALID_OBJECT, RPC_S_INVALID_OBJECT, ERRSRV, ERRerror, -1, "" )
S ( LW_RPC_NT_NO_INTERFACES, RPC_S_NO_INTERFACES, ERRSRV, ERRerror, -1, "" )
S ( LW_RPC_NT_CALL_CANCELLED, RPC_S_CALL_CANCELLED, ERRSRV, ERRerror, -1, "" )
S ( LW_RPC_NT_BINDING_INCOMPLETE, RPC_S_BINDING_INCOMPLETE, ERRSRV, ERRerror, -1, "" )
S ( LW_RPC_NT_COMM_FAILURE, RPC_S_COMM_FAILURE, ERRSRV, ERRerror, -1, "" )
S ( LW_RPC_NT_UNSUPPORTED_AUTHN_LEVEL, RPC_S_UNSUPPORTED_AUTHN_LEVEL, ERRSRV, ERRerror, -1, "" )
S ( LW_RPC_NT_NO_PRINC_NAME, RPC_S_NO_PRINC_NAME, ERRSRV, ERRerror, -1, "" )
S ( LW_RPC_NT_NOT_RPC_ERROR, RPC_S_NOT_RPC_ERROR, ERRSRV, ERRerror, -1, "" )
S ( LW_RPC_NT_SEC_PKG_ERROR, RPC_S_SEC_PKG_ERROR, ERRSRV, ERRerror, -1, "" )
S ( LW_RPC_NT_NOT_CANCELLED, RPC_S_NOT_CANCELLED, ERRSRV, ERRerror, -1, "" )
S ( LW_RPC_NT_INVALID_ASYNC_HANDLE, RPC_S_INVALID_ASYNC_HANDLE, ERRSRV, ERRerror, -1, "" )
S ( LW_RPC_NT_INVALID_ASYNC_CALL, RPC_S_INVALID_ASYNC_CALL, ERRSRV, ERRerror, -1, "" )
S ( LW_RPC_NT_PROXY_ACCESS_DENIED, ERROR_MR_MID_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_RPC_NT_NO_MORE_ENTRIES, RPC_X_NO_MORE_ENTRIES, ERRSRV, ERRerror, -1, "" )
S ( LW_RPC_NT_SS_CHAR_TRANS_OPEN_FAIL, RPC_X_SS_CHAR_TRANS_OPEN_FAIL, ERRSRV, ERRerror, -1, "" )
S ( LW_RPC_NT_SS_CHAR_TRANS_SHORT_FILE, RPC_X_SS_CHAR_TRANS_SHORT_FILE, ERRSRV, ERRerror, -1, "" )
S ( LW_RPC_NT_SS_IN_NULL_CONTEXT, ERROR_INVALID_HANDLE, ERRSRV, ERRerror, -1, "" )
S ( LW_RPC_NT_SS_CONTEXT_MISMATCH, ERROR_INVALID_HANDLE, ERRSRV, ERRerror, -1, "" )
S ( LW_RPC_NT_SS_CONTEXT_DAMAGED, RPC_X_SS_CONTEXT_DAMAGED, ERRSRV, ERRerror, -1, "" )
S ( LW_RPC_NT_SS_HANDLES_MISMATCH, RPC_X_SS_HANDLES_MISMATCH, ERRSRV, ERRerror, -1, "" )
S ( LW_RPC_NT_SS_CANNOT_GET_CALL_HANDLE, RPC_X_SS_CANNOT_GET_CALL_HANDLE, ERRSRV, ERRerror, -1, "" )
S ( LW_RPC_NT_NULL_REF_POINTER, RPC_X_NULL_REF_POINTER, ERRSRV, ERRerror, -1, "" )
S ( LW_RPC_NT_ENUM_VALUE_OUT_OF_RANGE, RPC_X_ENUM_VALUE_OUT_OF_RANGE, ERRSRV, ERRerror, -1, "" )
S ( LW_RPC_NT_BYTE_COUNT_TOO_SMALL, RPC_X_BYTE_COUNT_TOO_SMALL, ERRSRV, ERRerror, -1, "" )
S ( LW_RPC_NT_BAD_STUB_DATA, RPC_X_BAD_STUB_DATA, ERRSRV, ERRerror, -1, "" )
S ( LW_RPC_NT_INVALID_ES_ACTION, RPC_X_INVALID_ES_ACTION, ERRSRV, ERRerror, -1, "" )
S ( LW_RPC_NT_WRONG_ES_VERSION, RPC_X_WRONG_ES_VERSION, ERRSRV, ERRerror, -1, "" )
S ( LW_RPC_NT_WRONG_STUB_VERSION, RPC_X_WRONG_STUB_VERSION, ERRSRV, ERRerror, -1, "" )
S ( LW_RPC_NT_INVALID_PIPE_OBJECT, RPC_X_INVALID_PIPE_OBJECT, ERRSRV, ERRerror, -1, "" )
S ( LW_RPC_NT_INVALID_PIPE_OPERATION, RPC_X_WRONG_PIPE_ORDER, ERRSRV, ERRerror, -1, "" )
S ( LW_RPC_NT_WRONG_PIPE_VERSION, RPC_X_WRONG_PIPE_VERSION, ERRSRV, ERRerror, -1, "" )
S ( LW_RPC_NT_PIPE_CLOSED, RPC_X_PIPE_CLOSED, ERRSRV, ERRerror, -1, "" )
S ( LW_RPC_NT_PIPE_DISCIPLINE_ERROR, RPC_X_PIPE_DISCIPLINE_ERROR, ERRSRV, ERRerror, -1, "" )
S ( LW_RPC_NT_PIPE_EMPTY, RPC_X_PIPE_EMPTY, ERRSRV, ERRerror, -1, "" )

/* Win32 error codes without an equivalent NTStatus */
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_ARENA_TRASHED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_BAD_ENVIRONMENT, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_BAD_FORMAT, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_INVALID_ACCESS, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_INVALID_DATA, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_INVALID_DRIVE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_CURRENT_DIRECTORY, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_BAD_UNIT, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_SEEK, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_NOT_DOS_DISK, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_WRITE_FAULT, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_READ_FAULT, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_SHARING_BUFFER_EXCEEDED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_HANDLE_DISK_FULL, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_FILE_EXISTS, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_FAIL_I24, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_OUT_OF_STRUCTURES, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_ALREADY_ASSIGNED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_NO_PROC_SLOTS, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_TOO_MANY_SEMAPHORES, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_EXCL_SEM_ALREADY_OWNED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_SEM_IS_SET, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_TOO_MANY_SEM_REQUESTS, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_INVALID_AT_INTERRUPT_TIME, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_SEM_OWNER_DIED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_SEM_USER_LIMIT, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DISK_CHANGE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DRIVE_LOCKED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_OPEN_FAILED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_BUFFER_OVERFLOW, ERRDOS, ERRmoredata, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_NO_MORE_SEARCH_HANDLES, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_INVALID_TARGET_HANDLE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_INVALID_CATEGORY, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_INVALID_VERIFY_SWITCH, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_BAD_DRIVER_LEVEL, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_CALL_NOT_IMPLEMENTED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_NO_VOLUME_LABEL, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_WAIT_NO_CHILDREN, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_CHILD_NOT_COMPLETE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DIRECT_ACCESS_HANDLE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_NEGATIVE_SEEK, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_SEEK_ON_DEVICE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_IS_JOIN_TARGET, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_IS_JOINED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_IS_SUBSTED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_NOT_JOINED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_NOT_SUBSTED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_JOIN_TO_JOIN, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_SUBST_TO_SUBST, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_JOIN_TO_SUBST, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_SUBST_TO_JOIN, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_BUSY_DRIVE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_SAME_DRIVE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DIR_NOT_ROOT, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_IS_SUBST_PATH, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_IS_JOIN_PATH, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_PATH_BUSY, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_IS_SUBST_TARGET, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_SYSTEM_TRACE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_INVALID_EVENT_COUNT, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_TOO_MANY_MUXWAITERS, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_INVALID_LIST_FORMAT, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_TOO_MANY_TCBS, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DISCARDED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_BAD_THREADID_ADDR, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_BAD_ARGUMENTS, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_SIGNAL_PENDING, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_MAX_THRDS_REACHED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_LOCK_FAILED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_CANCEL_VIOLATION, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_ATOMIC_LOCKS_NOT_SUPPORTED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_INVALID_SEGMENT_NUMBER, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_INVALID_FLAG_NUMBER, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_SEM_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_INVALID_STARTING_CODESEG, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_INVALID_STACKSEG, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_INVALID_MODULETYPE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_INVALID_EXE_SIGNATURE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_EXE_MARKED_INVALID, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_ITERATED_DATA_EXCEEDS_64k, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_INVALID_MINALLOCSIZE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DYNLINK_FROM_INVALID_RING, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_IOPL_NOT_ENABLED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_INVALID_SEGDPL, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_AUTODATASEG_EXCEEDS_64k, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_RING2SEG_MUST_BE_MOVABLE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_RELOC_CHAIN_XEEDS_SEGLIM, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_INFLOOP_IN_RELOC_CHAIN, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_NO_SIGNAL_SENT, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_RING2_STACK_IN_USE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_META_EXPANSION_TOO_LONG, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_INVALID_SIGNAL_NUMBER, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_THREAD_1_INACTIVE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_LOCKED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_TOO_MANY_MODULES, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_NESTING_NOT_ALLOWED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_EXE_MACHINE_TYPE_MISMATCH, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_EXE_CANNOT_MODIFY_SIGNED_BINARY, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_EXE_CANNOT_MODIFY_STRONG_SIGNED_BINARY, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_FILE_CHECKED_OUT, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_CHECKOUT_REQUIRED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_BAD_FILE_TYPE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_FILE_TOO_LARGE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_FORMS_AUTH_REQUIRED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_VIRUS_INFECTED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_VIRUS_DELETED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_PIPE_LOCAL, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, WAIT_TIMEOUT, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_CANNOT_COPY, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_EAS_DIDNT_FIT, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_EA_FILE_CORRUPT, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_EA_TABLE_FULL, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_INVALID_EA_HANDLE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DELETE_PENDING, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_SCOPE_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_FAIL_NOACTION_REBOOT, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_FAIL_SHUTDOWN, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_FAIL_RESTART, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_MAX_SESSIONS_REACHED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_THREAD_MODE_ALREADY_BACKGROUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_THREAD_MODE_NOT_BACKGROUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_PROCESS_MODE_ALREADY_BACKGROUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_PROCESS_MODE_NOT_BACKGROUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_USER_PROFILE_LOAD, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_VERIFIER_STOP, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_ABIOS_ERROR, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_WX86_WARNING, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_WX86_ERROR, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_TIMER_NOT_CANCELED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_UNWIND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_BAD_STACK, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_INVALID_UNWIND_TARGET, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_INVALID_PORT_ATTRIBUTES, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_PORT_MESSAGE_TOO_LONG, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_INVALID_QUOTA_LOWER, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DEVICE_ALREADY_ATTACHED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_INSTRUCTION_MISALIGNMENT, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_PROFILING_NOT_STARTED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_PROFILING_NOT_STOPPED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_COULD_NOT_INTERPRET, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_PROFILING_AT_LIMIT, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_CANT_WAIT, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_CANT_TERMINATE_SELF, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_UNEXPECTED_MM_CREATE_ERR, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_UNEXPECTED_MM_MAP_ERROR, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_UNEXPECTED_MM_EXTEND_ERR, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_BAD_FUNCTION_TABLE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_NO_GUID_TRANSLATION, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_INVALID_LDT_SIZE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_INVALID_LDT_OFFSET, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_INVALID_LDT_DESCRIPTOR, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_TOO_MANY_THREADS, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_THREAD_NOT_IN_PROCESS, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_PAGEFILE_QUOTA_EXCEEDED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_LOGON_SERVER_CONFLICT, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_SYNCHRONIZATION_REQUIRED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_NET_OPEN_FAILED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_IO_PRIVILEGE_FAILED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_CONTROL_C_EXIT, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_MISSING_SYSTEMFILE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_UNHANDLED_EXCEPTION, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_APP_INIT_FAILURE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_PAGEFILE_CREATE_FAILED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_INVALID_IMAGE_HASH, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_NO_PAGEFILE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_ILLEGAL_FLOAT_CONTEXT, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_NO_EVENT_PAIR, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DOMAIN_CTRLR_CONFIG_ERROR, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_ILLEGAL_CHARACTER, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_UNDEFINED_CHARACTER, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_FLOPPY_VOLUME, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_BIOS_FAILED_TO_CONNECT_INTERRUPT, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_BACKUP_CONTROLLER, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_MUTANT_LIMIT_EXCEEDED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_FS_DRIVER_REQUIRED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_CANNOT_LOAD_REGISTRY_FILE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DEBUG_ATTACH_FAILED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_SYSTEM_PROCESS_TERMINATED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DATA_NOT_ACCEPTED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_VDM_HARD_ERROR, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DRIVER_CANCEL_TIMEOUT, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_REPLY_MESSAGE_MISMATCH, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_LOST_WRITEBEHIND_DATA, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_CLIENT_SERVER_PARAMETERS_INVALID, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_NOT_TINY_STREAM, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_STACK_OVERFLOW_READ, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_CONVERT_TO_LARGE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_FOUND_OUT_OF_SCOPE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_ALLOCATE_BUCKET, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_MARSHALL_OVERFLOW, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_INVALID_VARIANT, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_BAD_COMPRESSION_BUFFER, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_AUDIT_FAILED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_TIMER_RESOLUTION_NOT_SET, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_INSUFFICIENT_LOGON_INFO, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_BAD_DLL_ENTRYPOINT, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_BAD_SERVICE_ENTRYPOINT, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_IP_ADDRESS_CONFLICT1, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_IP_ADDRESS_CONFLICT2, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_REGISTRY_QUOTA_LIMIT, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_NO_CALLBACK_ACTIVE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_PWD_TOO_SHORT, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_PWD_TOO_RECENT, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_PWD_HISTORY_CONFLICT, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_UNSUPPORTED_COMPRESSION, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_INVALID_HW_PROFILE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_INVALID_PLUGPLAY_DEVICE_PATH, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_QUOTA_LIST_INCONSISTENT, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_EVALUATION_EXPIRATION, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_ILLEGAL_DLL_RELOCATION, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DLL_INIT_FAILED_LOGOFF, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_VALIDATE_CONTINUE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_NO_MORE_MATCHES, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_RANGE_LIST_CONFLICT, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_SERVER_SID_MISMATCH, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_CANT_ENABLE_DENY_ONLY, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_FLOAT_MULTIPLE_FAULTS, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_FLOAT_MULTIPLE_TRAPS, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_NOINTERFACE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DRIVER_FAILED_SLEEP, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_CORRUPT_SYSTEM_FILE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_COMMITMENT_MINIMUM, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_PNP_RESTART_ENUMERATION, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_SYSTEM_IMAGE_BAD_SIGNATURE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_PNP_REBOOT_REQUIRED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_INSUFFICIENT_POWER, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_MULTIPLE_FAULT_VIOLATION, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_SYSTEM_SHUTDOWN, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_PORT_NOT_SET, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DS_VERSION_CHECK_FAILURE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_RANGE_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_NOT_SAFE_MODE_DRIVER, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_FAILED_DRIVER_ENTRY, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DEVICE_ENUMERATION_ERROR, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_MOUNT_POINT_NOT_RESOLVED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_INVALID_DEVICE_OBJECT_PARAMETER, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_MCA_OCCURED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DRIVER_DATABASE_ERROR, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_SYSTEM_HIVE_TOO_LARGE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DRIVER_FAILED_PRIOR_UNLOAD, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_VOLSNAP_PREPARE_HIBERNATE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_HIBERNATION_FAILURE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_FILE_SYSTEM_LIMITATION, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_ASSERTION_FAILURE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_ACPI_ERROR, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_WOW_ASSERTION, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_PNP_BAD_MPS_TABLE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_PNP_TRANSLATION_FAILED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_PNP_IRQ_TRANSLATION_FAILED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_PNP_INVALID_ID, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_WAKE_SYSTEM_DEBUGGER, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_HANDLES_CLOSED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_EXTRANEOUS_INFORMATION, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_RXACT_COMMIT_NECESSARY, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_MEDIA_CHECK, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_GUID_SUBSTITUTION_MADE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_STOPPED_ON_SYMLINK, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_LONGJUMP, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_PLUGPLAY_QUERY_VETOED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_UNWIND_CONSOLIDATE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_REGISTRY_HIVE_RECOVERED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DLL_MIGHT_BE_INSECURE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DLL_MIGHT_BE_INCOMPATIBLE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DBG_EXCEPTION_NOT_HANDLED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DBG_REPLY_LATER, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DBG_UNABLE_TO_PROVIDE_HANDLE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DBG_TERMINATE_THREAD, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DBG_TERMINATE_PROCESS, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DBG_CONTROL_C, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DBG_PRINTEXCEPTION_C, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DBG_RIPEXCEPTION, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DBG_CONTROL_BREAK, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DBG_COMMAND_EXCEPTION, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_OBJECT_NAME_EXISTS, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_THREAD_WAS_SUSPENDED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_IMAGE_NOT_AT_BASE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_RXACT_STATE_CREATED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_SEGMENT_NOTIFICATION, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_BAD_CURRENT_DIRECTORY, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_FT_READ_RECOVERY_FROM_BACKUP, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_FT_WRITE_RECOVERY, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_IMAGE_MACHINE_TYPE_MISMATCH, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_RECEIVE_PARTIAL, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_RECEIVE_EXPEDITED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_RECEIVE_PARTIAL_EXPEDITED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_EVENT_DONE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_EVENT_PENDING, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_CHECKING_FILE_SYSTEM, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_FATAL_APP_EXIT, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_PREDEFINED_HANDLE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_WAS_UNLOCKED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_SERVICE_NOTIFICATION, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_WAS_LOCKED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_LOG_HARD_ERROR, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_ALREADY_WIN32, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_IMAGE_MACHINE_TYPE_MISMATCH_EXE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_NO_YIELD_PERFORMED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_TIMER_RESUME_IGNORED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_ARBITRATION_UNHANDLED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_CARDBUS_NOT_SUPPORTED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_MP_PROCESSOR_MISMATCH, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_HIBERNATED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_RESUME_HIBERNATION, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_FIRMWARE_UPDATED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DRIVERS_LEAKING_LOCKED_PAGES, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_WAKE_SYSTEM, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_WAIT_1, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_WAIT_2, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_WAIT_3, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_WAIT_63, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_ABANDONED_WAIT_0, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_ABANDONED_WAIT_63, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_USER_APC, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_KERNEL_APC, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_ALERTED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_ELEVATION_REQUIRED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_VOLUME_MOUNTED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_RXACT_COMMITTED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_NOTIFY_CLEANUP, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_PRIMARY_TRANSPORT_CONNECT_FAILED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_PAGE_FAULT_TRANSITION, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_PAGE_FAULT_DEMAND_ZERO, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_PAGE_FAULT_COPY_ON_WRITE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_PAGE_FAULT_GUARD_PAGE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_PAGE_FAULT_PAGING_FILE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_CACHE_PAGE_LOCKED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_CRASH_DUMP, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_BUFFER_ALL_ZEROS, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_REPARSE_OBJECT, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_RESOURCE_REQUIREMENTS_CHANGED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_TRANSLATION_COMPLETE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_NOTHING_TO_TERMINATE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_PROCESS_NOT_IN_JOB, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_PROCESS_IN_JOB, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_VOLSNAP_HIBERNATE_READY, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_FSFILTER_OP_COMPLETED_SUCCESSFULLY, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_INTERRUPT_VECTOR_ALREADY_CONNECTED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_INTERRUPT_STILL_CONNECTED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DBG_EXCEPTION_HANDLED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DBG_CONTINUE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_CALLBACK_POP_STACK, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_COMPRESSION_DISABLED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_CANTFETCHBACKWARDS, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_CANTSCROLLBACKWARDS, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_ROWSNOTRELEASED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_BAD_ACCESSOR_FLAGS, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_ERRORS_ENCOUNTERED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_NOT_CAPABLE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_REQUEST_OUT_OF_SEQUENCE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_VERSION_PARSE_ERROR, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_BADSTARTPOSITION, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_MEMORY_HARDWARE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DISK_REPAIR_DISABLED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_INSUFFICIENT_RESOURCE_FOR_SPECIFIED_SHARED_SECTION_SIZE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_SYSTEM_POWERSTATE_TRANSITION, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_SYSTEM_POWERSTATE_COMPLEX_TRANSITION, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_MCA_EXCEPTION, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_ACCESS_AUDIT_BY_POLICY, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_ACCESS_DISABLED_NO_SAFER_UI_BY_POLICY, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_ABANDON_HIBERFILE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_LOST_WRITEBEHIND_DATA_NETWORK_DISCONNECTED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_LOST_WRITEBEHIND_DATA_NETWORK_SERVER_ERROR, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_LOST_WRITEBEHIND_DATA_LOCAL_DISK_ERROR, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_EA_ACCESS_DENIED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_IO_INCOMPLETE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_INVALID_MESSAGE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_CAN_NOT_COMPLETE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_INVALID_FLAGS, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_BADKEY, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_CANTOPEN, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_CANTREAD, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_CANTWRITE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_REGISTRY_CORRUPT, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DEPENDENT_SERVICES_RUNNING, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_INVALID_SERVICE_CONTROL, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_SERVICE_REQUEST_TIMEOUT, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_SERVICE_NO_THREAD, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_SERVICE_DATABASE_LOCKED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_INVALID_SERVICE_ACCOUNT, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_CIRCULAR_DEPENDENCY, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_SERVICE_DOES_NOT_EXIST, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_SERVICE_CANNOT_ACCEPT_CTRL, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_SERVICE_NOT_ACTIVE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_FAILED_SERVICE_CONTROLLER_CONNECT, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_EXCEPTION_IN_SERVICE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DATABASE_DOES_NOT_EXIST, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_SERVICE_SPECIFIC_ERROR, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_PROCESS_ABORTED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_SERVICE_DEPENDENCY_FAIL, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_SERVICE_LOGON_FAILED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_SERVICE_START_HANG, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_INVALID_SERVICE_LOCK, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_SERVICE_MARKED_FOR_DELETE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_SERVICE_EXISTS, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_ALREADY_RUNNING_LKG, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_SERVICE_DEPENDENCY_DELETED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_BOOT_ALREADY_ACCEPTED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_SERVICE_NEVER_STARTED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DUPLICATE_SERVICE_NAME, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DIFFERENT_SERVICE_ACCOUNT, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_CANNOT_DETECT_DRIVER_FAILURE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_CANNOT_DETECT_PROCESS_ABORT, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_NO_RECOVERY_PROGRAM, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_SERVICE_NOT_IN_EXE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_NOT_SAFEBOOT_SERVICE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_NO_SHUTDOWN_IN_PROGRESS, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_SET_POWER_STATE_VETOED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_SET_POWER_STATE_FAILED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_OLD_WIN_VERSION, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_APP_WRONG_OS, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_SINGLE_INSTANCE_APP, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_RMODE_APP, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_INVALID_DLL, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_NO_ASSOCIATION, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DDE_FAIL, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DLL_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_NO_MORE_USER_HANDLES, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_MESSAGE_SYNC_ONLY, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_POINT_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_NO_VOLUME_ID, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_UNABLE_TO_REMOVE_REPLACED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_UNABLE_TO_MOVE_REPLACEMENT, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_UNABLE_TO_MOVE_REPLACEMENT_2, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_POTENTIAL_FILE_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_SHUTDOWN_IS_SCHEDULED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_SHUTDOWN_USERS_LOGGED_ON, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_BAD_DEVICE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DEVICE_ALREADY_REMEMBERED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_NO_NET_OR_BAD_PATH, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_BAD_PROVIDER, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_CANNOT_OPEN_PROFILE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_BAD_PROFILE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_NOT_CONTAINER, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_EXTENDED_ERROR, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_INVALID_GROUPNAME, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_INVALID_EVENTNAME, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_INVALID_DOMAINNAME, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_INVALID_SERVICENAME, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_INVALID_SHARENAME, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_INVALID_PASSWORDNAME, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_INVALID_MESSAGENAME, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_INVALID_MESSAGEDEST, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DUP_DOMAINNAME, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_NO_NETWORK, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_CANCELLED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_INCORRECT_ADDRESS, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_ALREADY_REGISTERED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_SERVICE_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_NOT_AUTHENTICATED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_NOT_LOGGED_ON, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_CONTINUE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_ALREADY_INITIALIZED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_NO_MORE_DEVICES, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_NO_SUCH_SITE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DOMAIN_CONTROLLER_EXISTS, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_OVERRIDE_NOCHANGES, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_BAD_USER_PROFILE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_NON_ACCOUNT_SID, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_NON_DOMAIN_SID, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_APPHELP_BLOCK, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_CSCSHARE_OFFLINE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_MACHINE_LOCKED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_CALLBACK_SUPPLIED_INVALID_DATA, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_SYNC_FOREGROUND_REFRESH_REQUIRED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_ACCESS_DISABLED_WEBBLADE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_ACCESS_DISABLED_WEBBLADE_TAMPER, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_RECOVERY_FAILURE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_ALREADY_FIBER, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_ALREADY_THREAD, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_STACK_BUFFER_OVERRUN, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_PARAMETER_QUOTA_EXCEEDED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DEBUGGER_INACTIVE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DELAY_LOAD_FAILED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_VDM_DISALLOWED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_UNIDENTIFIED_ERROR, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_INVALID_CRUNTIME_PARAMETER, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_BEYOND_VDL, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_INCOMPATIBLE_SERVICE_SID_TYPE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DRIVER_PROCESS_TERMINATED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_IMPLEMENTATION_LIMIT, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_PROCESS_IS_PROTECTED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_SERVICE_NOTIFY_CLIENT_LAGGING, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DISK_QUOTA_EXCEEDED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_CONTENT_BLOCKED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_INCOMPATIBLE_SERVICE_PRIVILEGE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_INVALID_LABEL, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_WRONG_PASSWORD, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_WRONG_TARGET_NAME, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_INVALID_WINDOW_HANDLE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_INVALID_MENU_HANDLE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_INVALID_CURSOR_HANDLE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_INVALID_ACCEL_HANDLE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_INVALID_HOOK_HANDLE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_INVALID_DWP_HANDLE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_TLW_WITH_WSCHILD, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_CANNOT_FIND_WND_CLASS, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_WINDOW_OF_OTHER_THREAD, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_HOTKEY_ALREADY_REGISTERED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_CLASS_ALREADY_EXISTS, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_CLASS_DOES_NOT_EXIST, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_CLASS_HAS_WINDOWS, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_INVALID_INDEX, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_INVALID_ICON_HANDLE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_PRIVATE_DIALOG_INDEX, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_LISTBOX_ID_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_NO_WILDCARD_CHARACTERS, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_CLIPBOARD_NOT_OPEN, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_HOTKEY_NOT_REGISTERED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_WINDOW_NOT_DIALOG, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_CONTROL_ID_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_INVALID_COMBOBOX_MESSAGE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_WINDOW_NOT_COMBOBOX, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_INVALID_EDIT_HEIGHT, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DC_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_INVALID_HOOK_FILTER, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_INVALID_FILTER_PROC, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_HOOK_NEEDS_HMOD, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_GLOBAL_ONLY_HOOK, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_JOURNAL_HOOK_SET, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_HOOK_NOT_INSTALLED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_INVALID_LB_MESSAGE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_SETCOUNT_ON_BAD_LB, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_LB_WITHOUT_TABSTOPS, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DESTROY_OBJECT_OF_OTHER_THREAD, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_CHILD_WINDOW_MENU, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_NO_SYSTEM_MENU, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_INVALID_MSGBOX_STYLE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_INVALID_SPI_VALUE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_SCREEN_ALREADY_LOCKED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_HWNDS_HAVE_DIFF_PARENT, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_NOT_CHILD_WINDOW, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_INVALID_GW_COMMAND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_NON_MDICHILD_WINDOW, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_POPUP_ALREADY_ACTIVE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_NO_SCROLLBARS, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_INVALID_SCROLLBAR_RANGE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_INVALID_SHOWWIN_COMMAND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_NO_SYSTEM_RESOURCES, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_NONPAGED_SYSTEM_RESOURCES, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_PAGED_SYSTEM_RESOURCES, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_MENU_ITEM_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_INVALID_KEYBOARD_HANDLE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_HOOK_TYPE_NOT_ALLOWED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_REQUIRES_INTERACTIVE_WINDOWSTATION, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_TIMEOUT, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_INVALID_MONITOR_HANDLE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_INCORRECT_SIZE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_SYMLINK_CLASS_DISABLED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_SYMLINK_NOT_SUPPORTED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_INVALID_TASK_NAME, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_INVALID_TASK_INDEX, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_THREAD_ALREADY_IN_TASK, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_INSTALL_SERVICE_FAILURE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_INSTALL_USEREXIT, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_INSTALL_FAILURE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_INSTALL_SUSPEND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_UNKNOWN_PRODUCT, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_UNKNOWN_FEATURE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_UNKNOWN_COMPONENT, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_UNKNOWN_PROPERTY, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_INVALID_HANDLE_STATE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_BAD_CONFIGURATION, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_INDEX_ABSENT, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_INSTALL_SOURCE_ABSENT, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_INSTALL_PACKAGE_VERSION, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_PRODUCT_UNINSTALLED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_BAD_QUERY_SYNTAX, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_INVALID_FIELD, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_INSTALL_ALREADY_RUNNING, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_INSTALL_PACKAGE_OPEN_FAILED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_INSTALL_PACKAGE_INVALID, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_INSTALL_UI_FAILURE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_INSTALL_LOG_FAILURE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_INSTALL_LANGUAGE_UNSUPPORTED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_INSTALL_TRANSFORM_FAILURE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_INSTALL_PACKAGE_REJECTED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_FUNCTION_NOT_CALLED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_FUNCTION_FAILED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_INVALID_TABLE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DATATYPE_MISMATCH, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_UNSUPPORTED_TYPE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_CREATE_FAILED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_INSTALL_TEMP_UNWRITABLE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_INSTALL_PLATFORM_UNSUPPORTED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_INSTALL_NOTUSED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_PATCH_PACKAGE_OPEN_FAILED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_PATCH_PACKAGE_INVALID, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_PATCH_PACKAGE_UNSUPPORTED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_PRODUCT_VERSION, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_INVALID_COMMAND_LINE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_INSTALL_REMOTE_DISALLOWED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_SUCCESS_REBOOT_INITIATED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_PATCH_TARGET_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_PATCH_PACKAGE_REJECTED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_INSTALL_TRANSFORM_REJECTED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_INSTALL_REMOTE_PROHIBITED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_PATCH_REMOVAL_UNSUPPORTED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_UNKNOWN_PATCH, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_PATCH_NO_SEQUENCE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_PATCH_REMOVAL_DISALLOWED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_INVALID_PATCH_XML, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_PATCH_MANAGED_ADVERTISED_PRODUCT, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_INSTALL_SERVICE_SAFEBOOT, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, RPC_S_INVALID_BINDING, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, RPC_S_PROXY_ACCESS_DENIED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, RPC_X_SS_IN_NULL_CONTEXT, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_REDIRECTOR_HAS_OPEN_HANDLES, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_PRINTER_DRIVER_ALREADY_INSTALLED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_UNKNOWN_PORT, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_UNKNOWN_PRINTER_DRIVER, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_UNKNOWN_PRINTPROCESSOR, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_INVALID_SEPARATOR_FILE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_INVALID_PRIORITY, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_INVALID_PRINTER_NAME, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_PRINTER_ALREADY_EXISTS, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_INVALID_PRINTER_COMMAND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_INVALID_DATATYPE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_INVALID_ENVIRONMENT, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_SERVER_HAS_OPEN_HANDLES, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_INVALID_TIME, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_INVALID_FORM_NAME, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_INVALID_FORM_SIZE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_ALREADY_WAITING, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_PRINTER_DELETED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_INVALID_PRINTER_STATE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, OR_INVALID_OXID, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, OR_INVALID_OID, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, OR_INVALID_SET, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_NO_SITENAME, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, RPC_S_ENTRY_TYPE_MISMATCH, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, RPC_S_NOT_ALL_OBJS_EXPORTED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, RPC_S_INTERFACE_NOT_EXPORTED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, RPC_S_PROFILE_NOT_ADDED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, RPC_S_PRF_ELT_NOT_ADDED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, RPC_S_PRF_ELT_NOT_REMOVED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, RPC_S_GRP_ELT_NOT_ADDED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, RPC_S_GRP_ELT_NOT_REMOVED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_KM_DRIVER_BLOCKED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_CONTEXT_EXPIRED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_PER_USER_TRUST_QUOTA_EXCEEDED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_ALL_USER_TRUST_QUOTA_EXCEEDED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_USER_DELETE_TRUST_QUOTA_EXCEEDED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_AUTHENTICATION_FIREWALL_FAILED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_REMOTE_PRINT_CONNECTIONS_BLOCKED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_INVALID_PIXEL_FORMAT, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_INVALID_WINDOW_STYLE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_METAFILE_NOT_SUPPORTED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_TRANSFORM_NOT_SUPPORTED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_CLIPPING_NOT_SUPPORTED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_INVALID_CMM, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_INVALID_PROFILE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_TAG_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_TAG_NOT_PRESENT, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DUPLICATE_TAG, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_PROFILE_NOT_ASSOCIATED_WITH_DEVICE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_PROFILE_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_INVALID_COLORSPACE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_ICM_NOT_ENABLED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DELETING_ICM_XFORM, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_INVALID_TRANSFORM, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_COLORSPACE_MISMATCH, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_INVALID_COLORINDEX, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_PROFILE_DOES_NOT_MATCH_DEVICE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, NERR_NetNotStarted, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, NERR_ShareMem, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, NERR_NoNetworkResource, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, NERR_RemoteOnly, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, NERR_DevNotRedirected, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_CONNECTED_OTHER_PASSWORD, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_CONNECTED_OTHER_PASSWORD_DEFAULT, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, NERR_ServerNotStarted, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, NERR_ItemNotFound, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, NERR_RedirectedPath, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, NERR_DuplicateShare, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, NERR_NoRoom, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, NERR_TooManyItems, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, NERR_InvalidMaxUsers, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, NERR_BufTooSmall, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, NERR_RemoteErr, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, NERR_LanmanIniError, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, NERR_NetworkError, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, NERR_WkstaInconsistentState, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, NERR_WkstaNotStarted, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, NERR_BrowserNotStarted, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, NERR_InternalError, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, NERR_BadTransactConfig, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, NERR_InvalidAPI, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, NERR_BadEventName, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, NERR_DupNameReboot, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, NERR_CfgCompNotFound, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, NERR_CfgParamNotFound, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, NERR_LineTooLong, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, NERR_QNotFound, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, NERR_JobNotFound, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, NERR_DestNotFound, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, NERR_DestExists, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, NERR_QExists, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, NERR_QNoRoom, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, NERR_JobNoRoom, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, NERR_DestNoRoom, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, NERR_DestIdle, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, NERR_DestInvalidOp, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, NERR_ProcNoRespond, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, NERR_SpoolerNotLoaded, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, NERR_DestInvalidState, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, NERR_QinvalidState, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, NERR_JobInvalidState, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, NERR_SpoolNoMemory, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, NERR_DriverNotFound, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, NERR_DataTypeInvalid, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, NERR_ProcNotFound, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, NERR_ServiceTableLocked, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, NERR_ServiceTableFull, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, NERR_ServiceInstalled, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, NERR_ServiceEntryLocked, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, NERR_ServiceNotInstalled, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, NERR_BadServiceName, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, NERR_ServiceCtlTimeout, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, NERR_ServiceCtlBusy, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, NERR_BadServiceProgName, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, NERR_ServiceNotCtrl, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, NERR_ServiceKillProc, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, NERR_ServiceCtlNotValid, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, NERR_NotInDispatchTbl, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, NERR_BadControlRecv, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, NERR_ServiceNotStarting, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, NERR_AlreadyLoggedOn, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, NERR_NotLoggedOn, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, NERR_BAD_USERNAME, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, NERR_BadPassword, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, NERR_UnableToAddName_W, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, NERR_UnableToAddName_F, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, NERR_UnableToDelName_W, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, NERR_UnableToDelName_F, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, NERR_LogonsPaused, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, NERR_LogonServerConflict, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, NERR_LogonNoUserPath, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, NERR_LogonScriptError, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, NERR_StandaloneLogon, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, NERR_LogonServerNotFound, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, NERR_LogonDomainExists, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, NERR_NonValidatedLogon, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, NERR_ACFNotFound, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, NERR_GroupNotFound, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, NERR_UserNotFound, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, NERR_ResourceNotFound, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, NERR_GroupExists, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, NERR_UserExists, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, NERR_ResourceExists, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, NERR_NotPrimary, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, NERR_ACFNotLoaded, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, NERR_ACFNoRoom, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, NERR_ACFFileIOFail, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, NERR_ACFTooManyLists, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, NERR_UserLogon, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, NERR_ACFNoParent, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, NERR_CanNotGrowSegment, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, NERR_SpeGroupOp, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, NERR_NotInCache, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, NERR_UserInGroup, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, NERR_UserNotInGroup, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, NERR_AccountUndefined, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, NERR_AccountExpired, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, NERR_InvalidWorkstation, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, NERR_InvalidLogonHours, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, NERR_PasswordExpired, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, NERR_PasswordCantChange, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, NERR_PasswordHistConflict, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, NERR_PasswordTooShort, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, NERR_PasswordTooRecent, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, NERR_InvalidDatabase, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, NERR_DatabaseUpToDate, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, NERR_SyncRequired, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, NERR_UseNotFound, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, NERR_BadAsgType, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, NERR_DeviceIsShared, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, NERR_NoComputerName, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, NERR_MsgAlreadyStarted, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, NERR_MsgInitFailed, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, NERR_NameNotFound, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, NERR_AlreadyForwarded, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, NERR_AddForwarded, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, NERR_AlreadyExists, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, NERR_TooManyNames, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, NERR_DelComputerName, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, NERR_LocalForward, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, NERR_GrpMsgProcessor, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, NERR_PausedRemote, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, NERR_BadReceive, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, NERR_NameInUse, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, NERR_MsgNotStarted, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, NERR_NotLocalName, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, NERR_NoForwardName, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, NERR_RemoteFull, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, NERR_NameNotForwarded, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, NERR_TruncatedBroadcast, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, NERR_InvalidDevice, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, NERR_WriteFault, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, NERR_DuplicateName, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, NERR_DeleteLater, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, NERR_IncompleteDel, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, NERR_MultipleNets, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, NERR_NetNameNotFound, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, NERR_DeviceNotShared, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, NERR_ClientNameNotFound, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, NERR_FileIdNotFound, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, NERR_ExecFailure, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, NERR_TmpFile, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, NERR_TooMuchData, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, NERR_DeviceShareConflict, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, NERR_BrowserTableIncomplete, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, NERR_NotLocalDomain, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, NERR_IsDfsShare, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, NERR_DevInvalidOpCode, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, NERR_DevNotFound, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, NERR_DevNotOpen, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, NERR_BadQueueDevString, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, NERR_BadQueuePriority, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, NERR_NoCommDevs, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, NERR_QueueNotFound, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, NERR_BadDevString, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, NERR_BadDev, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, NERR_InUseBySpooler, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, NERR_CommDevInUse, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, NERR_InvalidComputer, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, NERR_MaxLenExceeded, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, NERR_BadComponent, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, NERR_CantType, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, NERR_TooManyEntries, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, NERR_ProfileFileTooBig, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, NERR_ProfileOffset, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, NERR_ProfileCleanup, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, NERR_ProfileLoadErr, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, NERR_ProfileSaveErr, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, NERR_LogOverflow, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, NERR_LogFileChanged, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, NERR_LogFileCorrupt, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, NERR_SourceIsDir, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, NERR_BadSource, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, NERR_BadDest, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, NERR_DifferentServers, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, NERR_RunSrvPaused, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, NERR_ErrCommRunSrv, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, NERR_ErrorExecingGhost, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, NERR_ShareNotFound, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, NERR_InvalidLana, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, NERR_BadPasswordCore, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, NERR_LocalDrive, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, NERR_AlertExists, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, NERR_TooManyAlerts, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, NERR_NoSuchAlert, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, NERR_BadRecipient, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, NERR_AcctLimitExceeded, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, NERR_InvalidLogSeek, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, NERR_BadUasConfig, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, NERR_InvalidUASOp, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, NERR_LastAdmin, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, NERR_DCNotFound, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, NERR_LogonTrackingError, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, NERR_NetlogonNotStarted, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, NERR_CanNotGrowUASFile, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, NERR_TimeDiffAtDC, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, NERR_PasswordMismatch, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, NERR_NoSuchServer, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, NERR_NoSuchSession, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, NERR_NoSuchConnection, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, NERR_TooManyServers, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, NERR_TooManySessions, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, NERR_TooManyConnections, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, NERR_TooManyFiles, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, NERR_NoAlternateServers, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, NERR_TryDownLevel, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, NERR_UPSDriverNotStarted, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, NERR_UPSInvalidConfig, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, NERR_UPSInvalidCommPort, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, NERR_UPSSignalAsserted, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, NERR_UPSShutdownFailed, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, NERR_BadDosRetCode, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, NERR_ProgNeedsExtraMem, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, NERR_BadDosFunction, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, NERR_RemoteBootFailed, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, NERR_BadFileCheckSum, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, NERR_NoRplBootSystem, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, NERR_RplLoadrNetBiosErr, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, NERR_RplLoadrDiskErr, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, NERR_ImageParamErr, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, NERR_TooManyImageParams, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, NERR_NonDosFloppyUsed, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, NERR_RplBootRestart, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, NERR_RplSrvrCallFailed, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, NERR_CantConnectRplSrvr, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, NERR_CantOpenImageFile, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, NERR_CallingRplSrvr, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, NERR_StartingRplBoot, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, NERR_RplBootServiceTerm, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, NERR_RplBootStartFailed, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, NERR_RPL_CONNECTED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, NERR_BrowserConfiguredToNotRun, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, NERR_RplNoAdaptersStarted, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, NERR_RplBadRegistry, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, NERR_RplBadDatabase, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, NERR_RplRplfilesShare, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, NERR_RplNotRplServer, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, NERR_RplCannotEnum, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, NERR_RplWkstaInfoCorrupted, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, NERR_RplWkstaNotFound, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, NERR_RplWkstaNameUnavailable, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, NERR_RplProfileInfoCorrupted, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, NERR_RplProfileNotFound, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, NERR_RplProfileNameUnavailable, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, NERR_RplProfileNotEmpty, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, NERR_RplConfigInfoCorrupted, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, NERR_RplConfigNotFound, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, NERR_RplAdapterInfoCorrupted, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, NERR_RplInternal, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, NERR_RplVendorInfoCorrupted, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, NERR_RplBootInfoCorrupted, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, NERR_RplWkstaNeedsUserAcct, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, NERR_RplNeedsRPLUSERAcct, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, NERR_RplBootNotFound, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, NERR_RplIncompatibleProfile, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, NERR_RplAdapterNameUnavailable, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, NERR_RplConfigNotEmpty, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, NERR_RplBootInUse, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, NERR_RplBackupDatabase, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, NERR_RplAdapterNotFound, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, NERR_RplVendorNotFound, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, NERR_RplVendorNameUnavailable, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, NERR_RplBootNameUnavailable, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, NERR_RplConfigNameUnavailable, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, NERR_DfsInternalCorruption, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, NERR_DfsVolumeDataCorrupt, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, NERR_DfsNoSuchVolume, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, NERR_DfsVolumeAlreadyExists, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, NERR_DfsAlreadyShared, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, NERR_DfsNoSuchShare, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, NERR_DfsNotALeafVolume, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, NERR_DfsLeafVolume, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, NERR_DfsVolumeHasMultipleServers, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, NERR_DfsCantCreateJunctionPoint, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, NERR_DfsServerNotDfsAware, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, NERR_DfsBadRenamePath, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, NERR_DfsVolumeIsOffline, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, NERR_DfsNoSuchServer, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, NERR_DfsCyclicalName, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, NERR_DfsNotSupportedInServerDfs, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, NERR_DfsDuplicateService, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, NERR_DfsCantRemoveLastServerShare, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, NERR_DfsVolumeIsInterDfs, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, NERR_DfsInconsistent, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, NERR_DfsServerUpgraded, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, NERR_DfsDataIsIdentical, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, NERR_DfsCantRemoveDfsRoot, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, NERR_DfsChildOrParentInDfs, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, NERR_DfsInternalError, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, NERR_SetupAlreadyJoined, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, NERR_SetupNotJoined, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, NERR_SetupDomainController, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, NERR_DefaultJoinRequired, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, NERR_InvalidWorkgroupName, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, NERR_NameUsesIncompatibleCodePage, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, NERR_ComputerAccountNotFound, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, NERR_PersonalSku, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, NERR_PasswordMustChange, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, NERR_AccountLockedOut, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, NERR_PasswordTooLong, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, NERR_PasswordNotComplexEnough, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, NERR_PasswordFilterError, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_UNKNOWN_PRINT_MONITOR, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_PRINTER_DRIVER_IN_USE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_SPOOL_FILE_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_SPL_NO_STARTDOC, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_SPL_NO_ADDJOB, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_PRINT_PROCESSOR_ALREADY_INSTALLED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_PRINT_MONITOR_ALREADY_INSTALLED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_INVALID_PRINT_MONITOR, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_PRINT_MONITOR_IN_USE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_PRINTER_HAS_JOBS_QUEUED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_SUCCESS_REBOOT_REQUIRED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_SUCCESS_RESTART_REQUIRED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_PRINTER_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_PRINTER_DRIVER_WARNED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_PRINTER_DRIVER_BLOCKED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_PRINTER_DRIVER_PACKAGE_IN_USE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_CORE_DRIVER_PACKAGE_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_FAIL_REBOOT_REQUIRED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_FAIL_REBOOT_INITIATED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_IO_REISSUE_AS_CACHED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_WINS_INTERNAL, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_CAN_NOT_DEL_LOCAL_WINS, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_STATIC_INIT, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_INC_BACKUP, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_FULL_BACKUP, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_REC_NON_EXISTENT, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_RPL_NOT_ALLOWED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, PEERDIST_ERROR_CONTENTINFO_VERSION_UNSUPPORTED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, PEERDIST_ERROR_CANNOT_PARSE_CONTENTINFO, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, PEERDIST_ERROR_MISSING_DATA, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, PEERDIST_ERROR_NO_MORE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, PEERDIST_ERROR_NOT_INITIALIZED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, PEERDIST_ERROR_ALREADY_INITIALIZED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, PEERDIST_ERROR_SHUTDOWN_IN_PROGRESS, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, PEERDIST_ERROR_INVALIDATED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, PEERDIST_ERROR_ALREADY_EXISTS, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, PEERDIST_ERROR_OPERATION_NOTFOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, PEERDIST_ERROR_ALREADY_COMPLETED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, PEERDIST_ERROR_OUT_OF_BOUNDS, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, PEERDIST_ERROR_VERSION_UNSUPPORTED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, PEERDIST_ERROR_INVALID_CONFIGURATION, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, PEERDIST_ERROR_NOT_LICENSED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, PEERDIST_ERROR_SERVICE_UNAVAILABLE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DHCP_ADDRESS_CONFLICT, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_WMI_DP_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_WMI_UNRESOLVED_INSTANCE_REF, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_WMI_SERVER_UNAVAILABLE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_WMI_DP_FAILED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_WMI_INVALID_MOF, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_WMI_INVALID_REGINFO, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_INVALID_MEDIA, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_INVALID_LIBRARY, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_INVALID_MEDIA_POOL, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DRIVE_MEDIA_MISMATCH, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_MEDIA_OFFLINE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_LIBRARY_OFFLINE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_EMPTY, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_NOT_EMPTY, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_MEDIA_UNAVAILABLE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_RESOURCE_DISABLED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_INVALID_CLEANER, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_UNABLE_TO_CLEAN, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_OBJECT_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DATABASE_FAILURE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DATABASE_FULL, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_MEDIA_INCOMPATIBLE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_RESOURCE_NOT_PRESENT, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_INVALID_OPERATION, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_MEDIA_NOT_AVAILABLE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DEVICE_NOT_AVAILABLE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_REQUEST_REFUSED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_INVALID_DRIVE_OBJECT, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_LIBRARY_FULL, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_MEDIUM_NOT_ACCESSIBLE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_UNABLE_TO_LOAD_MEDIUM, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_UNABLE_TO_INVENTORY_DRIVE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_UNABLE_TO_INVENTORY_SLOT, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_UNABLE_TO_INVENTORY_TRANSPORT, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_CONTROLLING_IEPORT, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_UNABLE_TO_EJECT_MOUNTED_MEDIA, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_CLEANER_SLOT_SET, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_CLEANER_SLOT_NOT_SET, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_CLEANER_CARTRIDGE_SPENT, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_UNEXPECTED_OMID, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_CANT_DELETE_LAST_ITEM, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_MESSAGE_EXCEEDS_MAX_SIZE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_VOLUME_CONTAINS_SYS_FILES, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_INDIGENOUS_TYPE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_NO_SUPPORTING_DRIVES, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_IEPORT_FULL, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_VOLUME_NOT_SIS_ENABLED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DEPENDENT_RESOURCE_EXISTS, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DEPENDENCY_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DEPENDENCY_ALREADY_EXISTS, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_RESOURCE_NOT_ONLINE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_HOST_NODE_NOT_AVAILABLE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_RESOURCE_NOT_AVAILABLE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_RESOURCE_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_SHUTDOWN_CLUSTER, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_CANT_EVICT_ACTIVE_NODE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_OBJECT_ALREADY_EXISTS, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_OBJECT_IN_LIST, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_GROUP_NOT_AVAILABLE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_GROUP_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_GROUP_NOT_ONLINE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_HOST_NODE_NOT_RESOURCE_OWNER, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_HOST_NODE_NOT_GROUP_OWNER, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_RESMON_CREATE_FAILED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_RESMON_ONLINE_FAILED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_RESOURCE_ONLINE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_QUORUM_RESOURCE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_NOT_QUORUM_CAPABLE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_CLUSTER_SHUTTING_DOWN, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_INVALID_STATE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_RESOURCE_PROPERTIES_STORED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_NOT_QUORUM_CLASS, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_CORE_RESOURCE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_QUORUM_RESOURCE_ONLINE_FAILED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_QUORUMLOG_OPEN_FAILED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_CLUSTERLOG_CORRUPT, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_CLUSTERLOG_RECORD_EXCEEDS_MAXSIZE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_CLUSTERLOG_EXCEEDS_MAXSIZE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_CLUSTERLOG_CHKPOINT_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_CLUSTERLOG_NOT_ENOUGH_SPACE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_QUORUM_OWNER_ALIVE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_NETWORK_NOT_AVAILABLE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_NODE_NOT_AVAILABLE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_ALL_NODES_NOT_AVAILABLE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_RESOURCE_FAILED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_CLUSTER_INVALID_NODE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_CLUSTER_NODE_EXISTS, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_CLUSTER_JOIN_IN_PROGRESS, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_CLUSTER_NODE_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_CLUSTER_LOCAL_NODE_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_CLUSTER_NETWORK_EXISTS, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_CLUSTER_NETWORK_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_CLUSTER_NETINTERFACE_EXISTS, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_CLUSTER_NETINTERFACE_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_CLUSTER_INVALID_REQUEST, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_CLUSTER_INVALID_NETWORK_PROVIDER, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_CLUSTER_NODE_DOWN, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_CLUSTER_NODE_UNREACHABLE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_CLUSTER_NODE_NOT_MEMBER, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_CLUSTER_JOIN_NOT_IN_PROGRESS, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_CLUSTER_INVALID_NETWORK, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_CLUSTER_NODE_UP, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_CLUSTER_IPADDR_IN_USE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_CLUSTER_NODE_NOT_PAUSED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_CLUSTER_NO_SECURITY_CONTEXT, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_CLUSTER_NETWORK_NOT_INTERNAL, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_CLUSTER_NODE_ALREADY_UP, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_CLUSTER_NODE_ALREADY_DOWN, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_CLUSTER_NETWORK_ALREADY_ONLINE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_CLUSTER_NETWORK_ALREADY_OFFLINE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_CLUSTER_NODE_ALREADY_MEMBER, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_CLUSTER_LAST_INTERNAL_NETWORK, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_CLUSTER_NETWORK_HAS_DEPENDENTS, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_INVALID_OPERATION_ON_QUORUM, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DEPENDENCY_NOT_ALLOWED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_CLUSTER_NODE_PAUSED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_NODE_CANT_HOST_RESOURCE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_CLUSTER_NODE_NOT_READY, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_CLUSTER_NODE_SHUTTING_DOWN, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_CLUSTER_JOIN_ABORTED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_CLUSTER_INCOMPATIBLE_VERSIONS, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_CLUSTER_MAXNUM_OF_RESOURCES_EXCEEDED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_CLUSTER_SYSTEM_CONFIG_CHANGED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_CLUSTER_RESOURCE_TYPE_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_CLUSTER_RESTYPE_NOT_SUPPORTED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_CLUSTER_RESNAME_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_CLUSTER_NO_RPC_PACKAGES_REGISTERED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_CLUSTER_OWNER_NOT_IN_PREFLIST, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_CLUSTER_DATABASE_SEQMISMATCH, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_RESMON_INVALID_STATE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_CLUSTER_GUM_NOT_LOCKER, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_QUORUM_DISK_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DATABASE_BACKUP_CORRUPT, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_CLUSTER_NODE_ALREADY_HAS_DFS_ROOT, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_RESOURCE_PROPERTY_UNCHANGEABLE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_CLUSTER_MEMBERSHIP_INVALID_STATE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_CLUSTER_QUORUMLOG_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_CLUSTER_MEMBERSHIP_HALT, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_CLUSTER_INSTANCE_ID_MISMATCH, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_CLUSTER_NETWORK_NOT_FOUND_FOR_IP, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_CLUSTER_PROPERTY_DATA_TYPE_MISMATCH, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_CLUSTER_EVICT_WITHOUT_CLEANUP, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_CLUSTER_PARAMETER_MISMATCH, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_NODE_CANNOT_BE_CLUSTERED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_CLUSTER_WRONG_OS_VERSION, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_CLUSTER_CANT_CREATE_DUP_CLUSTER_NAME, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_CLUSCFG_ALREADY_COMMITTED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_CLUSCFG_ROLLBACK_FAILED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_CLUSCFG_SYSTEM_DISK_DRIVE_LETTER_CONFLICT, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_CLUSTER_OLD_VERSION, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_CLUSTER_MISMATCHED_COMPUTER_ACCT_NAME, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_CLUSTER_NO_NET_ADAPTERS, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_CLUSTER_POISONED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_CLUSTER_GROUP_MOVING, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_CLUSTER_RESOURCE_TYPE_BUSY, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_RESOURCE_CALL_TIMED_OUT, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_INVALID_CLUSTER_IPV6_ADDRESS, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_CLUSTER_INTERNAL_INVALID_FUNCTION, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_CLUSTER_PARAMETER_OUT_OF_BOUNDS, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_CLUSTER_PARTIAL_SEND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_CLUSTER_REGISTRY_INVALID_FUNCTION, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_CLUSTER_INVALID_STRING_TERMINATION, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_CLUSTER_INVALID_STRING_FORMAT, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_CLUSTER_DATABASE_TRANSACTION_IN_PROGRESS, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_CLUSTER_DATABASE_TRANSACTION_NOT_IN_PROGRESS, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_CLUSTER_NULL_DATA, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_CLUSTER_PARTIAL_READ, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_CLUSTER_PARTIAL_WRITE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_CLUSTER_CANT_DESERIALIZE_DATA, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DEPENDENT_RESOURCE_PROPERTY_CONFLICT, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_CLUSTER_NO_QUORUM, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_CLUSTER_INVALID_IPV6_NETWORK, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_CLUSTER_INVALID_IPV6_TUNNEL_NETWORK, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_QUORUM_NOT_ALLOWED_IN_THIS_GROUP, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_ENCRYPTION_FAILED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DECRYPTION_FAILED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_NO_RECOVERY_POLICY, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_NO_EFS, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_WRONG_EFS, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_NO_USER_KEYS, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_FILE_READ_ONLY, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DIR_EFS_DISALLOWED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_EFS_SERVER_NOT_TRUSTED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_BAD_RECOVERY_POLICY, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_VOLUME_NOT_SUPPORT_EFS, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_EFS_DISABLED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_EFS_VERSION_NOT_SUPPORT, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_CS_ENCRYPTION_INVALID_SERVER_RESPONSE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_CS_ENCRYPTION_UNSUPPORTED_SERVER, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_CS_ENCRYPTION_EXISTING_ENCRYPTED_FILE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_CS_ENCRYPTION_NEW_ENCRYPTED_FILE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_CS_ENCRYPTION_FILE_NOT_CSE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, SCHED_E_SERVICE_NOT_LOCALSYSTEM, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_LOG_SECTOR_INVALID, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_LOG_SECTOR_PARITY_INVALID, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_LOG_SECTOR_REMAPPED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_LOG_BLOCK_INCOMPLETE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_LOG_INVALID_RANGE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_LOG_BLOCKS_EXHAUSTED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_LOG_READ_CONTEXT_INVALID, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_LOG_RESTART_INVALID, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_LOG_BLOCK_VERSION, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_LOG_BLOCK_INVALID, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_LOG_READ_MODE_INVALID, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_LOG_NO_RESTART, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_LOG_METADATA_CORRUPT, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_LOG_METADATA_INVALID, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_LOG_METADATA_INCONSISTENT, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_LOG_RESERVATION_INVALID, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_LOG_CANT_DELETE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_LOG_CONTAINER_LIMIT_EXCEEDED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_LOG_START_OF_LOG, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_LOG_POLICY_ALREADY_INSTALLED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_LOG_POLICY_NOT_INSTALLED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_LOG_POLICY_INVALID, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_LOG_POLICY_CONFLICT, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_LOG_PINNED_ARCHIVE_TAIL, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_LOG_RECORD_NONEXISTENT, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_LOG_RECORDS_RESERVED_INVALID, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_LOG_SPACE_RESERVED_INVALID, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_LOG_TAIL_INVALID, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_LOG_FULL, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_COULD_NOT_RESIZE_LOG, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_LOG_MULTIPLEXED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_LOG_DEDICATED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_LOG_ARCHIVE_NOT_IN_PROGRESS, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_LOG_ARCHIVE_IN_PROGRESS, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_LOG_EPHEMERAL, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_LOG_NOT_ENOUGH_CONTAINERS, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_LOG_CLIENT_ALREADY_REGISTERED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_LOG_CLIENT_NOT_REGISTERED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_LOG_FULL_HANDLER_IN_PROGRESS, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_LOG_CONTAINER_READ_FAILED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_LOG_CONTAINER_WRITE_FAILED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_LOG_CONTAINER_OPEN_FAILED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_LOG_CONTAINER_STATE_INVALID, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_LOG_STATE_INVALID, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_LOG_PINNED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_LOG_METADATA_FLUSH_FAILED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_LOG_INCONSISTENT_SECURITY, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_LOG_APPENDED_FLUSH_FAILED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_LOG_PINNED_RESERVATION, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_INVALID_TRANSACTION, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_TRANSACTION_NOT_ACTIVE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_TRANSACTION_REQUEST_NOT_VALID, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_TRANSACTION_NOT_REQUESTED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_TRANSACTION_ALREADY_ABORTED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_TRANSACTION_ALREADY_COMMITTED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_TM_INITIALIZATION_FAILED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_RESOURCEMANAGER_READ_ONLY, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_TRANSACTION_NOT_JOINED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_TRANSACTION_SUPERIOR_EXISTS, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_CRM_PROTOCOL_ALREADY_EXISTS, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_TRANSACTION_PROPAGATION_FAILED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_CRM_PROTOCOL_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_TRANSACTION_INVALID_MARSHALL_BUFFER, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_CURRENT_TRANSACTION_NOT_VALID, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_TRANSACTION_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_RESOURCEMANAGER_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_ENLISTMENT_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_TRANSACTIONMANAGER_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_TRANSACTIONMANAGER_NOT_ONLINE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_TRANSACTIONMANAGER_RECOVERY_NAME_COLLISION, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_TRANSACTIONAL_CONFLICT, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_RM_NOT_ACTIVE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_RM_METADATA_CORRUPT, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DIRECTORY_NOT_RM, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_TRANSACTIONS_UNSUPPORTED_REMOTE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_LOG_RESIZE_INVALID_SIZE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_OBJECT_NO_LONGER_EXISTS, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_STREAM_MINIVERSION_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_STREAM_MINIVERSION_NOT_VALID, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_MINIVERSION_INACCESSIBLE_FROM_SPECIFIED_TRANSACTION, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_CANT_OPEN_MINIVERSION_WITH_MODIFY_INTENT, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_CANT_CREATE_MORE_STREAM_MINIVERSIONS, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_REMOTE_FILE_VERSION_MISMATCH, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_HANDLE_NO_LONGER_VALID, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_NO_TXF_METADATA, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_LOG_CORRUPTION_DETECTED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_CANT_RECOVER_WITH_HANDLE_OPEN, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_RM_DISCONNECTED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_ENLISTMENT_NOT_SUPERIOR, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_RECOVERY_NOT_NEEDED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_RM_ALREADY_STARTED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_FILE_IDENTITY_NOT_PERSISTENT, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_CANT_BREAK_TRANSACTIONAL_DEPENDENCY, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_CANT_CROSS_RM_BOUNDARY, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_TXF_DIR_NOT_EMPTY, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_INDOUBT_TRANSACTIONS_EXIST, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_TM_VOLATILE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_ROLLBACK_TIMER_EXPIRED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_TXF_ATTRIBUTE_CORRUPT, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_EFS_NOT_ALLOWED_IN_TRANSACTION, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_TRANSACTIONAL_OPEN_NOT_ALLOWED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_LOG_GROWTH_FAILED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_TRANSACTED_MAPPING_UNSUPPORTED_REMOTE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_TXF_METADATA_ALREADY_PRESENT, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_TRANSACTION_SCOPE_CALLBACKS_NOT_SET, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_TRANSACTION_REQUIRED_PROMOTION, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_CANNOT_EXECUTE_FILE_IN_TRANSACTION, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_TRANSACTIONS_NOT_FROZEN, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_TRANSACTION_FREEZE_IN_PROGRESS, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_NOT_SNAPSHOT_VOLUME, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_NO_SAVEPOINT_WITH_OPEN_FILES, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DATA_LOST_REPAIR, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_SPARSE_NOT_ALLOWED_IN_TRANSACTION, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_TM_IDENTITY_MISMATCH, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_FLOATED_SECTION, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_CANNOT_ACCEPT_TRANSACTED_WORK, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_CANNOT_ABORT_TRANSACTIONS, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_CTX_WINSTATION_NAME_INVALID, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_CTX_INVALID_PD, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_CTX_PD_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_CTX_WD_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_CTX_CANNOT_MAKE_EVENTLOG_ENTRY, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_CTX_SERVICE_NAME_COLLISION, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_CTX_CLOSE_PENDING, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_CTX_NO_OUTBUF, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_CTX_MODEM_INF_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_CTX_INVALID_MODEMNAME, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_CTX_MODEM_RESPONSE_ERROR, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_CTX_MODEM_RESPONSE_TIMEOUT, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_CTX_MODEM_RESPONSE_NO_CARRIER, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_CTX_MODEM_RESPONSE_NO_DIALTONE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_CTX_MODEM_RESPONSE_BUSY, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_CTX_MODEM_RESPONSE_VOICE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_CTX_TD_ERROR, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_CTX_WINSTATION_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_CTX_WINSTATION_ALREADY_EXISTS, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_CTX_WINSTATION_BUSY, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_CTX_BAD_VIDEO_MODE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_CTX_GRAPHICS_INVALID, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_CTX_LOGON_DISABLED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_CTX_NOT_CONSOLE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_CTX_CLIENT_QUERY_TIMEOUT, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_CTX_CONSOLE_DISCONNECT, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_CTX_CONSOLE_CONNECT, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_CTX_SHADOW_DENIED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_CTX_WINSTATION_ACCESS_DENIED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_CTX_INVALID_WD, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_CTX_SHADOW_INVALID, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_CTX_SHADOW_DISABLED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_CTX_CLIENT_LICENSE_IN_USE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_CTX_CLIENT_LICENSE_NOT_SET, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_CTX_LICENSE_CLIENT_INVALID, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_CTX_LICENSE_EXPIRED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_CTX_SHADOW_NOT_RUNNING, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_CTX_SHADOW_ENDED_BY_MODE_CHANGE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_ACTIVATION_COUNT_EXCEEDED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_CTX_WINSTATIONS_DISABLED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_CTX_ENCRYPTION_LEVEL_REQUIRED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_CTX_SESSION_IN_USE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_CTX_NO_FORCE_LOGOFF, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_CTX_ACCOUNT_RESTRICTION, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_RDP_PROTOCOL_ERROR, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_CTX_CDM_CONNECT, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_CTX_CDM_DISCONNECT, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_CTX_SECURITY_LAYER_ERROR, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_TS_INCOMPATIBLE_SESSIONS, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, FRS_ERR_INVALID_API_SEQUENCE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, FRS_ERR_STARTING_SERVICE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, FRS_ERR_STOPPING_SERVICE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, FRS_ERR_INTERNAL_API, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, FRS_ERR_INTERNAL, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, FRS_ERR_SERVICE_COMM, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, FRS_ERR_INSUFFICIENT_PRIV, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, FRS_ERR_AUTHENTICATION, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, FRS_ERR_PARENT_INSUFFICIENT_PRIV, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, FRS_ERR_PARENT_AUTHENTICATION, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, FRS_ERR_CHILD_TO_PARENT_COMM, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, FRS_ERR_PARENT_TO_CHILD_COMM, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, FRS_ERR_SYSVOL_POPULATE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, FRS_ERR_SYSVOL_POPULATE_TIMEOUT, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, FRS_ERR_SYSVOL_IS_BUSY, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, FRS_ERR_SYSVOL_DEMOTE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, FRS_ERR_INVALID_SERVICE_PARAMETER, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DS_NOT_INSTALLED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_PROMOTION_ACTIVE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_NO_PROMOTION_ACTIVE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DS_OPERATIONS_ERROR, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DS_PROTOCOL_ERROR, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DS_TIMELIMIT_EXCEEDED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DS_SIZELIMIT_EXCEEDED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DS_COMPARE_FALSE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DS_COMPARE_TRUE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DS_AUTH_METHOD_NOT_SUPPORTED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DS_STRONG_AUTH_REQUIRED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DS_INAPPROPRIATE_AUTH, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DS_AUTH_UNKNOWN, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DS_REFERRAL, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DS_UNAVAILABLE_CRIT_EXTENSION, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DS_CONFIDENTIALITY_REQUIRED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DS_INAPPROPRIATE_MATCHING, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DS_CONSTRAINT_VIOLATION, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DS_NO_SUCH_OBJECT, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DS_ALIAS_PROBLEM, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DS_INVALID_DN_SYNTAX, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DS_IS_LEAF, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DS_ALIAS_DEREF_PROBLEM, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DS_UNWILLING_TO_PERFORM, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DS_LOOP_DETECT, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DS_NAMING_VIOLATION, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DS_OBJECT_RESULTS_TOO_LARGE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DS_AFFECTS_MULTIPLE_DSAS, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DS_SERVER_DOWN, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DS_LOCAL_ERROR, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DS_ENCODING_ERROR, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DS_DECODING_ERROR, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DS_FILTER_UNKNOWN, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DS_PARAM_ERROR, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DS_NOT_SUPPORTED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DS_NO_RESULTS_RETURNED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DS_CONTROL_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DS_CLIENT_LOOP, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DS_REFERRAL_LIMIT_EXCEEDED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DS_SORT_CONTROL_MISSING, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DS_OFFSET_RANGE_ERROR, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DS_ROOT_MUST_BE_NC, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DS_ADD_REPLICA_INHIBITED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DS_ATT_NOT_DEF_IN_SCHEMA, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DS_MAX_OBJ_SIZE_EXCEEDED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DS_OBJ_STRING_NAME_EXISTS, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DS_NO_RDN_DEFINED_IN_SCHEMA, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DS_RDN_DOESNT_MATCH_SCHEMA, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DS_NO_REQUESTED_ATTS_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DS_USER_BUFFER_TO_SMALL, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DS_ATT_IS_NOT_ON_OBJ, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DS_ILLEGAL_MOD_OPERATION, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DS_OBJ_TOO_LARGE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DS_BAD_INSTANCE_TYPE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DS_MASTERDSA_REQUIRED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DS_OBJECT_CLASS_REQUIRED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DS_MISSING_REQUIRED_ATT, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DS_ATT_NOT_DEF_FOR_CLASS, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DS_ATT_ALREADY_EXISTS, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DS_CANT_ADD_ATT_VALUES, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DS_SINGLE_VALUE_CONSTRAINT, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DS_RANGE_CONSTRAINT, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DS_ATT_VAL_ALREADY_EXISTS, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DS_CANT_REM_MISSING_ATT, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DS_CANT_REM_MISSING_ATT_VAL, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DS_ROOT_CANT_BE_SUBREF, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DS_NO_CHAINING, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DS_NO_CHAINED_EVAL, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DS_NO_PARENT_OBJECT, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DS_PARENT_IS_AN_ALIAS, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DS_CANT_MIX_MASTER_AND_REPS, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DS_CHILDREN_EXIST, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DS_OBJ_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DS_ALIASED_OBJ_MISSING, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DS_BAD_NAME_SYNTAX, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DS_ALIAS_POINTS_TO_ALIAS, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DS_CANT_DEREF_ALIAS, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DS_OUT_OF_SCOPE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DS_OBJECT_BEING_REMOVED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DS_CANT_DELETE_DSA_OBJ, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DS_GENERIC_ERROR, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DS_DSA_MUST_BE_INT_MASTER, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DS_CLASS_NOT_DSA, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DS_INSUFF_ACCESS_RIGHTS, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DS_ILLEGAL_SUPERIOR, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DS_ATTRIBUTE_OWNED_BY_SAM, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DS_NAME_TOO_MANY_PARTS, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DS_NAME_TOO_LONG, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DS_NAME_VALUE_TOO_LONG, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DS_NAME_UNPARSEABLE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DS_NAME_TYPE_UNKNOWN, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DS_NOT_AN_OBJECT, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DS_SEC_DESC_TOO_SHORT, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DS_SEC_DESC_INVALID, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DS_NO_DELETED_NAME, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DS_SUBREF_MUST_HAVE_PARENT, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DS_NCNAME_MUST_BE_NC, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DS_CANT_ADD_SYSTEM_ONLY, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DS_CLASS_MUST_BE_CONCRETE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DS_INVALID_DMD, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DS_OBJ_GUID_EXISTS, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DS_NOT_ON_BACKLINK, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DS_NO_CROSSREF_FOR_NC, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DS_UNKNOWN_OPERATION, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DS_INVALID_ROLE_OWNER, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DS_COULDNT_CONTACT_FSMO, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DS_CROSS_NC_DN_RENAME, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DS_CANT_MOD_SYSTEM_ONLY, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DS_REPLICATOR_ONLY, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DS_OBJ_CLASS_NOT_DEFINED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DS_OBJ_CLASS_NOT_SUBCLASS, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DS_NAME_REFERENCE_INVALID, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DS_CROSS_REF_EXISTS, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DS_CANT_DEL_MASTER_CROSSREF, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DS_SUBTREE_NOTIFY_NOT_NC_HEAD, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DS_NOTIFY_FILTER_TOO_COMPLEX, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DS_DUP_RDN, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DS_DUP_OID, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DS_DUP_MAPI_ID, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DS_DUP_SCHEMA_ID_GUID, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DS_DUP_LDAP_DISPLAY_NAME, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DS_SEMANTIC_ATT_TEST, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DS_SYNTAX_MISMATCH, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DS_EXISTS_IN_MUST_HAVE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DS_EXISTS_IN_MAY_HAVE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DS_NONEXISTENT_MAY_HAVE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DS_NONEXISTENT_MUST_HAVE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DS_AUX_CLS_TEST_FAIL, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DS_NONEXISTENT_POSS_SUP, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DS_SUB_CLS_TEST_FAIL, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DS_BAD_RDN_ATT_ID_SYNTAX, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DS_EXISTS_IN_AUX_CLS, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DS_EXISTS_IN_SUB_CLS, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DS_EXISTS_IN_POSS_SUP, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DS_RECALCSCHEMA_FAILED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DS_TREE_DELETE_NOT_FINISHED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DS_CANT_DELETE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DS_ATT_SCHEMA_REQ_ID, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DS_BAD_ATT_SCHEMA_SYNTAX, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DS_CANT_CACHE_ATT, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DS_CANT_CACHE_CLASS, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DS_CANT_REMOVE_ATT_CACHE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DS_CANT_REMOVE_CLASS_CACHE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DS_CANT_RETRIEVE_DN, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DS_MISSING_SUPREF, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DS_CANT_RETRIEVE_INSTANCE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DS_CODE_INCONSISTENCY, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DS_DATABASE_ERROR, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DS_GOVERNSID_MISSING, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DS_MISSING_EXPECTED_ATT, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DS_NCNAME_MISSING_CR_REF, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DS_SECURITY_CHECKING_ERROR, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DS_SCHEMA_NOT_LOADED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DS_SCHEMA_ALLOC_FAILED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DS_ATT_SCHEMA_REQ_SYNTAX, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DS_GCVERIFY_ERROR, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DS_DRA_SCHEMA_MISMATCH, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DS_CANT_FIND_DSA_OBJ, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DS_CANT_FIND_EXPECTED_NC, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DS_CANT_FIND_NC_IN_CACHE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DS_CANT_RETRIEVE_CHILD, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DS_SECURITY_ILLEGAL_MODIFY, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DS_CANT_REPLACE_HIDDEN_REC, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DS_BAD_HIERARCHY_FILE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DS_BUILD_HIERARCHY_TABLE_FAILED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DS_CONFIG_PARAM_MISSING, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DS_COUNTING_AB_INDICES_FAILED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DS_HIERARCHY_TABLE_MALLOC_FAILED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DS_INTERNAL_FAILURE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DS_UNKNOWN_ERROR, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DS_ROOT_REQUIRES_CLASS_TOP, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DS_REFUSING_FSMO_ROLES, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DS_MISSING_FSMO_SETTINGS, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DS_UNABLE_TO_SURRENDER_ROLES, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DS_DRA_GENERIC, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DS_DRA_INVALID_PARAMETER, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DS_DRA_BUSY, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DS_DRA_BAD_DN, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DS_DRA_BAD_NC, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DS_DRA_DN_EXISTS, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DS_DRA_INTERNAL_ERROR, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DS_DRA_INCONSISTENT_DIT, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DS_DRA_CONNECTION_FAILED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DS_DRA_BAD_INSTANCE_TYPE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DS_DRA_OUT_OF_MEM, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DS_DRA_MAIL_PROBLEM, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DS_DRA_REF_ALREADY_EXISTS, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DS_DRA_REF_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DS_DRA_OBJ_IS_REP_SOURCE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DS_DRA_DB_ERROR, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DS_DRA_NO_REPLICA, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DS_DRA_ACCESS_DENIED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DS_DRA_NOT_SUPPORTED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DS_DRA_RPC_CANCELLED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DS_DRA_SOURCE_DISABLED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DS_DRA_SINK_DISABLED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DS_DRA_NAME_COLLISION, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DS_DRA_SOURCE_REINSTALLED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DS_DRA_MISSING_PARENT, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DS_DRA_PREEMPTED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DS_DRA_ABANDON_SYNC, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DS_DRA_SHUTDOWN, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DS_DRA_INCOMPATIBLE_PARTIAL_SET, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DS_DRA_SOURCE_IS_PARTIAL_REPLICA, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DS_DRA_EXTN_CONNECTION_FAILED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DS_INSTALL_SCHEMA_MISMATCH, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DS_DUP_LINK_ID, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DS_NAME_ERROR_RESOLVING, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DS_NAME_ERROR_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DS_NAME_ERROR_NOT_UNIQUE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DS_NAME_ERROR_NO_MAPPING, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DS_NAME_ERROR_DOMAIN_ONLY, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DS_NAME_ERROR_NO_SYNTACTICAL_MAPPING, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DS_CONSTRUCTED_ATT_MOD, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DS_WRONG_OM_OBJ_CLASS, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DS_DRA_REPL_PENDING, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DS_INVALID_LDAP_DISPLAY_NAME, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DS_NON_BASE_SEARCH, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DS_CANT_RETRIEVE_ATTS, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DS_BACKLINK_WITHOUT_LINK, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DS_EPOCH_MISMATCH, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DS_SRC_NAME_MISMATCH, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DS_SRC_AND_DST_NC_IDENTICAL, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DS_DST_NC_MISMATCH, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DS_NOT_AUTHORITIVE_FOR_DST_NC, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DS_SRC_GUID_MISMATCH, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DS_CANT_MOVE_DELETED_OBJECT, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DS_PDC_OPERATION_IN_PROGRESS, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DS_CROSS_DOMAIN_CLEANUP_REQD, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DS_ILLEGAL_XDOM_MOVE_OPERATION, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DS_CANT_WITH_ACCT_GROUP_MEMBERSHPS, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DS_NC_MUST_HAVE_NC_PARENT, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DS_CR_IMPOSSIBLE_TO_VALIDATE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DS_DST_DOMAIN_NOT_NATIVE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DS_MISSING_INFRASTRUCTURE_CONTAINER, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DS_CANT_MOVE_ACCOUNT_GROUP, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DS_CANT_MOVE_RESOURCE_GROUP, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DS_INVALID_SEARCH_FLAG, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DS_NO_TREE_DELETE_ABOVE_NC, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DS_COULDNT_LOCK_TREE_FOR_DELETE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DS_COULDNT_IDENTIFY_OBJECTS_FOR_TREE_DELETE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DS_ILLEGAL_BASE_SCHEMA_MOD, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DS_NONSAFE_SCHEMA_CHANGE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DS_SCHEMA_UPDATE_DISALLOWED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DS_CANT_CREATE_UNDER_SCHEMA, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DS_INSTALL_NO_SRC_SCH_VERSION, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DS_INSTALL_NO_SCH_VERSION_IN_INIFILE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DS_STRING_SD_CONVERSION_FAILED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DS_NAMING_MASTER_GC, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DS_DNS_LOOKUP_FAILURE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DS_COULDNT_UPDATE_SPNS, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DS_CANT_RETRIEVE_SD, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DS_KEY_NOT_UNIQUE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DS_WRONG_LINKED_ATT_SYNTAX, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DS_NO_PKT_PRIVACY_ON_CONNECTION, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DS_SOURCE_DOMAIN_IN_FOREST, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DS_DESTINATION_DOMAIN_NOT_IN_FOREST, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DS_DESTINATION_AUDITING_NOT_ENABLED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DS_CANT_FIND_DC_FOR_SRC_DOMAIN, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DS_SRC_OBJ_NOT_GROUP_OR_USER, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DS_SRC_SID_EXISTS_IN_FOREST, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DS_SRC_AND_DST_OBJECT_CLASS_MISMATCH, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DS_DRA_SCHEMA_INFO_SHIP, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DS_DRA_SCHEMA_CONFLICT, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DS_DRA_EARLIER_SCHEMA_CONFLICT, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DS_DRA_OBJ_NC_MISMATCH, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DS_NC_STILL_HAS_DSAS, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DS_CANT_ADD_TO_GC, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DS_NO_CHECKPOINT_WITH_PDC, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DS_SOURCE_AUDITING_NOT_ENABLED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DS_CANT_CREATE_IN_NONDOMAIN_NC, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DS_INVALID_NAME_FOR_SPN, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DS_FILTER_USES_CONTRUCTED_ATTRS, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DS_UNICODEPWD_NOT_IN_QUOTES, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DS_MUST_BE_RUN_ON_DST_DC, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DS_SRC_DC_MUST_BE_SP4_OR_GREATER, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DS_CANT_TREE_DELETE_CRITICAL_OBJ, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DS_FOREST_VERSION_TOO_HIGH, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DS_DOMAIN_VERSION_TOO_HIGH, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DS_FOREST_VERSION_TOO_LOW, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DS_DOMAIN_VERSION_TOO_LOW, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DS_INCOMPATIBLE_VERSION, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DS_LOW_DSA_VERSION, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DS_NO_BEHAVIOR_VERSION_IN_MIXEDDOMAIN, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DS_NOT_SUPPORTED_SORT_ORDER, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DS_NAME_NOT_UNIQUE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DS_OUT_OF_VERSION_STORE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DS_INCOMPATIBLE_CONTROLS_USED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DS_NO_REF_DOMAIN, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DS_RESERVED_LINK_ID, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DS_LINK_ID_NOT_AVAILABLE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DS_MODIFYDN_DISALLOWED_BY_INSTANCE_TYPE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DS_NO_OBJECT_MOVE_IN_SCHEMA_NC, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DS_MODIFYDN_DISALLOWED_BY_FLAG, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DS_MODIFYDN_WRONG_GRANDPARENT, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DS_NAME_ERROR_TRUST_REFERRAL, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_NOT_SUPPORTED_ON_STANDARD_SERVER, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DS_CANT_ACCESS_REMOTE_PART_OF_AD, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DS_CR_IMPOSSIBLE_TO_VALIDATE_V2, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DS_THREAD_LIMIT_EXCEEDED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DS_NOT_CLOSEST, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DS_CANT_DERIVE_SPN_WITHOUT_SERVER_REF, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DS_SINGLE_USER_MODE_FAILED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DS_NTDSCRIPT_SYNTAX_ERROR, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DS_NTDSCRIPT_PROCESS_ERROR, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DS_DIFFERENT_REPL_EPOCHS, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DS_DRS_EXTENSIONS_CHANGED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DS_REPLICA_SET_CHANGE_NOT_ALLOWED_ON_DISABLED_CR, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DS_NO_MSDS_INTID, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DS_DUP_MSDS_INTID, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DS_EXISTS_IN_RDNATTID, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DS_AUTHORIZATION_FAILED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DS_INVALID_SCRIPT, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DS_REMOTE_CROSSREF_OP_FAILED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DS_CROSS_REF_BUSY, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DS_CANT_DERIVE_SPN_FOR_DELETED_DOMAIN, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DS_CANT_DEMOTE_WITH_WRITEABLE_NC, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DS_DUPLICATE_ID_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DS_INSUFFICIENT_ATTR_TO_CREATE_OBJECT, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DS_GROUP_CONVERSION_ERROR, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DS_CANT_MOVE_APP_BASIC_GROUP, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DS_CANT_MOVE_APP_QUERY_GROUP, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DS_ROLE_NOT_VERIFIED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DS_WKO_CONTAINER_CANNOT_BE_SPECIAL, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DS_DOMAIN_RENAME_IN_PROGRESS, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DS_EXISTING_AD_CHILD_NC, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DS_REPL_LIFETIME_EXCEEDED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DS_DISALLOWED_IN_SYSTEM_CONTAINER, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DS_LDAP_SEND_QUEUE_FULL, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DS_DRA_OUT_SCHEDULE_WINDOW, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DS_POLICY_NOT_KNOWN, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_NO_SITE_SETTINGS_OBJECT, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_NO_SECRETS, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_NO_WRITABLE_DC_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DS_NO_SERVER_OBJECT, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DS_NO_NTDSA_OBJECT, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DS_NON_ASQ_SEARCH, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DS_AUDIT_FAILURE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DS_INVALID_SEARCH_FLAG_SUBTREE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DS_INVALID_SEARCH_FLAG_TUPLE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, DNS_ERROR_RCODE_FORMAT_ERROR, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, DNS_ERROR_RCODE_SERVER_FAILURE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, DNS_ERROR_RCODE_NAME_ERROR, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, DNS_ERROR_RCODE_NOT_IMPLEMENTED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, DNS_ERROR_RCODE_REFUSED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, DNS_ERROR_RCODE_YXDOMAIN, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, DNS_ERROR_RCODE_YXRRSET, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, DNS_ERROR_RCODE_NXRRSET, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, DNS_ERROR_RCODE_NOTAUTH, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, DNS_ERROR_RCODE_NOTZONE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, DNS_ERROR_RCODE_BADSIG, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, DNS_ERROR_RCODE_BADKEY, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, DNS_ERROR_RCODE_BADTIME, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, DNS_INFO_NO_RECORDS, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, DNS_ERROR_BAD_PACKET, ERRSRV, ERRerror, -1, "A bad packet was received from a DNS server. Potentially the requested address does not exist." )
S ( LW_STATUS_MESSAGE_NOT_FOUND, DNS_ERROR_NO_PACKET, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, DNS_ERROR_RCODE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, DNS_ERROR_UNSECURE_PACKET, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, DNS_ERROR_INVALID_TYPE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, DNS_ERROR_INVALID_IP_ADDRESS, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, DNS_ERROR_INVALID_PROPERTY, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, DNS_ERROR_TRY_AGAIN_LATER, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, DNS_ERROR_NOT_UNIQUE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, DNS_ERROR_NON_RFC_NAME, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, DNS_STATUS_FQDN, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, DNS_STATUS_DOTTED_NAME, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, DNS_STATUS_SINGLE_PART_NAME, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, DNS_ERROR_INVALID_NAME_CHAR, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, DNS_ERROR_NUMERIC_NAME, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, DNS_ERROR_NOT_ALLOWED_ON_ROOT_SERVER, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, DNS_ERROR_NOT_ALLOWED_UNDER_DELEGATION, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, DNS_ERROR_CANNOT_FIND_ROOT_HINTS, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, DNS_ERROR_INCONSISTENT_ROOT_HINTS, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, DNS_ERROR_DWORD_VALUE_TOO_SMALL, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, DNS_ERROR_DWORD_VALUE_TOO_LARGE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, DNS_ERROR_BACKGROUND_LOADING, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, DNS_ERROR_NOT_ALLOWED_ON_RODC, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, DNS_ERROR_ZONE_DOES_NOT_EXIST, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, DNS_ERROR_NO_ZONE_INFO, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, DNS_ERROR_INVALID_ZONE_OPERATION, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, DNS_ERROR_ZONE_CONFIGURATION_ERROR, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, DNS_ERROR_ZONE_HAS_NO_SOA_RECORD, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, DNS_ERROR_ZONE_HAS_NO_NS_RECORDS, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, DNS_ERROR_ZONE_LOCKED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, DNS_ERROR_ZONE_CREATION_FAILED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, DNS_ERROR_ZONE_ALREADY_EXISTS, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, DNS_ERROR_AUTOZONE_ALREADY_EXISTS, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, DNS_ERROR_INVALID_ZONE_TYPE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, DNS_ERROR_SECONDARY_REQUIRES_MASTER_IP, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, DNS_ERROR_ZONE_NOT_SECONDARY, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, DNS_ERROR_NEED_SECONDARY_ADDRESSES, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, DNS_ERROR_WINS_INIT_FAILED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, DNS_ERROR_NEED_WINS_SERVERS, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, DNS_ERROR_NBSTAT_INIT_FAILED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, DNS_ERROR_SOA_DELETE_INVALID, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, DNS_ERROR_FORWARDER_ALREADY_EXISTS, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, DNS_ERROR_ZONE_REQUIRES_MASTER_IP, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, DNS_ERROR_ZONE_IS_SHUTDOWN, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, DNS_ERROR_PRIMARY_REQUIRES_DATAFILE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, DNS_ERROR_INVALID_DATAFILE_NAME, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, DNS_ERROR_DATAFILE_OPEN_FAILURE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, DNS_ERROR_FILE_WRITEBACK_FAILED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, DNS_ERROR_DATAFILE_PARSING, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, DNS_ERROR_RECORD_DOES_NOT_EXIST, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, DNS_ERROR_RECORD_FORMAT, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, DNS_ERROR_NODE_CREATION_FAILED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, DNS_ERROR_UNKNOWN_RECORD_TYPE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, DNS_ERROR_RECORD_TIMED_OUT, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, DNS_ERROR_NAME_NOT_IN_ZONE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, DNS_ERROR_CNAME_LOOP, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, DNS_ERROR_NODE_IS_CNAME, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, DNS_ERROR_CNAME_COLLISION, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, DNS_ERROR_RECORD_ONLY_AT_ZONE_ROOT, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, DNS_ERROR_RECORD_ALREADY_EXISTS, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, DNS_ERROR_SECONDARY_DATA, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, DNS_ERROR_NO_CREATE_CACHE_DATA, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, DNS_ERROR_NAME_DOES_NOT_EXIST, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, DNS_WARNING_PTR_CREATE_FAILED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, DNS_WARNING_DOMAIN_UNDELETED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, DNS_ERROR_DS_UNAVAILABLE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, DNS_ERROR_DS_ZONE_ALREADY_EXISTS, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, DNS_ERROR_NO_BOOTFILE_IF_DS_ZONE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, DNS_INFO_AXFR_COMPLETE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, DNS_ERROR_AXFR, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, DNS_INFO_ADDED_LOCAL_WINS, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, DNS_STATUS_CONTINUE_NEEDED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, DNS_ERROR_NO_TCPIP, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, DNS_ERROR_NO_DNS_SERVERS, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, DNS_ERROR_DP_DOES_NOT_EXIST, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, DNS_ERROR_DP_ALREADY_EXISTS, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, DNS_ERROR_DP_NOT_ENLISTED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, DNS_ERROR_DP_ALREADY_ENLISTED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, DNS_ERROR_DP_NOT_AVAILABLE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, DNS_ERROR_DP_FSMO_ERROR, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, WSAEINTR, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, WSAEBADF, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, WSAEACCES, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, WSAEFAULT, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, WSAEINVAL, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, WSAEMFILE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, WSAEWOULDBLOCK, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, WSAEINPROGRESS, ERRSRV, ERRerror, EINPROGRESS, "Operation now in progress" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, WSAEALREADY, ERRSRV, ERRerror, EALREADY, "Operation already in progress" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, WSAENOTSOCK, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, WSAEDESTADDRREQ, ERRSRV, ERRerror, EDESTADDRREQ, "Destination address required" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, WSAEMSGSIZE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, WSAEPROTOTYPE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, WSAENOPROTOOPT, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, WSAEPROTONOSUPPORT, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, WSAESOCKTNOSUPPORT, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, WSAEOPNOTSUPP, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, WSAEPFNOSUPPORT, ERRSRV, ERRerror, EAFNOSUPPORT, "Protocol family not supported" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, WSAEADDRINUSE, ERRSRV, ERRerror, EADDRINUSE, "Address already in use" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, WSAEADDRNOTAVAIL, ERRSRV, ERRerror, EADDRNOTAVAIL, "Cannot assign requested address" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, WSAENETDOWN, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, WSAENETUNREACH, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, WSAENETRESET, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, WSAECONNABORTED, ERRSRV, ERRerror, ECONNABORTED, "connection abort" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, WSAECONNRESET, ERRSRV, ERRerror, ECONNRESET, "Connection reset by peer" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, WSAENOBUFS, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, WSAEISCONN, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, WSAENOTCONN, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, WSAESHUTDOWN, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, WSAETOOMANYREFS, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, WSAETIMEDOUT, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, WSAECONNREFUSED, ERRSRV, ERRerror, ECONNREFUSED, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, WSAELOOP, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, WSAENAMETOOLONG, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, WSAEHOSTDOWN, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, WSAEHOSTUNREACH, ERRSRV, ERRerror, EHOSTUNREACH, "Host unreachable" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, WSAENOTEMPTY, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, WSAEPROCLIM, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, WSAEUSERS, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, WSAEDQUOT, ERRSRV, ERRerror, EDQUOT, "Disk quota exceeded" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, WSAESTALE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, WSAEREMOTE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, WSASYSNOTREADY, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, WSAVERNOTSUPPORTED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, WSANOTINITIALISED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, WSAEDISCON, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, WSAENOMORE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, WSAECANCELLED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, WSAEINVALIDPROCTABLE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, WSAEINVALIDPROVIDER, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, WSAEPROVIDERFAILEDINIT, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, WSASYSCALLFAILURE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, WSASERVICE_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, WSATYPE_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, WSA_E_NO_MORE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, WSA_E_CANCELLED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, WSAEREFUSED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, WSAHOST_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, WSATRY_AGAIN, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, WSANO_RECOVERY, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, WSANO_DATA, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, WSA_QOS_RECEIVERS, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, WSA_QOS_SENDERS, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, WSA_QOS_NO_SENDERS, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, WSA_QOS_NO_RECEIVERS, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, WSA_QOS_REQUEST_CONFIRMED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, WSA_QOS_ADMISSION_FAILURE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, WSA_QOS_POLICY_FAILURE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, WSA_QOS_BAD_STYLE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, WSA_QOS_BAD_OBJECT, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, WSA_QOS_TRAFFIC_CTRL_ERROR, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, WSA_QOS_GENERIC_ERROR, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, WSA_QOS_ESERVICETYPE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, WSA_QOS_EFLOWSPEC, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, WSA_QOS_EPROVSPECBUF, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, WSA_QOS_EFILTERSTYLE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, WSA_QOS_EFILTERTYPE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, WSA_QOS_EFILTERCOUNT, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, WSA_QOS_EOBJLENGTH, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, WSA_QOS_EFLOWCOUNT, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, WSA_QOS_EUNKOWNPSOBJ, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, WSA_QOS_EPOLICYOBJ, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, WSA_QOS_EFLOWDESC, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, WSA_QOS_EPSFLOWSPEC, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, WSA_QOS_EPSFILTERSPEC, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, WSA_QOS_ESDMODEOBJ, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, WSA_QOS_ESHAPERATEOBJ, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, WSA_QOS_RESERVED_PETYPE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, WSA_IPSEC_NAME_POLICY_ERROR, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_IPSEC_QM_POLICY_EXISTS, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_IPSEC_QM_POLICY_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_IPSEC_QM_POLICY_IN_USE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_IPSEC_MM_POLICY_EXISTS, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_IPSEC_MM_POLICY_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_IPSEC_MM_POLICY_IN_USE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_IPSEC_MM_FILTER_EXISTS, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_IPSEC_MM_FILTER_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_IPSEC_TRANSPORT_FILTER_EXISTS, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_IPSEC_TRANSPORT_FILTER_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_IPSEC_MM_AUTH_EXISTS, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_IPSEC_MM_AUTH_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_IPSEC_MM_AUTH_IN_USE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_IPSEC_DEFAULT_MM_POLICY_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_IPSEC_DEFAULT_MM_AUTH_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_IPSEC_DEFAULT_QM_POLICY_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_IPSEC_TUNNEL_FILTER_EXISTS, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_IPSEC_TUNNEL_FILTER_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_IPSEC_MM_FILTER_PENDING_DELETION, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_IPSEC_TRANSPORT_FILTER_PENDING_DELETION, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_IPSEC_TUNNEL_FILTER_PENDING_DELETION, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_IPSEC_MM_POLICY_PENDING_DELETION, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_IPSEC_MM_AUTH_PENDING_DELETION, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_IPSEC_QM_POLICY_PENDING_DELETION, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, WARNING_IPSEC_MM_POLICY_PRUNED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, WARNING_IPSEC_QM_POLICY_PRUNED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_IPSEC_IKE_NEG_STATUS_BEGIN, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_IPSEC_IKE_AUTH_FAIL, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_IPSEC_IKE_ATTRIB_FAIL, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_IPSEC_IKE_NEGOTIATION_PENDING, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_IPSEC_IKE_GENERAL_PROCESSING_ERROR, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_IPSEC_IKE_TIMED_OUT, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_IPSEC_IKE_NO_CERT, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_IPSEC_IKE_SA_DELETED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_IPSEC_IKE_SA_REAPED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_IPSEC_IKE_MM_ACQUIRE_DROP, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_IPSEC_IKE_QM_ACQUIRE_DROP, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_IPSEC_IKE_QUEUE_DROP_MM, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_IPSEC_IKE_QUEUE_DROP_NO_MM, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_IPSEC_IKE_DROP_NO_RESPONSE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_IPSEC_IKE_MM_DELAY_DROP, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_IPSEC_IKE_QM_DELAY_DROP, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_IPSEC_IKE_ERROR, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_IPSEC_IKE_CRL_FAILED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_IPSEC_IKE_INVALID_KEY_USAGE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_IPSEC_IKE_INVALID_CERT_TYPE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_IPSEC_IKE_NO_PRIVATE_KEY, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_IPSEC_IKE_SIMULTANEOUS_REKEY, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_IPSEC_IKE_DH_FAIL, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_IPSEC_IKE_CRITICAL_PAYLOAD_NOT_RECOGNIZED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_IPSEC_IKE_INVALID_HEADER, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_IPSEC_IKE_NO_POLICY, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_IPSEC_IKE_INVALID_SIGNATURE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_IPSEC_IKE_KERBEROS_ERROR, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_IPSEC_IKE_NO_PUBLIC_KEY, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_IPSEC_IKE_PROCESS_ERR, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_IPSEC_IKE_PROCESS_ERR_SA, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_IPSEC_IKE_PROCESS_ERR_PROP, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_IPSEC_IKE_PROCESS_ERR_TRANS, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_IPSEC_IKE_PROCESS_ERR_KE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_IPSEC_IKE_PROCESS_ERR_ID, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_IPSEC_IKE_PROCESS_ERR_CERT, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_IPSEC_IKE_PROCESS_ERR_CERT_REQ, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_IPSEC_IKE_PROCESS_ERR_HASH, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_IPSEC_IKE_PROCESS_ERR_SIG, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_IPSEC_IKE_PROCESS_ERR_NONCE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_IPSEC_IKE_PROCESS_ERR_NOTIFY, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_IPSEC_IKE_PROCESS_ERR_DELETE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_IPSEC_IKE_PROCESS_ERR_VENDOR, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_IPSEC_IKE_INVALID_PAYLOAD, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_IPSEC_IKE_LOAD_SOFT_SA, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_IPSEC_IKE_SOFT_SA_TORN_DOWN, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_IPSEC_IKE_INVALID_COOKIE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_IPSEC_IKE_NO_PEER_CERT, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_IPSEC_IKE_PEER_CRL_FAILED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_IPSEC_IKE_POLICY_CHANGE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_IPSEC_IKE_NO_MM_POLICY, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_IPSEC_IKE_NOTCBPRIV, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_IPSEC_IKE_SECLOADFAIL, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_IPSEC_IKE_FAILSSPINIT, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_IPSEC_IKE_FAILQUERYSSP, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_IPSEC_IKE_SRVACQFAIL, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_IPSEC_IKE_SRVQUERYCRED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_IPSEC_IKE_GETSPIFAIL, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_IPSEC_IKE_INVALID_FILTER, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_IPSEC_IKE_OUT_OF_MEMORY, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_IPSEC_IKE_ADD_UPDATE_KEY_FAILED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_IPSEC_IKE_INVALID_POLICY, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_IPSEC_IKE_UNKNOWN_DOI, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_IPSEC_IKE_INVALID_SITUATION, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_IPSEC_IKE_DH_FAILURE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_IPSEC_IKE_INVALID_GROUP, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_IPSEC_IKE_ENCRYPT, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_IPSEC_IKE_DECRYPT, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_IPSEC_IKE_POLICY_MATCH, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_IPSEC_IKE_UNSUPPORTED_ID, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_IPSEC_IKE_INVALID_HASH, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_IPSEC_IKE_INVALID_HASH_ALG, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_IPSEC_IKE_INVALID_HASH_SIZE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_IPSEC_IKE_INVALID_ENCRYPT_ALG, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_IPSEC_IKE_INVALID_AUTH_ALG, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_IPSEC_IKE_INVALID_SIG, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_IPSEC_IKE_LOAD_FAILED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_IPSEC_IKE_RPC_DELETE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_IPSEC_IKE_BENIGN_REINIT, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_IPSEC_IKE_INVALID_RESPONDER_LIFETIME_NOTIFY, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_IPSEC_IKE_INVALID_MAJOR_VERSION, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_IPSEC_IKE_INVALID_CERT_KEYLEN, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_IPSEC_IKE_MM_LIMIT, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_IPSEC_IKE_NEGOTIATION_DISABLED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_IPSEC_IKE_QM_LIMIT, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_IPSEC_IKE_MM_EXPIRED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_IPSEC_IKE_PEER_MM_ASSUMED_INVALID, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_IPSEC_IKE_CERT_CHAIN_POLICY_MISMATCH, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_IPSEC_IKE_UNEXPECTED_MESSAGE_ID, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_IPSEC_IKE_INVALID_UMATTS, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_IPSEC_IKE_DOS_COOKIE_SENT, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_IPSEC_IKE_SHUTTING_DOWN, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_IPSEC_IKE_CGA_AUTH_FAILED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_IPSEC_IKE_PROCESS_ERR_NATOA, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_IPSEC_IKE_INVALID_MM_FOR_QM, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_IPSEC_IKE_QM_EXPIRED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_IPSEC_IKE_TOO_MANY_FILTERS, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_IPSEC_IKE_NEG_STATUS_END, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_IPSEC_IKE_KILL_DUMMY_NAP_TUNNEL, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_IPSEC_IKE_INNER_IP_ASSIGNMENT_FAILURE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_IPSEC_IKE_REQUIRE_CP_PAYLOAD_MISSING, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_IPSEC_KEY_MODULE_IMPERSONATION_NEGOTIATION_PENDING, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_IPSEC_IKE_COEXISTENCE_SUPPRESS, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_IPSEC_IKE_RATELIMIT_DROP, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_IPSEC_IKE_PEER_DOESNT_SUPPORT_MOBIKE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_IPSEC_IKE_AUTHORIZATION_FAILURE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_IPSEC_IKE_STRONG_CRED_AUTHORIZATION_FAILURE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_IPSEC_IKE_AUTHORIZATION_FAILURE_WITH_OPTIONAL_RETRY, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_IPSEC_IKE_STRONG_CRED_AUTHORIZATION_AND_CERTMAP_FAILURE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_IPSEC_IKE_NEG_STATUS_EXTENDED_END, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_IPSEC_BAD_SPI, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_IPSEC_SA_LIFETIME_EXPIRED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_IPSEC_WRONG_SA, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_IPSEC_REPLAY_CHECK_FAILED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_IPSEC_INVALID_PACKET, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_IPSEC_INTEGRITY_CHECK_FAILED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_IPSEC_CLEAR_TEXT_DROP, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_IPSEC_AUTH_FIREWALL_DROP, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_IPSEC_THROTTLE_DROP, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_IPSEC_DOSP_BLOCK, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_IPSEC_DOSP_RECEIVED_MULTICAST, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_IPSEC_DOSP_INVALID_PACKET, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_IPSEC_DOSP_STATE_LOOKUP_FAILED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_IPSEC_DOSP_MAX_ENTRIES, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_IPSEC_DOSP_KEYMOD_NOT_ALLOWED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_IPSEC_DOSP_NOT_INSTALLED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_IPSEC_DOSP_MAX_PER_IP_RATELIMIT_QUEUES, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_SXS_SECTION_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_SXS_CANT_GEN_ACTCTX, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_SXS_INVALID_ACTCTXDATA_FORMAT, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_SXS_ASSEMBLY_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_SXS_MANIFEST_FORMAT_ERROR, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_SXS_MANIFEST_PARSE_ERROR, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_SXS_ACTIVATION_CONTEXT_DISABLED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_SXS_KEY_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_SXS_VERSION_CONFLICT, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_SXS_WRONG_SECTION_TYPE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_SXS_THREAD_QUERIES_DISABLED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_SXS_PROCESS_DEFAULT_ALREADY_SET, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_SXS_UNKNOWN_ENCODING_GROUP, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_SXS_UNKNOWN_ENCODING, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_SXS_INVALID_XML_NAMESPACE_URI, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_SXS_ROOT_MANIFEST_DEPENDENCY_NOT_INSTALLED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_SXS_LEAF_MANIFEST_DEPENDENCY_NOT_INSTALLED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_SXS_INVALID_ASSEMBLY_IDENTITY_ATTRIBUTE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_SXS_MANIFEST_MISSING_REQUIRED_DEFAULT_NAMESPACE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_SXS_MANIFEST_INVALID_REQUIRED_DEFAULT_NAMESPACE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_SXS_PRIVATE_MANIFEST_CROSS_PATH_WITH_REPARSE_POINT, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_SXS_DUPLICATE_DLL_NAME, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_SXS_DUPLICATE_WINDOWCLASS_NAME, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_SXS_DUPLICATE_CLSID, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_SXS_DUPLICATE_IID, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_SXS_DUPLICATE_TLBID, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_SXS_DUPLICATE_PROGID, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_SXS_DUPLICATE_ASSEMBLY_NAME, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_SXS_FILE_HASH_MISMATCH, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_SXS_POLICY_PARSE_ERROR, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_SXS_XML_E_MISSINGQUOTE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_SXS_XML_E_COMMENTSYNTAX, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_SXS_XML_E_BADSTARTNAMECHAR, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_SXS_XML_E_BADNAMECHAR, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_SXS_XML_E_BADCHARINSTRING, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_SXS_XML_E_XMLDECLSYNTAX, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_SXS_XML_E_BADCHARDATA, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_SXS_XML_E_MISSINGWHITESPACE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_SXS_XML_E_EXPECTINGTAGEND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_SXS_XML_E_MISSINGSEMICOLON, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_SXS_XML_E_UNBALANCEDPAREN, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_SXS_XML_E_INTERNALERROR, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_SXS_XML_E_UNEXPECTED_WHITESPACE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_SXS_XML_E_INCOMPLETE_ENCODING, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_SXS_XML_E_MISSING_PAREN, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_SXS_XML_E_EXPECTINGCLOSEQUOTE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_SXS_XML_E_MULTIPLE_COLONS, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_SXS_XML_E_INVALID_DECIMAL, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_SXS_XML_E_INVALID_HEXIDECIMAL, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_SXS_XML_E_INVALID_UNICODE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_SXS_XML_E_WHITESPACEORQUESTIONMARK, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_SXS_XML_E_UNEXPECTEDENDTAG, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_SXS_XML_E_UNCLOSEDTAG, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_SXS_XML_E_DUPLICATEATTRIBUTE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_SXS_XML_E_MULTIPLEROOTS, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_SXS_XML_E_INVALIDATROOTLEVEL, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_SXS_XML_E_BADXMLDECL, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_SXS_XML_E_MISSINGROOT, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_SXS_XML_E_UNEXPECTEDEOF, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_SXS_XML_E_BADPEREFINSUBSET, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_SXS_XML_E_UNCLOSEDSTARTTAG, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_SXS_XML_E_UNCLOSEDENDTAG, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_SXS_XML_E_UNCLOSEDSTRING, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_SXS_XML_E_UNCLOSEDCOMMENT, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_SXS_XML_E_UNCLOSEDDECL, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_SXS_XML_E_UNCLOSEDCDATA, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_SXS_XML_E_RESERVEDNAMESPACE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_SXS_XML_E_INVALIDENCODING, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_SXS_XML_E_INVALIDSWITCH, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_SXS_XML_E_BADXMLCASE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_SXS_XML_E_INVALID_STANDALONE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_SXS_XML_E_UNEXPECTED_STANDALONE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_SXS_XML_E_INVALID_VERSION, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_SXS_XML_E_MISSINGEQUALS, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_SXS_PROTECTION_RECOVERY_FAILED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_SXS_PROTECTION_PUBLIC_KEY_TOO_SHORT, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_SXS_PROTECTION_CATALOG_NOT_VALID, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_SXS_UNTRANSLATABLE_HRESULT, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_SXS_PROTECTION_CATALOG_FILE_MISSING, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_SXS_MISSING_ASSEMBLY_IDENTITY_ATTRIBUTE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_SXS_INVALID_ASSEMBLY_IDENTITY_ATTRIBUTE_NAME, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_SXS_ASSEMBLY_MISSING, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_SXS_CORRUPT_ACTIVATION_STACK, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_SXS_CORRUPTION, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_SXS_EARLY_DEACTIVATION, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_SXS_INVALID_DEACTIVATION, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_SXS_MULTIPLE_DEACTIVATION, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_SXS_PROCESS_TERMINATION_REQUESTED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_SXS_RELEASE_ACTIVATION_CONTEXT, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_SXS_SYSTEM_DEFAULT_ACTIVATION_CONTEXT_EMPTY, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_SXS_INVALID_IDENTITY_ATTRIBUTE_VALUE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_SXS_INVALID_IDENTITY_ATTRIBUTE_NAME, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_SXS_IDENTITY_DUPLICATE_ATTRIBUTE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_SXS_IDENTITY_PARSE_ERROR, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_MALFORMED_SUBSTITUTION_STRING, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_SXS_INCORRECT_PUBLIC_KEY_TOKEN, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_UNMAPPED_SUBSTITUTION_STRING, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_SXS_ASSEMBLY_NOT_LOCKED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_SXS_COMPONENT_STORE_CORRUPT, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_ADVANCED_INSTALLER_FAILED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_XML_ENCODING_MISMATCH, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_SXS_MANIFEST_IDENTITY_SAME_BUT_CONTENTS_DIFFERENT, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_SXS_IDENTITIES_DIFFERENT, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_SXS_ASSEMBLY_IS_NOT_A_DEPLOYMENT, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_SXS_FILE_NOT_PART_OF_ASSEMBLY, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_SXS_MANIFEST_TOO_BIG, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_SXS_SETTING_NOT_REGISTERED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_SXS_TRANSACTION_CLOSURE_INCOMPLETE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_EVT_INVALID_CHANNEL_PATH, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_EVT_INVALID_QUERY, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_EVT_PUBLISHER_METADATA_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_EVT_EVENT_TEMPLATE_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_EVT_INVALID_PUBLISHER_NAME, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_EVT_INVALID_EVENT_DATA, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_EVT_CHANNEL_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_EVT_MALFORMED_XML_TEXT, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_EVT_SUBSCRIPTION_TO_DIRECT_CHANNEL, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_EVT_CONFIGURATION_ERROR, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_EVT_QUERY_RESULT_STALE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_EVT_QUERY_RESULT_INVALID_POSITION, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_EVT_NON_VALIDATING_MSXML, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_EVT_FILTER_ALREADYSCOPED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_EVT_FILTER_NOTELTSET, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_EVT_FILTER_INVARG, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_EVT_FILTER_INVTEST, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_EVT_FILTER_INVTYPE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_EVT_FILTER_PARSEERR, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_EVT_FILTER_UNSUPPORTEDOP, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_EVT_FILTER_UNEXPECTEDTOKEN, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_EVT_INVALID_OPERATION_OVER_ENABLED_DIRECT_CHANNEL, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_EVT_INVALID_CHANNEL_PROPERTY_VALUE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_EVT_INVALID_PUBLISHER_PROPERTY_VALUE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_EVT_CHANNEL_CANNOT_ACTIVATE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_EVT_FILTER_TOO_COMPLEX, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_EVT_MESSAGE_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_EVT_MESSAGE_ID_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_EVT_UNRESOLVED_VALUE_INSERT, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_EVT_UNRESOLVED_PARAMETER_INSERT, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_EVT_MAX_INSERTS_REACHED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_EVT_EVENT_DEFINITION_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_EVT_MESSAGE_LOCALE_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_EVT_VERSION_TOO_OLD, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_EVT_VERSION_TOO_NEW, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_EVT_CANNOT_OPEN_CHANNEL_OF_QUERY, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_EVT_PUBLISHER_DISABLED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_EVT_FILTER_OUT_OF_RANGE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_EC_SUBSCRIPTION_CANNOT_ACTIVATE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_EC_LOG_DISABLED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_EC_CIRCULAR_FORWARDING, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_EC_CREDSTORE_FULL, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_EC_CRED_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_EC_NO_ACTIVE_CHANNEL, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_MUI_FILE_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_MUI_INVALID_FILE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_MUI_INVALID_RC_CONFIG, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_MUI_INVALID_LOCALE_NAME, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_MUI_INVALID_ULTIMATEFALLBACK_NAME, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_MUI_FILE_NOT_LOADED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_RESOURCE_ENUM_USER_STOP, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_MUI_INTLSETTINGS_UILANG_NOT_INSTALLED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_MUI_INTLSETTINGS_INVALID_LOCALE_NAME, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_MCA_INVALID_CAPABILITIES_STRING, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_MCA_INVALID_VCP_VERSION, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_MCA_MONITOR_VIOLATES_MCCS_SPECIFICATION, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_MCA_MCCS_VERSION_MISMATCH, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_MCA_UNSUPPORTED_MCCS_VERSION, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_MCA_INTERNAL_ERROR, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_MCA_INVALID_TECHNOLOGY_TYPE_RETURNED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_MCA_UNSUPPORTED_COLOR_TEMPERATURE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_AMBIGUOUS_SYSTEM_DEVICE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_SYSTEM_DEVICE_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DHCP_SUBNET_NOT_PRESENT, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DHCP_ELEMENT_CANT_REMOVE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DHCP_OPTION_EXITS, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DHCP_OPTION_NOT_PRESENT, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DHCP_ADDRESS_NOT_AVAILABLE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DHCP_JET_ERROR, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DHCP_CLIENT_EXISTS, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DHCP_NOT_RESERVED_CLIENT, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DHCP_RESERVED_CLIENT, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DHCP_IPRANGE_EXITS, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DHCP_RESERVEDIP_EXITS, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DHCP_INVALID_RANGE, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DHCP_CLASS_NOT_FOUND, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DHCP_CLASS_ALREADY_EXISTS, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DHCP_SCOPE_NAME_TOO_LONG, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DHCP_IPRANGE_CONV_ILLEGAL, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DHCP_NETWORK_CHANGED, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DHCP_CANNOT_MODIFY_BINDINGS, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DHCP_SUBNET_EXISTS, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_DHCP_MSCOPE_EXISTS, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, ERROR_MSCOPE_RANGE_TOO_SMALL, ERRSRV, ERRerror, -1, "" )
S ( LW_STATUS_MESSAGE_NOT_FOUND, SEC_E_NO_CREDENTIALS, ERRSRV, ERRerror, -1, "" )

/* End autogenerated portion */

#undef S
#endif


/*
local variables:
mode: c
c-basic-offset: 4
indent-tabs-mode: nil
tab-width: 4
end:
*/
